////////////////////////////////////////////////////////////////////////////
// Title         : HT Common
// Project       : HT Testbench
////////////////////////////////////////////////////////////////////////////
// File          : csco_ht_common.e
// Author        : Brian Van Essen  <vanessen@cisco.com>
// Created       : 04.05.2001
// Last modified : 04.05.2001
////////////////////////////////////////////////////////////////////////////
// Description :
// A place to define many of the common elements to the ht testbench.
// Error handlind and reporting is define here.
// Scenario frameworks and fundamentals are defined here.
// Test Control is also defined here.
// Scoreboard is defined here.
////////////////////////////////////////////////////////////////////////////////
// Modification history :
// 04.05.2001 : created
// 04.22.2002 : fixed  Obsolete field access syntax i.e. ' casting. for Specman 4.0
////////////////////////////////////////////////////////////////////////////

// ***** BEGIN LICENSE BLOCK *****
// Version: MPL 1.1
// The contents of this file are subject to the Mozilla Public License Version
// 1.1 (the "License"); you may not use this file except in compliance with
// the License. You may obtain a copy of the License at
// http:                                                   // www.mozilla.org/MPL/ or ./mpl_1_1.txt
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
// for the specific language governing rights and limitations under the License.
// The Original Code is Cisco Hypertransport eVC.
// The Initial Developer of the Original Code is
// Brian Van Essen, Lisa Huang, and Sean Smith.
// Portions created by the Initial Developer are Copyright (C) 2002
// the Initial Developer. All Rights Reserved.
// Contributor(s): Brian Van Essen, Lisa Huang, and Sean Smith
// ***** END LICENSE BLOCK *****

<'
////////////////////////////////////////////////////////////////////////////////
// extend sys
// Set up the basic error control and test control objects.
////////////////////////////////////////////////////////////////////////////////
extend sys {
  // This is a flag that will indicate that the clock provided to the 'e' model is clock that is running at twice
  // the desired link frequency.  One effect of this is that the model will not be able to dynamically change link
  // speed during the simulation.  Any link frequency shift will have to be done by the clock source.
  // If this flag is false, then a clock that is 1/12th ns needs to be provided so that it can be multiplied to achieve
  // any one of the legal HT values.  (i.e. 200MHz, 300MHz, 400MHz, 500MH, 600MHz, 800MHz, 1000MHz).
  csco_ht_use_direct_csco_ht_clock : bool;
    keep soft csco_ht_use_direct_csco_ht_clock == TRUE;


  // Flags used to indicate if the testbench is connected to a verilog DUT or is running against itself.
  csco_ht_development_test : bool;
    keep soft me.csco_ht_development_test == FALSE;

  csco_ht_using_verilog_simulator : bool;
    keep soft me.csco_ht_using_verilog_simulator != me.csco_ht_development_test;


  csco_ht_test_control : csco_ht_test_control_block is instance;

  csco_ht_errors : csco_ht_error_handler is instance;



  event csco_ht_sys_power_change;
  event csco_ht_sys_reset_change;
  csco_ht_sys_PWROK : bool;
  csco_ht_sys_RESET_L : bool;


};
////////////////////////////////////////////////////////////////////////////////
// extend the packing to support big endian and big endian w/ bit matching
////////////////////////////////////////////////////////////////////////////////
extend packing {
  // new packing options to handle different endianness
  // csco_ht_big_endian pack takes data in big endian (w/ byte match lane)
  // format and swap the bytes accordingly so it is sent out on
  // the bus correctly
  //
  //   63                                    0
  //  |////|////|////|////|////|////|////|//               // |
  //  | B0 | B1 | B2 | B3 | B4 | B5 | B6 | B7 |
  //  |////|////|////|////|////|////|////|////|
  //
  //           |////|////|////|////|
  //  HT bus  | B3 | B2 | B1 | B0 |
  //  A[2] = 0 |////|////|////|////|
  //
  //           |////|////|////|////|
  //  HT bus  | B7 | B6 | B5 | B4 |
  //  A[2] = 1 |////|////|////|////|
  csco_ht_big_endian           : pack_options;

  // csco_ht_big_endian_bit pack takes data in big endian w/ bit match lane
  // format and swaps the bytes accordingly so it is sent out on
  // the bus correctly
  //
  //   63                                    0
  //  |////|////|////|////|////|////|////|////|
  //  | B0 | B1 | B2 | B3 | B4 | B5 | B6 | B7 |
  //  |////|////|////|////|////|////|////|////|
  //
  //           |////|////|////|////|
  //  HT bus  | B0 | B1 | B2 | B3 |
  //  A[2] = 0 |////|////|////|////|
  //
  //           |////|////|////|////|
  //  HT bus  | B4 | B5 | B6 | B7 |
  //  A[2] = 1 |////|////|////|////|
  csco_ht_big_endian_bit_match : pack_options;

  init() is also {
    csco_ht_big_endian = low.copy();
    csco_ht_big_endian.reverse_fields = FALSE;
    csco_ht_big_endian.reverse_list_items = FALSE;
    csco_ht_big_endian.scalar_reorder = {};
    csco_ht_big_endian.final_reorder = {8; UNDEF;};

    csco_ht_big_endian_bit_match = low.copy();
    csco_ht_big_endian_bit_match.reverse_fields = FALSE;
    csco_ht_big_endian_bit_match.reverse_list_items = FALSE;
    csco_ht_big_endian_bit_match.scalar_reorder = {};
    csco_ht_big_endian_bit_match.final_reorder = {32; UNDEF;};
  };
};
////////////////////////////////////////////////////////////////////////////////
// Scenarios - A directed way to provide random traffic generation.
////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
// csco_ht_scenario provides a basic framework for a device to generate traffic.
////////////////////////////////////////////////////////////////////////////////
struct csco_ht_scenario {
  parent_device : csco_ht_device;
  keep parent_device == get_enclosing_unit(csco_ht_device);
  !channel : csco_ht_channel_interface;
  kind : csco_ht_scenario_kind;
  done : bool;
  keep done == FALSE;
  // All address and offsets must be 4 byte aligned, as the HT protocol cannot handle any finer granularity.
  addr_map : csco_ht_address_map;
  // soft constrain the memory map to something small
  keep soft addr_map.base_addr == 40'h00_0000_0000;
  keep soft addr_map.top_addr == 40'h00_0000_00ff;

  // The dense_target_address_space flag is used to determine if the address space should be exhaustively tested, or sparsely.
  // The default is to test it sparsely.
  dense_target_address_space : bool;
  keep soft dense_target_address_space == FALSE;
  allow_gen_repeat_Address: bool;
  keep soft allow_gen_repeat_Address == FALSE;

  max_num_transactions : uint (bits:64);
  keep allow_gen_repeat_Address == FALSE => max_num_transactions == ((addr_map.top_addr - addr_map.base_addr)>>2)+1;
  num_transactions : uint (bits:64);
  keep gen (dense_target_address_space) before (num_transactions);
  keep num_transactions <= value(max_num_transactions);
  keep num_transactions > 0;
  // In the case of a dense address space set the number or transactions to the maximum possible so that the the
  // legal address region will be exhausted.
  keep (dense_target_address_space == TRUE) => (num_transactions == value(max_num_transactions));
  // Maintain a list of all legal addresses within the address map.
  keep soft num_transactions == select {
    10 : pass; // Allow a normal distribution of number of transactions
    90 : [10..200]; // Normally, try to keep the number of transactions in a reasonable limit
  };

  legal_address_region : list of csco_ht_address_map;
  keep legal_address_region.size() == 0;
  post_generate() is also {
    legal_address_region.add(addr_map.copy());
  };

  // Provide delay constraints to control the rate of traffic
  !min_random_tx_delay : uint;
  max_tx_delay : uint; // The max_tx_delay is a multiplier for the min_random_tx_delay
  keep max_tx_delay >= 2;
  keep soft max_tx_delay >= 3;
  keep soft max_tx_delay <= 5;

  target_address_list : list of csco_ht_scenario_element;
  keep soft target_address_list.size() == 0;
  // This is a flag that will inhibit the generation of the target address list. If that list is generated, then don't
  // allow the automatic generation to occur.
  target_address_gen_complete : bool;
  keep soft target_address_gen_complete == FALSE;
  keep target_address_list.size() != 0 => target_address_gen_complete == TRUE;
  !transaction_history : list of csco_ht_packet;
  ////////////////////////////////
  // for random
  ////////////////////////////////
  random_coherent_percent: uint;
  keep random_coherent_percent in [0..100];
  keep soft random_coherent_percent == 0;
  random_isochronous_percent: uint;
  keep random_isochronous_percent in [0..100];
  keep soft random_isochronous_percent == 0;
  random_NonPosted_request_percent: uint;
  keep random_NonPosted_request_percent in [0..100];
  keep soft random_NonPosted_request_percent == 0;
  random_Address_aligned: bool;
  keep soft random_Address_aligned == FALSE;
  random_fence_percent: uint;
  keep random_fence_percent in [0..100];
  random_flush_percent: uint;
  keep random_flush_percent in [0..100];
  !fence_count: uint;
  !flush_count: uint;
  !calPercentFence: uint;
  !calPercentFlush: uint;


  !all_uids: list (key:it) of uint;
  !listOfAccessType: list of csco_ht_mem_access_type;;
  !listOfBWLatencyReq: list of csco_ht_BW_latency_req;
  !listOfVChanType: list of csco_ht_virtual_channel;
  !randomIndex:uint;
  ////////////////////////////////////////////////////////////////////////////////////////////////////////////-
  //get_next_uid() - This method will generate a random number. If this number already exists in the
  //list continue generating random numbers. If it is a new number - add it to the list, and return this value.
  ////////////////////////////////////////////////////////////////////////////////////////////////////////////-
  get_next_uid(): uint  is {
    var next_uid : uint;
    result = 0xffff_ffff;
    while result == 0xffff_ffff {
      gen next_uid keeping {it < num_transactions; };
      if not all_uids.key_exists(next_uid) then {
        all_uids.add(next_uid);
        result = next_uid;
        //print all_uids;
      };
    };
  };


  ////////////////////////////////////////////////////////////////////////////////
  gen_target_address_list() is {
    var current_target : csco_ht_scenario_element;
    var current_addr : uint (bits:40);

    if(target_address_gen_complete == TRUE) {
      // The list has already been generated, don't do anything.
      return;
    };
    var temp_address_list : list of csco_ht_scenario_element;
    var temp_scenario_element : csco_ht_scenario_element;
    var tmp1_address_list : list of csco_ht_scenario_element;
    while(legal_address_region.size() > 0 && temp_address_list.size() < num_transactions) {
      temp_scenario_element = gen_random_address();
      temp_address_list.add(temp_scenario_element);
      if (allow_gen_repeat_Address == FALSE) {
        tmp1_address_list = temp_address_list.sort(it.addr).unique(it.addr);
      } else {
        tmp1_address_list = temp_address_list;
      };
    };
    //target_address_list = temp_address_list.sort(it.addr);
    target_address_list = tmp1_address_list;
  };
  ////////////////////////////////////////////////////////////////////////////////
  gen_random_address():csco_ht_scenario_element is {
    var current_target : csco_ht_scenario_element;
    var current_address_region : csco_ht_address_map;
    var addr_range_index : uint;
    var lower_address_map : csco_ht_address_map;
    var upper_address_map : csco_ht_address_map;

    gen addr_range_index keeping {
      it >= 0;
      it < legal_address_region.size();
    };
    current_address_region = legal_address_region[addr_range_index];

    if (allow_gen_repeat_Address == FALSE) {
      legal_address_region.delete(addr_range_index);
    };
    gen current_target keeping {
      it.addr >= current_address_region.base_addr;
      it.addr <= current_address_region.top_addr;
      //it.tmp == current_address_region.top_addr+1;
      //it.tmp1 == (it.tmp - it.addr)>>2 - 1;
      it.tmp1 <= ((current_address_region.top_addr+1-it.addr)>>2)-1;
      it.tmp1 <= 0 => it.num_doublewords == 0;
      it.tmp1 >  0 => it.num_doublewords == it.tmp1;
      //it.num_doublewords <= ((current_address_region.top_addr+1-it.addr)>>2)-1;
      //it.num_doublewords <= it.tmp1;
    };
    if(current_target.addr == current_address_region.base_addr
       && ((current_target.addr + ((current_target.num_doublewords+1)<<2))-1 == current_address_region.top_addr)) {
         // The whole memory region is consumed by this transaction
    }else if(current_target.addr == current_address_region.base_addr) {
    if current_target.num_doublewords == 0 then {
        gen upper_address_map keeping {
          it.base_addr == current_target.addr;
          it.top_addr == current_address_region.top_addr;
        };
      } else {
          gen upper_address_map keeping {
          it.base_addr == value(current_target.addr + ((current_target.num_doublewords+1)<<2));
          it.top_addr == current_address_region.top_addr;
        };
      };
      legal_address_region.add(upper_address_map);
    }else if((current_target.addr + ((current_target.num_doublewords+1)<<2))-1 == current_address_region.top_addr) {
      gen lower_address_map keeping {
        it.base_addr == current_address_region.base_addr;
        it.top_addr == value(current_target.addr - 1);
      };
      legal_address_region.add(lower_address_map);
    }else {
      gen lower_address_map keeping {
        it.base_addr == current_address_region.base_addr;
        it.top_addr == value(current_target.addr - 1);
      };
      if current_target.num_doublewords == 0 then {
        gen upper_address_map keeping {
          it.base_addr == current_target.addr;
          it.top_addr == current_address_region.top_addr;
        };
      } else {
          gen upper_address_map keeping {
          it.base_addr == value(current_target.addr + ((current_target.num_doublewords+1)<<2));
          it.top_addr == current_address_region.top_addr;
        };
      };
      legal_address_region.add(lower_address_map);
      legal_address_region.add(upper_address_map);
    };
    result = current_target;
  };
  ////////////////////////////////////////////////////////////////////////////////
  gen_replacement_address(transaction_size : uint (bits:4), current_target : csco_ht_scenario_element) is {
    if(transaction_size < current_target.num_doublewords) {
      var replacement_target : csco_ht_scenario_element;
      var replacement_addr : uint (bits:40);

      replacement_addr = current_target.addr+((transaction_size+1)<<2);

      gen replacement_target keeping {
        //random_Address_aligned == TRUE => it.addr[2:2] == 1'b0;
        it.addr == replacement_addr;
        it.num_doublewords == current_target.num_doublewords - (transaction_size+1);
      };
      target_address_list.push0(replacement_target); // Push the remainding address block back onto the top of the list.
      var AccessType: csco_ht_mem_access_type;
      var BWLatencyReq: csco_ht_BW_latency_req;
      var vChan: csco_ht_virtual_channel;
      num_transactions += 1;
      gen AccessType keeping {it == NonCoherent;};
      listOfAccessType.add(AccessType);
      gen BWLatencyReq keeping {it == Normal;};
      listOfBWLatencyReq.add(BWLatencyReq);
      gen vChan keeping {it == Posted;};
      listOfVChanType.add(vChan);
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  schedule_transaction(transaction : csco_ht_packet) is {
    parent_device.schedule_packet(channel,transaction );
    //parent_device.schedule_packet(channel, transaction);
    transaction_history.add(transaction);
  };
  ////////////////////////////////////////////////////////////////////////////////
  run_scenario(channel : csco_ht_channel_interface)@parent_device.csco_ht_device_clock is {
    me.channel = channel;
    // Before starting a new scenario wait for the tx_scheduling queue to empty and the scoreboards are empty.
    // Let the entire system flush before starting the new scenario.
    emit sys.csco_ht_test_control.csco_ht_model_initialized;
    emit sys.csco_ht_test_control.csco_ht_start_traffic;
    wait true(channel.tx_scheduling_queue.size() == 0 &&
              parent_device.scoreboard.pending_transactions.is_empty() && parent_device.scoreboard.completed_transactions.is_empty());
    messagef(HIGH, "beginning %s scenario", kind);
    gen_target_address_list();

    var max_random_tx_delay : uint;
    var queue_depth : uint;
    var delay : uint;
    min_random_tx_delay = 32/(parent_device.control_block.read_current_tx_width(channel.channel_identifier).as_a(uint));
    max_random_tx_delay = max_tx_delay * min_random_tx_delay;
    //////////////////////////////////////////////////////////-
    // for random
    //////////////////////////////////////////////////////////-
    var j:uint=0;
    var k:uint=0;
    var n:uint=0;
    var AccessType: csco_ht_mem_access_type;
    var BWLatencyReq: csco_ht_BW_latency_req;
    var vChan: csco_ht_virtual_channel;
    var calPercentAccessType: uint;
    var calPercentBWLatencyReq: uint;
    var calPercentVChan: uint;
    calPercentAccessType = (random_coherent_percent * num_transactions)/100;
    calPercentBWLatencyReq = (random_isochronous_percent * num_transactions)/100;
    calPercentVChan = (random_NonPosted_request_percent * num_transactions)/100;
    calPercentFence =  (random_fence_percent * num_transactions)/100;
    calPercentFlush =  (random_flush_percent * num_transactions)/100;

    for i from 1 to num_transactions do {
      if j<calPercentAccessType then {
        gen AccessType keeping {it == Coherent;};
        listOfAccessType.add(AccessType);
      } else {
        gen AccessType keeping {it == NonCoherent;};
        listOfAccessType.add(AccessType);
      };
      j += 1;
      if k<calPercentBWLatencyReq then {
        gen BWLatencyReq keeping {it == Isochronous;};
        listOfBWLatencyReq.add(BWLatencyReq);
      } else {
        gen BWLatencyReq keeping {it == Normal;};
        listOfBWLatencyReq.add(BWLatencyReq);
      };
      k += 1;
      if n<calPercentVChan then {
        gen vChan keeping {it == NonPosted;};
        listOfVChanType.add(vChan);
      } else {
        gen vChan keeping {it == Posted;};
        listOfVChanType.add(vChan);
      };
      n += 1;
    };

    while(!done) {
      //randomIndex = get_next_uid();
      wait true(channel.tx_scheduling_queue.size()<=parent_device.TxQueueDepth);
      generate_transactions();
      // Wait for some random amount of time before creating the next transactions.  To provide a base value
      // for the delay use the depth of the tx_scheduling_queue, if this is not done, then the queue will quickly
      // reach an obscene depth as it takes at least 4 cycles for a transaction to be transmitted on a 8 bit link.
      queue_depth = channel.tx_scheduling_queue.size();
      gen delay keeping {
        it >= queue_depth + min_random_tx_delay;
        it <= queue_depth + max_random_tx_delay;
      };
      wait [delay] * cycle;
    };
    messagef(HIGH, "completing %s scenario", kind);
  };
  ////////////////////////////////////////////////////////////////////////////////
  // init_scenario() is a method that is used to start the run_scenario.  It provides
  // a wrapper functionality that allows a scenario to be aborted by the reset sequence
  // at any time.
  ////////////////////////////////////////////////////////////////////////////////
  init_scenario(channel : csco_ht_channel_interface)@parent_device.csco_ht_device_clock is {
    first of {
      {
        // Have the TCM monitor and wait for a reset signal.  This will allow the
        // device to reinitialize itself at any time.  Otherwise, the TCM could get stuck waiting
        // for an event that may never happen.
        wait @parent_device.reset @parent_device.reset;
      };
      {
        // Call the scenario's run method.
        run_scenario(channel);
      };
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  generate_transactions()@parent_device.csco_ht_device_clock is empty;

  ////////////////////////////////////////////////////////////////////////////////
//KMM added an empty definition for 4.x compatability

  initialize_csco_ht_chain()@parent_device.csco_ht_device_clock is empty;
////////////////////////////////////////////////////////////////

};
////////////////////////////////////////////////////////////////////////////////
// A structure used to fragment the target address space into regions appropriate for HT requests.
////////////////////////////////////////////////////////////////////////////////
struct csco_ht_scenario_element {
  scenario: csco_ht_scenario;
  addr : uint (bits:40);
  !tmp : uint (bits:40);
  tmp1 : int (bits:4);
  keep addr[1:0] == 2'b00;
  keep scenario.random_Address_aligned == TRUE => soft addr[2:2] == 1'b0;
  num_doublewords : uint (bits:4); // This is a 0 based number so that 0 indicates that there is 1 DW of data.
};
////////////////////////////////////////////////////////////////////////////////
// Define the address space that the HT testbench can interact with.
////////////////////////////////////////////////////////////////////////////////
struct csco_ht_address_map {
  // An address range for HT must start on a doubleword boundary and include
  // all 4 bytes in that range.
  region : csco_ht_address_region;
  keep soft region == Generic;
  base_addr : uint (bits:40); // First legal address in region
  keep base_addr[1:0] == 2'b00;
  top_addr : uint (bits:40); // Last legal address in region
  keep soft top_addr[1:0] == 2'b11;
  keep top_addr >= base_addr;

  endian_type : csco_ht_endian_type; // endian type of this region

  when Generic'region csco_ht_address_map { // Used for any non-specific region
  };
  when Memory_Mapped_IO'region csco_ht_address_map {
    keep base_addr == 40'h00_0000_0000;
    keep top_addr  == 40'hFC_FFFF_FFFF;
  };
  when Interrupts'region csco_ht_address_map {
    keep base_addr == 40'hFD_F800_0000;
    keep top_addr  == 40'hFD_F8FF_FFFF;
  };
  when IACK'region csco_ht_address_map {
    keep base_addr == 40'hFD_F900_0000;
    keep top_addr  == 40'hFD_F90F_FFFF;
  };
  when System_Management'region csco_ht_address_map {
    keep base_addr == 40'hFD_F910_0000;
    keep top_addr  == 40'hFD_F91F_FFFF;
  };
  when PCI_IO'region csco_ht_address_map {
    keep base_addr == 40'hFD_FC00_0000;
    keep top_addr  == 40'hFD_FDFF_FFFF;
  };
  when PCI_Configuration'region csco_ht_address_map {
    keep base_addr == 40'hFD_FE00_0000;
    keep top_addr  == 40'hFD_FFFF_FFFF;
  };
};
////////////////////////////////////////////////////////////////////////////////
// Scoreboard - a method of managing and tracking device traffic.  Allows the testbench
// to create self checking tests.
////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
// csco_ht_scoreboard provides all of the methods for tracking request and responses.  Additionally
// it will check the data so that self checking test can be created.
////////////////////////////////////////////////////////////////////////////////
struct csco_ht_scoreboard {
  parent_device : csco_ht_device;
  keep parent_device == get_enclosing_unit(csco_ht_device);

  // item which to score or to be added to the response or broadcast
  // scoreboard so that the expected response can be tracked
  //!score_packet : csco_ht_packet;

  !pending_transactions : list (key:token) of csco_ht_transaction;
  !completed_transactions : list (key:token) of csco_ht_transaction;

  // A current copy of the memory outside of this device.  For the Tunnel or EOC devices, this is an expected copy of the Host's
  // memory, based upon the transactions sent by this device.  For the Host this should be an aggregated copy of the memory
  // of all devices on the chain, based upon the transactions sent by the Host.  This structure is used for self checking write
  // requests that are sent out.
  !shadow_memory : list (key:addr) of csco_ht_memory_element;

  // fixme!!!
  data_scoreboarding : bool; // If this flag is set to false no data will be scoreboarded or checked.
  keep soft data_scoreboarding == TRUE;

  // define level of scoreboarding to use, soft constrain this based on data_scoreboarding flag
  // for backward compatibility
  score_level : csco_ht_scoreboard_level;
  keep (data_scoreboarding) => (score_level == csco_ht_full_scoring);
  keep soft score_level == csco_ht_register_scoring;
  ////////////////////////////////////////////////////////////////////////////////
  // reset() reset all of the lists
  ////////////////////////////////////////////////////////////////////////////////
  reset() is {
    pending_transactions = {};
    completed_transactions = {};
    shadow_memory = {};
  };
  ////////////////////////////////////////////////////////////////////////////////
  // add_request() is called by score() to add the packet to either the response
  // or broadcase scoreboard
  ////////////////////////////////////////////////////////////////////////////////
  add_request(packet : csco_ht_packet) is inline {
    var new_transaction : csco_ht_transaction;
    gen new_transaction keeping {
      .request == packet;
    };
    if(packet.command() == Broadcast) {
      check that (parent_device.device_function == Host_Bridge) else
      dut_error("A HT device other than a Host Bridge is trying to issue an HT Broadcast.");
        //KMM explicit type cast to subtype Host_Bridge'device_function csco_ht_device
      new_transaction.response_count = parent_device.as_a(Host_Bridge'device_function csco_ht_device).chain_topology.size();
    };
    if(transaction_pending(new_transaction.token) == TRUE) {
      // There already exists a transaction with this unique UnitID/SrcTag Pair.
      dut_error("Attempting to scoreboard a request with a non-unique transaction token (UnitID/SrcTag) pair.");
    };
    pending_transactions.add(new_transaction);
  };
  ////////////////////////////////////////////////////////////////////////////////
  // Return NO_ERROR if the response was successfully added to the scoreboard.
  // If the corresponding request does not exist, return UNMATCHED_RESPONSE.
  ////////////////////////////////////////////////////////////////////////////////
  add_response(packet : csco_ht_packet): csco_ht_error_code is {
    check that (packet.vChan() == Response) else
    dut_error("A request packet was being added to the response portion of the HT scoreboard.");

    var token : uint (bits:10);
    token = packet.generate_token();

    // FIXME!!!! change this to use some other defined flag???
    // if scoreboard level is not set to score all transactions, then
    // just return since the transaction will not be on the pending queue
    if((score_level != csco_ht_full_scoring) && !pending_transactions.key_exists(token)) {
      return(NO_ERROR);
    };

    if(pending_transactions.key_exists(token)) {
      var temp_transaction : csco_ht_transaction = pending_transactions.key(token);
      check that (temp_transaction.response.size() < temp_transaction.response_count) else
      dut_error("Attempting to add a response to a transaction where response_count is exceeded.");
      temp_transaction.response.add(packet);
      if(temp_transaction.response.size() == temp_transaction.response_count) {
        if(completed_transactions.key_exists(token)) {
          dut_error("Attempting to complete a transaction while another transaction is already completed and has the same token.");
        };
        completed_transactions.add(temp_transaction);
        pending_transactions.delete(pending_transactions.key_index(token));
      };
      result = NO_ERROR;
    }else {
      result = UNMATCHED_RESPONSE;
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  get_completed_request(token : uint (bits:10)):csco_ht_packet is {
    if((score_level != csco_ht_full_scoring) && !completed_transactions.key_exists(token)) {
      result = NULL;
    } else {
      check that (completed_transactions.key_exists(token)) else
      dut_error("HT scoreboard is attempting to return a request for a transaction that does not exist on completed list.");
      result = completed_transactions.key(token).request;
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  get_completed_response(token : uint (bits:10)):csco_ht_packet is {
    check that (completed_transactions.key_exists(token)) else
    dut_error("HT scoreboard is attempting to return a response for a transaction that does not exist on completed list.");
    // This should only really be used for read responses and target dones, but not for nonposted broadcasts.
    result = completed_transactions.key(token).response[0];
  };
  ////////////////////////////////////////////////////////////////////////////////
  retire_transaction(token : uint (bits:10)) is {
    check that (completed_transactions.key_exists(token)) else
    dut_error("HT scoreboard is attempting to remove a transaction that does not exist on completed list.");
    completed_transactions.delete(completed_transactions.key_index(token));
  };
  ////////////////////////////////////////////////////////////////////////////////
  // check_response_data() is used for the random read/write tests created by the generate_random_traffic() method.
  // When a write is created the data to be written is placed on a local copy of memory that is then handed to this method.
  // Once a response comes back, the returned data is checked against the local copy of the data.  If the data matched
  // nothing is done, and the test proceeds as normal, but if there is a data mismatch, a csco_ht_transaction_error is created.
  ////////////////////////////////////////////////////////////////////////////////
  check_response_data(token : uint (bits:10)) is {
    var xaction : csco_ht_transaction;
    var request : csco_ht_packet;
    var response : csco_ht_packet;
    var addr : uint (bits:40);
    var data : csco_ht_packet;
    var read_count : uint (bits:4);

    check that (completed_transactions.key_exists(token)) else
    dut_error("Attempting to check the results of a read request/response pair that does not exist or is not completed.");

    xaction = completed_transactions.key(token);
    request = xaction.request;
    response = xaction.response[0];

    addr = request.Address();
    data = response.retrieve_data_packet();
    read_count = request.Mask_Count();


    if(xaction.register) {
      var reg : csco_ht_basic_reg = parent_device.read_reg_req.key(addr);
      reg.check_nonblocking_read_ht(token);
      parent_device.read_reg_req.delete(parent_device.read_reg_req.key_index(addr));
    };

    //if(!data_scoreboarding) {
      // The data has not been scoreboarded, so don't check anything.
    //  return;
    //};

    if(request.DataType() == Byte_Read) {
      // Byte Reads are always only 1 DoubleWord of Data.
      if(!shadow_memory.key_exists(addr[39:2])) {
        var error_msg : string;
        error_msg = appendf("SCOREBOARD Read Error occurred : Data Not Expected : Addr=0x%x", addr);
        sys.csco_ht_errors.report_transaction_error(UNEXPECTED_READ_DATA, error_msg, parent_device.control_block, request, response);
      }else {
        for i from 0 to 3 do {
          if(read_count[i:i] == 1) {
            if(data.as_a(Byte_Read'command csco_ht_packet).Data[((i+1)*8)-1:(i*8)] != shadow_memory.key(addr[39:2]).data[(i+1)*8-1:i*8]) {
              var error_msg : string;
              error_msg = appendf("SCOREBOARD Read Error occurred : Data Mismatch : Addr=0x%x Mask=0x%x Expected Data=0x%x Received Data=0x%x", addr + (i<<2), read_count, shadow_memory.key(addr[39:2]).data[(i+1)*8-1:i*8], data.as_a(Byte_Read'command csco_ht_packet).Data[((i+1)*8)-1:(i*8)]);
              sys.csco_ht_errors.report_transaction_error(UNMATCHED_READ_DATA, error_msg, parent_device.control_block, request, response);
              return;
            };
          };
        };
      };
    } else {
      for i from 0 to read_count do {
        if(!shadow_memory.key_exists(addr[39:2]+i)) {
          var error_msg : string;
          error_msg = appendf("SCOREBOARD Read Error occurred : Data Unexpected : Addr=0x%x Received Data=0x%x", addr + (i<<2), data.as_a(DoubleWord_Read'command csco_ht_packet).Data[((i+1)*32)-1:(i*32)]);
          sys.csco_ht_errors.report_transaction_error(UNEXPECTED_READ_DATA, error_msg, parent_device.control_block, request, response);
        }else {
          if(data.as_a(DoubleWord_Read'command csco_ht_packet).Data[((i+1)*32)-1:(i*32)] != shadow_memory.key(addr[39:2]+i).data) {
            var error_msg : string;
            error_msg = appendf("SCOREBOARD Read Error occurred : Data Mismatch : Addr=0x%x Expected Data=0x%x Received Data=0x%x", addr + (i<<2), shadow_memory.key(addr[39:2]).data, data.as_a(DoubleWord_Read'command csco_ht_packet).Data[((i+1)*32)-1:(i*32)]);
            sys.csco_ht_errors.report_transaction_error(UNMATCHED_READ_DATA, error_msg, parent_device.control_block, request, response);
            return;
          };
        };
      };
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  // Check to see if a transaction is still outstanding, or if it has been completed.
  // TRUE will be returned if the transaction still exists in the scoreboard.  Otherwise
  // FALSE will be returned.
  ////////////////////////////////////////////////////////////////////////////////
  transaction_pending(token : uint (bits:10)):bool is {
    if(pending_transactions.key_exists(token) == TRUE) {
      result = TRUE;
    }else {
      result = FALSE;
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  // Check to see if a transaction has been completed.
  // TRUE will be returned if it has.  Otherwise FALSE will be returned.
  ////////////////////////////////////////////////////////////////////////////////
  transaction_complete(token : uint (bits:10)):bool is {
    if(completed_transactions.key_exists(token) == TRUE) {
      result = TRUE;
    }else {
      result = FALSE;
    };
  };
  ////////////////////////////////////////////////////////////////////////////////////
  // post() is a method used to place a marker in either the response or broadcast
  // scoreboard so that the expected response can be tracked.  This is not for use
  // with the write scoreboard or any of the other random data generation scoreboards.
  ////////////////////////////////////////////////////////////////////////////////////
  post(packet: csco_ht_packet) is {
    case(packet.command()) {
      Write: {
        if(packet.vChan() == NonPosted) {
          add_request(packet);
        };
        if(score_level == csco_ht_full_scoring) {
          // Update the local target memory
          parent_device.generic_memory_write(shadow_memory, packet);
        };
      };
      Read: {
        add_request(packet);
      };
      Broadcast: {
        if(packet.vChan() == NonPosted) {
          add_request(packet);
        };
      };
      Flush: {
        add_request(packet);
      };
      default: {
        // This packet does not need to be scoreboarded.
      };
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  collect_config_response(token : uint (bits:10)):bool is {
    var request : csco_ht_packet;
    var response : csco_ht_packet;

    request = completed_transactions.key(token).request;
    response = completed_transactions.key(token).response[0];

    if(request.command() == Read || request.command() == Write) {
      // This response is part of a csco_ht_basic_reg transaction.  Don't check it.
      return (completed_transactions.key(token).register_config);
    };
    return FALSE;
  };
  ////////////////////////////////////////////////////////////////////////////////
  mark_config_transaction(token : uint(bits:10)) is {
    if(transaction_pending(token) == TRUE) {
      pending_transactions.key(token).register_config = TRUE;
    }else if(transaction_complete(token) == TRUE) {
      completed_transactions.key(token).register_config = TRUE;
    }else {
      dut_error("The scoreboard was trying to mark a config transaction that does not exist.");
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  mark_reg_transaction(token : uint(bits:10)) is {
    if(transaction_pending(token) == TRUE) {
      pending_transactions.key(token).register = TRUE;
    }else if(transaction_complete(token) == TRUE) {
      completed_transactions.key(token).register = TRUE;
    }else {
      dut_error("The scoreboard was trying to mark a config transaction that does not exist.");
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  // Return the number of outstanding transactions.
  ////////////////////////////////////////////////////////////////////////////////
  check_scoreboard() is {
    sys.csco_ht_test_control.csco_ht_model_outstanding_transactions += pending_transactions.size();
    sys.csco_ht_test_control.csco_ht_model_outstanding_transactions += completed_transactions.size();
  };
  ////////////////////////////////////////////////////////////////////////////////
};
////////////////////////////////////////////////////////////////////////////////
// When a Non Posted request is generated the transmitter needs to track the expected response.  This structure
// is used to keep track of those request/response pairs and would be in a keyed list that is indexed by the
// aggregation of the UnitID and SrcTag.
////////////////////////////////////////////////////////////////////////////////
struct csco_ht_transaction {
  !token : uint (bits:10);

  request : csco_ht_packet;
  !response : list of csco_ht_packet;
  response_count : uint;
  keep soft response_count == 1;
  register_config : bool; // This is a flag to indicate that the source of this transaction was to config a csco_ht_basic_reg object.
  keep soft register_config == FALSE;
  register : bool; // This is a flag to indicate that the source of this transaction was the csco_ht_basic_reg object.
  keep soft register == FALSE;

  post_generate() is also {
    token = request.generate_token();
  };
};
////////////////////////////////////////////////////////////////////////////////
// Error - These objects and methods provide a common way to report errors.
////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
// csco_ht_error_handler provides methods for the device to report different classes of errors.
// Additionally, it will signal the test control flags when a FATAL error has occured.
////////////////////////////////////////////////////////////////////////////////
unit csco_ht_error_handler {
  short_name(): string is { result = "HT_ERROR_HANDLER" };
  logger : message_logger;
    keep soft logger.verbosity == HIGH;
    keep soft logger.tags == {NORMAL};

  !errors : list of csco_ht_basic_error;
  !warnings : list of csco_ht_basic_error;
  !ignores : list of csco_ht_basic_error;
  // Flags to allow the user to determine the severity of the errors.
  initialization_error_severity : csco_ht_error_severity;
  keep soft initialization_error_severity == Error_Stop;
  protocol_error_severity : csco_ht_error_severity;
  keep soft protocol_error_severity == Error_Stop;
  transaction_error_severity : csco_ht_error_severity;
  keep soft transaction_error_severity == Error_Stop;
  ////////////////////////////////////////////////////////////////////////////////
  // reset() print out the summary of errors and reset all of the counters
  ////////////////////////////////////////////////////////////////////////////////
  reset() is {
    var error_flag : bool;

    if(!errors.is_empty() || !warnings.is_empty() || !ignores.is_empty()) {
      // If all of the list are empty, do nothing, otherwise dump the contents and reset them.
      error_flag = report_summary();
      errors = {};
      warnings = {};
      ignores = {};
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  // report_init_error() is a method used to record HT protocol errors during initialization.
  ////////////////////////////////////////////////////////////////////////////////
  report_init_error(error_code : csco_ht_error_code, message : string, control_block : csco_ht_device_control_block, link_id : uint) is {
    var temp_error : csco_ht_basic_error;
    gen temp_error keeping {
      .kind == Initialization;
      .error_code == error_code;
      .device_reference_name == control_block.device_reference_name;
      .device_function == control_block.device_function;
      .sim_time == sys.time;
      it is a Initialization'kind csco_ht_basic_error (init_err) =>
        (init_err.link_number == link_id) and
        (init_err.link_orientation == control_block.read_link_orientation(link_id));
      .error_string == message;
    };
    classify_errors(initialization_error_severity, temp_error);
  };
  ////////////////////////////////////////////////////////////////////////////////
  // report_protocol_error() is a method used to record HT protocol errors.
  ////////////////////////////////////////////////////////////////////////////////
  report_protocol_error(error_code : csco_ht_error_code, message : string, control_block : csco_ht_device_control_block, link_id : uint) is {
    var temp_error : csco_ht_basic_error;
    gen temp_error keeping {
      .kind == Protocol;
      .error_code == error_code;
      .device_reference_name == control_block.device_reference_name;
      .device_function == control_block.device_function;
      .sim_time == sys.time;
      .as_a(Protocol'kind csco_ht_basic_error).link_number == link_id;
      .as_a(Protocol'kind csco_ht_basic_error).link_orientation == control_block.read_link_orientation(link_id);
      .error_string == message;
    };
    classify_errors(protocol_error_severity, temp_error);
  };
  ////////////////////////////////////////////////////////////////////////////////
  // report_transaction_error() is a method for reporting an error in read/write data.  This is
  // primarily used for the standard random generation method which employs self checking of the data.
  ////////////////////////////////////////////////////////////////////////////////
  report_transaction_error(error_code : csco_ht_error_code, message : string, control_block : csco_ht_device_control_block, request : csco_ht_packet, response : csco_ht_packet) is {
    var temp_error : csco_ht_basic_error;
    gen temp_error keeping {
      .kind == Transaction;
      .error_code == error_code;
      .device_reference_name == control_block.device_reference_name;
      .device_function == control_block.device_function;
      .sim_time == sys.time;
      .error_string == message;
      .as_a(Transaction'kind csco_ht_basic_error).request == request;
      .as_a(Transaction'kind csco_ht_basic_error).response == response;
    };
    classify_errors(transaction_error_severity, temp_error);
  };
  ////////////////////////////////////////////////////////////////////////////////
  classify_errors(severity : csco_ht_error_severity, temp_error : csco_ht_basic_error) is {
    case(severity) {
      Error_Stop: {
        temp_error.print_error();
        errors.add(temp_error);
        sys.csco_ht_test_control.csco_ht_model_error = TRUE;
      };
      Error: {
        temp_error.print_error();
        errors.add(temp_error);
      };
      Warning: {
        temp_error.print_error();
        warnings.add(temp_error);
      };
      Ignore: {
        ignores.add(temp_error);
      };
      default: {
        dut_error("Unable to classify error code.");
      };
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  check_csco_ht_error_handler(error_class : csco_ht_error_classification):list of csco_ht_basic_error is {
    result = errors.all(.kind == error_class);
  };
  ////////////////////////////////////////////////////////////////////////////////
  check_csco_ht_warning_handler(error_class : csco_ht_error_classification):list of csco_ht_basic_error is {
    result = warnings.all(.kind == error_class);
  };
  ////////////////////////////////////////////////////////////////////////////////
  report_summary(): bool is {
    result = TRUE;
    // Dump out info about errors that occured.
    var protocol_errors : list of csco_ht_basic_error = check_csco_ht_error_handler(Protocol);
    var transaction_errors : list of csco_ht_basic_error = check_csco_ht_error_handler(Transaction);

    if (sys.csco_ht_test_control.csco_ht_model_outstanding_transactions != 0) {
      dut_error("HT Model has outstanding transactions.");
    };

    if(protocol_errors.is_empty() == FALSE) {
      messagef(LOW, "ERROR: Protocol Errors were found!\n") {
        for each in protocol_errors {
          it.print_error();
          outf("\n");
        };
      };
      result = FALSE;
    };
    if(transaction_errors.is_empty() == FALSE) {
      messagef(LOW, "ERROR: Transaction Errors were found!\n") {
        for each in transaction_errors {
          it.print_full_error();
          outf("\n");
        };
      };
      result = FALSE;
    };
    // Dump out info about warnings that occured.
    var protocol_warnings : list of csco_ht_basic_error = check_csco_ht_warning_handler(Protocol);
    var transaction_warnings : list of csco_ht_basic_error = check_csco_ht_warning_handler(Transaction);
    if(protocol_warnings.is_empty() == FALSE) {
      messagef(LOW, "WARNING: Protocol Warnings were found!\n") {
        for each in protocol_warnings {
          it.print_error();
          outf("\n");
        };
      };
      result = FALSE;
    };
    if(transaction_warnings.is_empty() == FALSE) {
      messagef(LOW, "WARNING: Transaction Warnings were found!\n") {
        for each in transaction_warnings {
          it.print_full_error();
          outf("\n");
        };
      };
      result = FALSE;
    };
    messagef(LOW, "********************************************************************************\n");
    messagef(LOW, "HT Report : %d Protocol Errors Occurred : %d Transaction Errors Occurred\n", protocol_errors.size(), transaction_errors.size());
    messagef(LOW, "HT Report : %d Protocol Warnings Occurred : %d Transaction Warnings Occurred.\n", protocol_warnings.size(), transaction_warnings.size());
    messagef(LOW, "********************************************************************************\n");
  };
  ////////////////////////////////////////////////////////////////////////////////
};
////////////////////////////////////////////////////////////////////////////////
// csco_ht_basic_errors define an object that contains all of the relevant information
// for different categories of errors.
////////////////////////////////////////////////////////////////////////////////
struct csco_ht_basic_error {
  kind : csco_ht_error_classification;
  error_code : csco_ht_error_code;
  device_reference_name : string;
  device_function : csco_ht_device_encoding;
  sim_time : uint;
  error_string : string;
  print_error() is empty;
  print_full_error() is empty;
  print_error_header() is {
    var device_func : string;
    case(device_function) {
      Host_Bridge: {
        device_func = "Host Dev";
      };
      Tunnel_IO_Device: {
        device_func = "Tunnel Dev";
      };
      EOC_IO_Device: {
        device_func = "EOC Dev";
      };
    };
    outf("%12d ns %12s(%10s) ERROR: %s : ", sim_time, device_reference_name, device_func, error_code);
  };
  ////////////////////////////////////////////////////////////////////////////////
  when Initialization'kind csco_ht_basic_error {
    link_number : uint;
    link_orientation : csco_ht_link_encoding;

    print_error() is {
      var link_dir : string;
      if(link_orientation == TX_Upstream_RX_Downstream_Interface) {
        link_dir = "Upstream";
      }else {
        link_dir = "Downstream";
      };
      print_error_header();
      outf("%s Link %s\n", link_dir, error_string);
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  when Protocol'kind csco_ht_basic_error {
    link_number : uint;
    link_orientation : csco_ht_link_encoding;

    print_error() is {
      var link_dir : string;
      if(link_orientation == TX_Upstream_RX_Downstream_Interface) {
        link_dir = "Upstream";
      }else {
        link_dir = "Downstream";
      };
      print_error_header();
      outf("%s Link %s\n", link_dir, error_string);
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  when Transaction'kind csco_ht_basic_error {
    request : csco_ht_packet;
    response : csco_ht_packet;

    print_error() is {
      print_error_header();
      outf("%s\n", error_string);
    };

    print_brief_error() is {
      print_error();
      print_error_header();
      if(request != NULL) {
        outf("Request Packet: %s : ", request.print_csco_ht_packet_string());
      };
      if(response != NULL) {
        outf("Response Packet: %s", response.print_csco_ht_packet_string());
      };
      outf("\n\n");
    };

    print_full_error() is {
      var tmp_data: uint;

      print_error();
      if(request != NULL) {
        outf("%40s  Request Packet: Cmd=%s Pkt_type=%s vChan=%s Len=%d Weight=%x Sched_id=%x\n",
             "", request.command(), request.packet_type, request.vChan(), request.packet_length,
             request.weight, request.scheduling_index);
        outf("%56s  %s\n", " ", request.print_csco_ht_packet_string());
        if(request.command() == Write) { // There is associated data, print that out.
          outf("%56s  %s\n", "", (request.retrieve_data_packet()).print_csco_ht_packet_string());
        };
      };
      if(response != NULL) {
        outf("%40s Response Packet: Cmd=%s Pkt_type=%s vChan=%s Len=%d\n",
             "", response.command(), response.packet_type, response.vChan(), response.packet_length);
        outf("%56s  %s\n", " ", response.print_csco_ht_packet_string());
        if(response.command() == RdResponse) { // There is associated data, print that out.
          outf("%56s  %s\n", "", (response.retrieve_data_packet()).print_csco_ht_packet_string());
        };
      };
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
};
////////////////////////////////////////////////////////////////////////////////
// csco_ht_test_control block implements the system wide HT timers for traffic generation and system activity.
////////////////////////////////////////////////////////////////////////////////
unit csco_ht_test_control_block {
  short_name(): string is { result = "HT_TEST_CONTROL" };
  logger : message_logger;
    keep soft logger.verbosity == HIGH;
    keep soft logger.tags == {NORMAL};

  // These are the two flags that the external test bench needs to monitor.
  csco_ht_model_done : bool;  // Flag used to indicate that the HT model is done
  keep soft csco_ht_model_done == FALSE;
  csco_ht_model_inactive : bool;  // Flag used to indicate that the HT model is hung.
  keep soft csco_ht_model_inactive == FALSE;
  csco_ht_model_error : bool;  // Flag used to indicate that the HT model detected an error that the user marked as fatal.
  keep soft csco_ht_model_error == FALSE;
  csco_ht_model_initialized : bool;
  keep soft csco_ht_model_initialized == FALSE;

  // An event to indicate that the HT system has been initialized.
  event csco_ht_model_initialized;
  on csco_ht_model_initialized {
    csco_ht_model_initialized = TRUE;
    messagef(LOW, "HT models are initialized\n");
  };
  // An event to indicate that the HT system can start generating traffic.
  event csco_ht_start_traffic;
  on csco_ht_start_traffic {
    messagef(LOW, "HT models are starting traffic generators\n");
  };

  // A Counter to track how many transaction are outstanding when the model completes.
  csco_ht_model_outstanding_transactions : uint;
  keep csco_ht_model_outstanding_transactions == 0;

  csco_ht_watchdog_started : bool;
  keep soft csco_ht_watchdog_started == FALSE;

  csco_ht_bus_timeout : uint; // The reset value for the activity timer.
  keep csco_ht_bus_timeout > 256; // The largest possible HT packet is a 64 byte data packet.  Which will take 64 cycles on an 8 bit link
                              // and 256 cycles on a 2 bit link.  Thus the minimum timeout has to be large enough to send this packet.
  keep soft csco_ht_bus_timeout == 300;
  csco_ht_activity_timer : uint;  // Counter used to count down since the last activity was detected.
  keep csco_ht_activity_timer == csco_ht_bus_timeout;  // Use a seperate variable to set the timeout value.
  // On the ht bus, there will always be idle packets sent even if there is no other traffic.
  // Therefore, activity will be defined as any non-idle ht packet.
  // An idle packet is a nop with 0 buffer releases.  The idle counter must be at least 4 cycles
  // long, since on an 8 bit link, a idle packet will take 4 cycles to transmit.

  // Two internal flags used to track the state of the model.
  csco_ht_activity_timer_expired : bool;
  keep soft csco_ht_activity_timer_expired == FALSE;
  csco_ht_traffic_generation_done : bool;  // Flag that indicates that all devices have complete generating new traffic.
  keep soft csco_ht_traffic_generation_done == FALSE;

  // Track the number of 'e' devices in the HT testbench.
  num_csco_ht_e_devices : uint;
  keep num_csco_ht_e_devices == 0;
  // This is a list that tracks the state of each 'e' device in the csco_ht_testbench.
  // For each entry in the list, the index indicates the testbench ID of the device and TRUE
  // indicates that the device is done generating traffic.  A FALSE indicates that the device is
  // still generating traffic.
  !csco_ht_traffic_generation_monitor : list of bool;

  // Derive this from the csco_ht_device_clock's so that it is the logical OR of all of them.  This will allow it to
  // sample at a rate independent of each device.
  event csco_ht_watchdog_clock;
  ////////////////////////////////////////////////////////////////////////////////
  // Watchdog TCM and monitoring for errors.
  monitor_csco_ht_model()@csco_ht_watchdog_clock is {
    // Wait until some or all of the devices have been initialized and the watchdogs have started.
    wait @sys.csco_ht_test_control.csco_ht_start_traffic;
    // Start two watchdogs to monitor the state of the HT testbench.  The first monitor checks for
    // activity and if a set time elapses set a flag to indicate that the model has hung.
    // The second monitor checks to see if the devices are done generating traffic.  Set a flag
    // that the individual devices will reset if they are still generating traffic.  Once they are done,
    // wait for the activity timer to run out and then set a flag to indicate that the testbench is complete.
    first of {
      {
        // Monitor and decrement the activity timer to check to see if the model is hung.
        monitor_csco_ht_activity();
      };
      {
        // Monitor the activity of the device generation.
        wait true(csco_ht_traffic_generation_done == TRUE);
        // As long as the model is not done generation traffic, keep the watchdog alive.
        // Once the traffic generation is done exit both the traffic monitor and the activity monitor
      };
    };

    if(csco_ht_traffic_generation_done == TRUE) {
      // The watchdogs are complete because the devices are done with traffic generation.
      // Wait until the activity timer expires and then set the model done flag.
      monitor_csco_ht_activity();
      // There is no more traffic on the bus, so set the csco_ht_model_done flag to indicate that the ht part of the test is done.
      csco_ht_model_done = TRUE;
    }else {
      // The watchdogs are complete, but the devices are not done with traffic generation.  (e.g. the model must have hung.)
      csco_ht_model_inactive = TRUE;
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  monitor_csco_ht_activity()@csco_ht_watchdog_clock is {
    while(csco_ht_activity_timer_expired == FALSE) {
      if(csco_ht_activity_timer == 0) {
        // The HT model has hung, set a flag to indicate this.
        csco_ht_activity_timer_expired = TRUE;
      }else {
        // Decrement the timer to monitor for model stalls or hangs.
        csco_ht_activity_timer -= 1;
      };
      wait cycle;
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  reset_csco_ht_activity_timer() is {
    csco_ht_activity_timer = csco_ht_bus_timeout;
  };
  ////////////////////////////////////////////////////////////////////////////////
  // When a device is instantiated in needs to register with the test control so that
  // the traffic generation can be tracked.
  ////////////////////////////////////////////////////////////////////////////////
  register_csco_ht_device(): uint is {
    result = num_csco_ht_e_devices;
    num_csco_ht_e_devices += 1;
    csco_ht_traffic_generation_monitor.add(FALSE);
  };
  ////////////////////////////////////////////////////////////////////////////////
  // Allow a device to indicate that it is done with traffic generation.
  ////////////////////////////////////////////////////////////////////////////////
  device_traffic_generation_completed(ID: uint) is {
    csco_ht_traffic_generation_monitor[ID] = TRUE;
    check_traffic_generation_progress();
  };
  ////////////////////////////////////////////////////////////////////////////////
  check_traffic_generation_progress() is {
    check that (csco_ht_traffic_generation_monitor.size() == num_csco_ht_e_devices) else
    dut_error("The size of list csco_ht_traffic_generation_monitor does not equal num_csco_ht_e_devices.");

    var csco_ht_traffic_done : bool = TRUE;
    for each (monitor) in csco_ht_traffic_generation_monitor do {
      if(monitor == FALSE) {
        csco_ht_traffic_done = FALSE;
      };
    };
    if(csco_ht_traffic_done) {
      csco_ht_traffic_generation_done = TRUE;
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  run() is also {
    start monitor_csco_ht_model();
  };
  ////////////////////////////////////////////////////////////////////////////////
};
////////////////////////////////////////////////////////////////////////////////
// Misc - These are small objects that are used throughout the model.
////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
// A management structure used to hold a packet and record a system time.
////////////////////////////////////////////////////////////////////////////////
struct csco_ht_packet_record {
  packet : csco_ht_packet;
  time : uint;
  keep me.time == sys.time;

  print_record() is {
    outf("%12d ns", time);
    packet.print_csco_ht_packet();
    outf("\n");
  };
};
////////////////////////////////////////////////////////////////////////////////
// This is a structure that is used to model a generic memory for HT devices.  It would be an entry
// in a keyed list, uses a 38 bit address because each entry is 32 bits.
////////////////////////////////////////////////////////////////////////////////
struct csco_ht_memory_element {
  // Use addr[39:2] to index into the memory hierarchy
  addr : uint (bits:38);
  full_addr : uint (bits:40);
  keep full_addr == addr << 2;
  data : uint;

  print_mem() is {
    outf("Addr=0x%x Data=0x%x\n", full_addr, data);
  };
};
////////////////////////////////////////////////////////////////////////////////
'>
