////////////////////////////////////////////////////////////////////////////
// Title         : HT Declarations
// Project       : HT Testbench
////////////////////////////////////////////////////////////////////////////
// File          : csco_ht_declarations.e
// Author        : Brian Van Essen  <vanessen@cisco.com>
// Created       : 25.01.2001
// Last modified : 02.05.2001
////////////////////////////////////////////////////////////////////////////
// Description :
// The primary file for the HT testbench.  It instantiates all of the commonly
// used enumerated types, and provides the initial definition of each unit/struct
// used in the testbench.  The reason for this is that most objects have
// backpointers for their parent objects.  Thus each object has to be defined
// before the use of the backpointers.
////////////////////////////////////////////////////////////////////////////////
// Modification history :
// 25.01.2001 : created
////////////////////////////////////////////////////////////////////////////

// ***** BEGIN LICENSE BLOCK *****
// Version: MPL 1.1
// The contents of this file are subject to the Mozilla Public License Version
// 1.1 (the "License"); you may not use this file except in compliance with
// the License. You may obtain a copy of the License at
// http:                                                   // www.mozilla.org/MPL/ or ./mpl_1_1.txt
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
// for the specific language governing rights and limitations under the License.
// The Original Code is Cisco Hypertransport eVC.
// The Initial Developer of the Original Code is
// Brian Van Essen, Lisa Huang, and Sean Smith.
// Portions created by the Initial Developer are Copyright (C) 2002
// the Initial Developer. All Rights Reserved.
// Contributor(s): Brian Van Essen, Lisa Huang, and Sean Smith
// ***** END LICENSE BLOCK *****


<'
// Define the orientation of a HT link
type csco_ht_link_encoding : [TX_Upstream_RX_Downstream_Interface, TX_Downstream_RX_Upstream_Interface];
// Possible packet types
type csco_ht_packet_encoding : [Nop, Flush, Write, Read, RdResponse, TgtDone, Broadcast, Fence, Sync, Reserved, DoubleWord_Read, DoubleWord_Write, Byte_Read, Byte_Write];
type csco_ht_packet_type : [Info, Ctl, Data];
// Interrupt types
type csco_ht_interrupt_type : [Fixed=3'b000, Lowest_Priority_Mode=3'b001, SMI=3'b010, NMI=3'b011, INIT=3'b100, Startup=3'b101, Ext_Int=3'b110, EOI=3'b111];
type csco_ht_interrupt_mode : [Edge=1'b0, Level=1'b1];
type csco_ht_interrupt_destination : [Physical=1'b0, Logical=1'b1];
// Attributes for HT packets
type csco_ht_packet_attributes : [NonPosted, Posted, Resp_NoPassPostedReq, Resp_PassPostedReq, Byte, DoubleWord, Normal, Isochronous, Noncoherent, Coherent, NoPassPW, PassPW];
type csco_ht_passing_rules : [NA, NoPassPW, PassPW];
type csco_ht_data_type : [Byte, DoubleWord];
type csco_ht_BW_latency_req : [Normal, Isochronous];
type csco_ht_mem_access_type : [NonCoherent, Coherent];
type csco_ht_virtual_channel : [NA, NonPosted, Posted, Response];
// Possible HT device types
type csco_ht_device_encoding : [Host_Bridge, Tunnel_IO_Device, EOC_IO_Device];
// Possible HT device types that the Host Bridge directly connects to - HB for SES, Tunnel or NA
// this comes in handy when instantiating more than 1 HT model...
type csco_ht_hostbridge_type : [NA, HB_SES, HB_TUNNEL];

// Legal link widths (Note: 2 & 4 bit links are not used.)
// Assign the enumerated types as unique id equal to their bit width.  This allows for constrained generation.
type csco_ht_link_size : [Eight_Bits = 8, Sixteen_Bits = 16, ThirtyTwo_Bits = 32];
// HT capability register formats
//type csco_ht_capability_type : [Slave_Primary, Host_Secondary];
type csco_ht_capability_type : [Primary, Secondary];
type csco_ht_reset_type : [Cold, Warm];
// Legal clock speeds for HT links
// These are the multiplication factors needed to achieve a clock of a given frequency when using a reference clock of 1/12 ns.
type csco_ht_clock_multiplier : [csco_ht_200MHz=60, csco_ht_300MHz=40, csco_ht_400MHz=30, csco_ht_500MHz=24, csco_ht_600MHz=20, csco_ht_800MHz=15, csco_ht_1000MHz=12, csco_ht_DirectDrive=2];
// Memory ranges in memory map
type csco_ht_memory_range : [Memory_Mapped_IO, Interrupts, PCI_IO, PCI_Configuration, Reserved];
type csco_ht_traffic_type : [tx, rx, forwarding];
// States for Host Bridge Initialization state machine
type csco_ht_init_stages : [Generic_0, Generic_1, HT_1, HT_Slave_Primary_1,  HT_Host_1, HT_EOC_1, HT_Tunnel_1, Non_HT_1, Chain_Init_Complete, Chain_Init_Error];

// Physical Link Initialization Parameters
define CSCO_HT_MIN_SYNC_SEQUENCE_CYCLES 16;
define CSCO_HT_NUM_TX_RX_POINTER_INIT_CYCLES 512;
define CSCO_HT_NUM_PACKET_FRAMING_CYCLES 4;

type csco_ht_reg_size : [Eight = 8, Sixteen = 16, TwentyFour = 24, ThirtyTwo = 32, SixtyFour = 64];
// Standard CSR values
type csco_ht_csr_access_type : [Type0 = 16'hFDFE, Type1 = 16'hFDFF];       // : uint (bits:16);
define CSCO_HT_CAPABILITIES_OFFSET 8'h40;
// These are the masks for unused BAR's
define CSCO_HT_BASE_ADDRESS_REGISTER_MASK_UNUSED 32'h0000_0000;    // This is an unused BAR
// These are the masks for 32 bit BAR's
define CSCO_HT_BASE_ADDRESS_REGISTER_MASK_32 32'hF000_0000;// This is a 256 MB Addressable region
// These are the masks for 64 bit BAR's
define CSCO_HT_BASE_ADDRESS_REGISTER_MASK_LOW 32'hF000_0000;       // This is a 256 MB Addressable region
define CSCO_HT_BASE_ADDRESS_REGISTER_MASK_HIGH 32'hFFFF_FFFF;      // This is the continuation of a 256 MB Addressable region
// These are the masks for 40 bit BAR's
define CSCO_HT_BASE_ADDRESS_REGISTER_MASK_HT 32'hFF00_0000; // This is a 256 MB Addressable region in the 40 shifted reserved BAR's

// HT Address Map
type csco_ht_address_region : [Generic, Memory_Mapped_IO, Interrupts, IACK, System_Management, PCI_IO, PCI_Configuration];

type csco_ht_endian_type : [little_endian, big_endian, big_endian_bit_match];

type csco_ht_device_reg_type : [VendorID, DeviceID, Command, Status, RevisionID, ClassCode, Cache_Line_Size,
  Latency_Timer, Header_Type, BIST,  BAR0, BAR1, BAR2, BAR3, BAR4, BAR5, Cardbus_CIS_Pointer,
  SubSystem_Vendor_ID, SubSystem_ID, Expansion_ROM_Base_Address, Capabilities_Pointer, Reserved_34h,
  Reserved_38h, Interrupt_Line, Interrupt_Pin, Min_Gnt, Max_Lat];

type csco_ht_capability_reg_type : [Capability_ID, Capabilities_Pointer, Command, Link_Control0, Link_Config0,
  Link_Control1, Link_Config1, HT_Rev_ID, Link_Freq, Reserved];

type csco_ht_scenario_kind : [];

// different scoring thresholds to turn on and off parts of the HT scoreboard
// this also allows for an external scoreboard to be used
//    no_scoring       : turns off all scoreboarding
//    register_scoring : only score register transactions
//    full_scoring     : score all transactions
type csco_ht_scoreboard_level : [csco_ht_no_scoring, csco_ht_register_scoring, csco_ht_full_scoring];

type csco_ht_error_classification : [Initialization, Protocol, Transaction];
type csco_ht_error_severity : [Ignore, Warning, Error, Error_Stop];

type csco_ht_error_code : [NO_ERROR             = 32'h0000_0000,
// Transaction Error Codes
                       WRITE_NXA            = 32'h0000_0001,       // Non Existent Address
                       WRITE_ERROR          = 32'h0000_0002,
                       READ_NXA             = 32'h0000_0004,       // Non Existent Address
                       READ_ERROR           = 32'h0000_0008,
                       FLUSH_NXA            = 32'h0000_0010,       // Never Expect to see this
                       FLUSH_ERROR          = 32'h0000_0020,
                       BROADCAST_NXA        = 32'h0000_0040,       // Never Expect to see this
                       BROADCAST_ERROR      = 32'h0000_0080,
                       UNMATCHED_RESPONSE   = 32'h0000_0100,       // There is no corresponding Request for this response packet.
                       UNMATCHED_REQUEST    = 32'h0000_0200,       // No response was received for this request.
                       UNMATCHED_READ_DATA  = 32'h0000_0400,       // Scorebaord received read data that did not match data in scoreboard memory.
                       UNEXPECTED_READ_DATA = 32'h0000_0800,       // Scoreboard received read data that did not exist in scoreboard memory.
// Protocol Error Codes
                       ILLEGAL_INTERRUPT    = 32'h0000_1000,
                       ILLEGAL_EOI          = 32'h0000_2000,
                       ILLEGAL_DATA_PACKET  = 32'h0000_4000,       // A data packet is being received when non is expected.
                       ILLEGAL_CTL_CHANGE   = 32'h0001_0000,       // The CTL line changed on an illegal 4-byte boundary.
                       RSV_COMMAND_FIELD    = 32'h0002_0000,       // HT packet had a reserved value in the command field.
                       ILLEGAL_PACKET_SIZE  = 32'h0004_0000,       // Packet has an illegal size for its type.
                       CRC_ERROR            = 32'h0008_0000,       // Bad CRC
// Initialization Errors (Protocol Errors)
                       INIT_ILLEGAL_CTL_CHANGE = 32'h0010_0000,// Indicates that the CTL line was not asserted across a rising clock edge when indicating a initialization sequence.
                       INIT_MIN_SYNC_SEQ       = 32'h0020_0000,    // Indicates that the minimum number of bit times in the sync sequence was violated.
                       INIT_MIN_LOAD_PTR_SEQ   = 32'h0040_0000,// Indicates that the minimum number of bit times in the load pointer initialization sequence was violated.
                       INIT_MAX_LOAD_PTR_SEQ   = 32'h0080_0000,// Indicates that the maximum number of bit times in the load pointer initialization sequence was violated.
                       INIT_MIN_FRAMING_SEQ    = 32'h0100_0000,    // Indicates that the minimun number of bit times in the framing sequence was violated.
                       INIT_MAX_FRAMING_SEQ    = 32'h0200_0000,    // Indicates that the minimun number of bit times in the framing sequence was violated.
                       INITIALIZATION_ERROR    = 32'h0400_0000,    // Indicate a general error in the initialization sequence.
// Chain Configuration Errors
                       CHAIN_CONFIG_ERROR      = 32'h1000_0000,    // Indicates a general error in the chain configuration sequence.
                       ILLEGAL_MEM_ACCESS      = 32'h2000_0000,    // Indicates that a read or write did not stay within the appropriate 32/64 byte boundary.
                       EVC_HT_CHAIN_RESET     = 32'h4000_0000,     // Indicates that the ht chain was reset by the eVC.
                       UNKNOWN_ERROR        = 32'h8000_0000];      // Indicate that an error occured of unknown type.

// Reset Sequence Constants
define CSCO_HT_RESET_ASSERTION_LENGTH 100;                 // This should actually be 1 ms.
////////////////////////////////////////////////////////////////////////////////
'>
