////////////////////////////////////////////////////////////////////////////
// Title         : HT Device
// Project       : HT Testbench
////////////////////////////////////////////////////////////////////////////
// File          : csco_ht_device.e
// Author        : Brian Van Essen  <vanessen@cisco.com>
// Created       : 25.01.2001
// Last modified : 04.05.2001
////////////////////////////////////////////////////////////////////////////
// Description :
// The Device is the brains of a node sitting on the HT chain.  It is responsible for generating
// and managing traffic at a high level.  It will then pass the desired traffic to the HT Channel
// for interaction with the actual HT protocol/bus.
////////////////////////////////////////////////////////////////////////////
// Modification history :
// 25.01.2001 : created
////////////////////////////////////////////////////////////////////////////

// ***** BEGIN LICENSE BLOCK *****
// Version: MPL 1.1
// The contents of this file are subject to the Mozilla Public License Version
// 1.1 (the "License"); you may not use this file except in compliance with
// the License. You may obtain a copy of the License at
// http:                                                   // www.mozilla.org/MPL/ or ./mpl_1_1.txt
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
// for the specific language governing rights and limitations under the License.
// The Original Code is Cisco Hypertransport eVC.
// The Initial Developer of the Original Code is
// Brian Van Essen, Lisa Huang, and Sean Smith.
// Portions created by the Initial Developer are Copyright (C) 2002
// the Initial Developer. All Rights Reserved.
// Contributor(s): Brian Van Essen, Lisa Huang, and Sean Smith
// ***** END LICENSE BLOCK *****

<'
////////////////////////////////////////////////////////////////////////////////
// The Device is the brains of a node sitting on the HT chain.  It is responsible for generating
// and managing traffic at a high level.  It will then pass the desired traffic to the HT Channel
// for interaction with the actual HT protocol/bus.
////////////////////////////////////////////////////////////////////////////////
unit csco_ht_device {

  // Determine what manner of device this is, a Host, Tunnel or EOC
  device_function : csco_ht_device_encoding;
  // A multiplier that will determine what frequency the Device clock should run at.  It is setup to run at the same rate
  // as the fastest clock for any of the devices link's.
  device_clock_freq : uint;
  keep device_clock_freq == control_block.read_max_current_tx_clock_freq();
  // A single clock for the device.  It runs at a rate equal to the fastest link clock.
  event csco_ht_device_clock is {@csco_ht_device_clock; [device_clock_freq]} or {@session.start_of_test} @csco_ht_device_fundamental_clock;
  // One every instance of the device clock, stimulate the watchdog clock so that the system can check to see if the
  // devices are still alive and generating traffic.
  on csco_ht_device_clock {
    emit sys.csco_ht_test_control.csco_ht_watchdog_clock;
  };

  // A uint that tracks the devices unique ID number in the system.
  testbench_ID : uint;
  // added 09/05/02 for support remote boot
  remote_boot: bool;
  keep soft remote_boot == FALSE;
  // added 09/05/02 for gen crc error
  genCrcError: bool;
  keep soft genCrcError == FALSE;
  doneGenCrcError: bool;
  keep soft doneGenCrcError == TRUE;
  // added 09/05/02 for gen_direct (not to restrict byte mask when using directed test)
  gen_direct: bool;
  keep soft gen_direct == FALSE;


  // A reference to the CSR and general control block.
  control_block : csco_ht_device_control_block;
  keep me.control_block.num_device_csco_ht_interfaces in [1..2];
  keep me.control_block.device_function == device_function;
  keep sys.csco_ht_development_test == TRUE => me.control_block.device_reference_name == (str_split(me.e_path(), "'")).top();
  keep sys.csco_ht_development_test == FALSE => me.control_block.device_reference_name == (str_split(me.e_path(), ".")).top();
  // A list of the HT links/interfaces associated with the device.
  csco_ht_interfaces : list of csco_ht_channel_interface is instance;
  keep me.csco_ht_interfaces.size() in [1..2];
  keep for each in csco_ht_interfaces {
    it.channel_identifier == index;
  };
  // A flag used to indicate if the transmit and receive traffic should be recorded.
  record_device_history : bool;
  keep soft record_device_history == FALSE;
  // A flag used to indicate that device level information should be printed.
  display_records : bool;
  keep soft display_records == FALSE;
  // Lists of all of the transactions handled by this device.  These are only filled if the record_device_history flag is set.
  !tx_history : list of csco_ht_packet_record; // Generated Packets
  !rx_history : list of csco_ht_packet_record; // Accepted Packets
  !forwarding_history : list of csco_ht_packet_record; // Packets that are forwarded/tunneled don't appear on either the tx or rx boards.

  // list of register read requests
  !read_reg_req : list (key: Addr) of csco_ht_basic_reg;

  // list of csco_ht_address_maps to define the host bridges memory map and
  // endianness for each region
  !device_mem_map : list (key: base_addr) of csco_ht_address_map;

  // device memory map mask, used to mask off the address bits to decode base address
  // of a region
  !device_mem_map_mask: uint(bits: 40);

  // Support for Sync floods/resets
  event Sync_Flood_Init;
  on Sync_Flood_Init {
    // If a HT Link on a device receives sync packets, it needs to terminate all current
    // transactions and propogate the sync flood to all interfaces.
    for each (link) in csco_ht_interfaces {
      link.initiate_sync_flood();
    };
  };
  event Sync_Flood_Complete;
  on Sync_Flood_Complete {
    start reset_sequence();
  };

  ////////////////////////////////////////////////////////////////////////////-
  // method to repeat a patten of data
  ////////////////////////////////////////////////////////////////////////////-
  repeat(numRepeat:int, patten:int(bits:8)):list of bit is {
    var listn: list of bit;
    var n:uint(bits:512) = 512'h0;

    for i from 1 to numRepeat/8 do {
      n=%{n,patten};
    };
    listn = %{n[numRepeat-1:0]};
    return listn;

  };
  ////////////////////////////////////////////////////////////////////////////-
  // method to check what region a given address falls into the host bridge's
  // memory map. it returns the endian type for that region.
  // the address is and-ed with the mask to produce the base address of a
  // memory region. the result is used to do a lookup into hb_memory_map
  ////////////////////////////////////////////////////////////////////////////-
  decode_endian_type(address: uint(bits: 40)):csco_ht_endian_type is {
    var base_addr: uint(bits: 40) = (address & device_mem_map_mask);
    if(device_mem_map.key_exists(base_addr)) {
      result = device_mem_map.key(base_addr).endian_type;
    } else {
      result = little_endian;
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  // Clocks and Initialization
  ////////////////////////////////////////////////////////////////////////////////
  // A fundamental clock for a given device.  In general this clock is derived from a common sys.csco_ht_fundamental_clock,
  // but if that were a direct drive clock then every device would be forced to run at the same rate.  Therefore if a
  // direct drive clock is to be used, this method can be overridden to provide a unique clock for each device.
  event csco_ht_device_fundamental_clock;
  ////////////////////////////////////////////////////////////////////////////////
  // Use different events to track the power and reset signals in a dut versus a 'e' system
  ////////////////////////////////////////////////////////////////////////////////
  // Monitor the Power and Reset signals and watch them for changes.
  event power_change is change('(PowerOk)')@sim;
  on power_change {
    if('(PowerOk)' == 1) {
      PWROK = TRUE;
    }else {
      PWROK = FALSE;
    };
  };
  event e_power_change is @sys.csco_ht_sys_power_change;
  on e_power_change {
    if(sys.csco_ht_using_verilog_simulator == FALSE) {
      PWROK = sys.csco_ht_sys_PWROK;
    };
  };
  event reset_change is change('(Reset_L)')@sim;
  on reset_change {
    if('(Reset_L)' == 0) {
      RESET_L = TRUE;
    }else {
      RESET_L = FALSE;
    };
  };
  event e_reset_change is @sys.csco_ht_sys_reset_change;
  on e_reset_change {
    if(sys.csco_ht_using_verilog_simulator == FALSE) {
      RESET_L = sys.csco_ht_sys_RESET_L;
    };
  };
  // These are flags used by the testbench to maintain the current state of the reset and power signals
  PWROK : bool;
  keep PWROK == FALSE;
  RESET_L : bool;  // reset is an active low signal.
  keep RESET_L == FALSE;

  Reset_L : string;
  keep soft Reset_L == "HT_RESET_L";
  PowerOk : string;
  keep soft PowerOk == "HT_POWEROK";

  // Device level event used to initiate a complete reset.
  event reset;
  // Persistant counter that records the number of resets of the device.
  reset_count : uint;
  keep reset_count == 0;
  on reset {
    reset_count += 1;
  };
  ////////////////////////////////////////////////////////////////////////////////
  // General Device Scoreboards
  ////////////////////////////////////////////////////////////////////////////////
  // Use this scoreboard to keep track of any outstanding transactions via tracking their requests and response.
  // This is primarily for reads and non-posted writes.
  scoreboard : csco_ht_scoreboard;
  ////////////////////////////////////////////////////////////////////////////////
  // Device Counters
  ////////////////////////////////////////////////////////////////////////////////
  config_write_counter : uint;
  keep config_write_counter == 0;
  config_read_counter : uint;
  keep config_read_counter == 0;

  flush_counter : uint;
  keep flush_counter == 0;
  write_counter : uint;
  keep write_counter == 0;
  read_counter : uint;
  keep read_counter == 0;
  broadcast_counter : uint;
  keep broadcast_counter == 0;
  fence_counter : uint;
  keep fence_counter == 0;

  flush_error_counter : uint;
  keep flush_error_counter == 0;
  write_error_counter : uint;
  keep write_error_counter == 0;
  read_error_counter : uint;
  keep read_error_counter == 0;
  broadcast_error_counter : uint;
  keep broadcast_error_counter == 0;
  ////////////////////////////////////////////////////////////////////////////////
  // Scenario List - A list of all scenarios that this device will execute.
  ////////////////////////////////////////////////////////////////////////////////
  csco_ht_scene : list of csco_ht_scenario;
  keep soft csco_ht_scene.size() == 0;
  ReadWrite_double: bool;
  keep soft ReadWrite_double == FALSE;
  NoFence: bool;
  keep soft NoFence == FALSE;
  TxQueueDepth: uint;
  keep soft TxQueueDepth == 12;
  ////////////////////////////////////////////////////////////////////////////////
  // A few parameters that define the behavior of the device with respect to inboud traffic.
  ////////////////////////////////////////////////////////////////////////////////
  max_memory_latency : uint;
  keep soft max_memory_latency == 10;
  max_io_delay : uint;
  keep soft max_io_delay == 10;
  max_tunneling_delay : uint;
  keep soft max_tunneling_delay == 10;
  ////////////////////////////////////////////////////////////////////////////////
  // gen_tx_SrcTag() - This method will generate a new unique combination of UnitID and SrcTag to allow a new NonPosted transaction
  // to be generated.  It will stall in the method indefinitly until a valid combination can be created.
  ////////////////////////////////////////////////////////////////////////////////
  gen_tx_SrcTag(UnitID : uint (bits:5)):uint (bits:5) @csco_ht_device_clock is {
    if(control_block.free_SrcTag(UnitID) == FALSE) {
      wait true(control_block.free_SrcTag(UnitID) == TRUE);
    };
    result = control_block.new_SrcTag(UnitID);
  };
  ////////////////////////////////////////////////////////////////////////////////
  // schedule_packet() - This method will take a packet that is to be transmitted and place it in the channels scheduling queue.
  // Additionally it will maintain the appropriate scoreboards and counters for the transmit path.  This method should be the
  // only means of accessing the channel on the transmit side.
  ////////////////////////////////////////////////////////////////////////////////
  schedule_packet(channel : csco_ht_channel_interface, packet : csco_ht_packet) is {
    // fixme, moved scoreboard.score() call to scenario
    //scoreboard.post(packet);
    record_history(tx, packet);
    case(packet.command()) {
      Flush: {
        flush_counter += 1;
      };
      Write: {
        if(packet.Address()[39:24] == 16'hFDFE) {
          config_write_counter += 1;
        }else {
          write_counter += 1;
        };
      };
      Read: {
        if(packet.Address()[39:24] == 16'hFDFE) {
          config_read_counter += 1;
        }else {
          read_counter += 1;
        };
      };
      Broadcast: {
        broadcast_counter += 1;
      };
      Fence: {
        fence_counter += 1;
      };
    };
    channel.schedule_csco_ht_tx_packet(packet);
  };
  ////////////////////////////////////////////////////////////////////////////////
  // score_tx_pkt() - method to allow for an external scoreboard hook
  // score_rx_pkt() - method to allow for an external scoreboard hook
  ////////////////////////////////////////////////////////////////////////////////
  score_tx_pkt(pkt: csco_ht_packet) is empty;
  score_rx_pkt(pkt: csco_ht_packet) is empty;
  ////////////////////////////////////////////////////////////////////////////////
  // accept_packets() - This TCM is the primary method to handle all inbound packets for this device.  It will retrieve
  // packets from the receive channel and will direct them to the appropriate memory and then generate the necessary response.
  // Secondary TCMs are called from this method to perform device specific actions on the accepted requests.
  ////////////////////////////////////////////////////////////////////////////////
  accept_packets(channel : csco_ht_channel_interface)@csco_ht_device_clock is {
    var current_packet : csco_ht_packet;
    var current_address : uint (bits:40);

    var response : csco_ht_packet;

    first of {
      {
        // Have the TCM monitor and wait for a reset signal.  This will allow the
        // device to reinitialize itself at any time.  Otherwise, the TCM could get stuck waiting
        // for an event that may never happen.
        wait @reset @reset;
      };
      {
        wait true(control_block.physical_link_initialized(channel.channel_identifier) == TRUE);
        ////////////////////////////////////////////////////////////////////////////////
        while(control_block.physical_link_initialized(channel.channel_identifier) && control_block.EOC(channel.channel_identifier) == FALSE) {
          wait true(!channel.rx_acceptance_queue.is_empty());
          // Once there are packets ready, pull them from the queue, classifiy them, execute them and then respond to them.
          current_packet = channel.rx_acceptance_queue.pop0();
          record_history(rx, current_packet);
          case(current_packet.command()) {
            Flush: {
              // A Flush is used to make sure that all posted writes have been pushed up
              // to memory, so by this point it's purpose is done, just send a TgtDone.
              var UnitID : uint (bits:5);

              // call external scoreboard if full scoring is not enabled
              if(scoreboard.score_level != csco_ht_full_scoring) {
                score_rx_pkt(current_packet);
              };

              if(device_function == Host_Bridge) {
                UnitID = current_packet.UnitID();
              }else {
                UnitID = control_block.current_UnitID[0];
              };
              response = generate_TgtDone(PassPW, UnitID, current_packet.SrcTag(), 0, 0);

              // call external scoreboard if full scoring is not enabled
              if(scoreboard.score_level != csco_ht_full_scoring) {
                score_tx_pkt(response);
              };

              schedule_packet(channel, response);
            };
            Write: {
              var write_error : bool;
              var current_data_packet : csco_ht_packet;
              var data_length : uint;
              var write_mask : uint (bits:4);
              var write_data : list of bit;

              // call external scoreboard if full scoring is not enabled
              if(scoreboard.score_level != csco_ht_full_scoring) {
                score_rx_pkt(current_packet);
              };

              current_address = current_packet.Address();
              current_data_packet = current_packet.retrieve_data_packet();
              data_length = current_packet.Mask_Count();

              if(current_packet.DataType() == DoubleWord_Write) {
                // Check to see if the write exceeds a 64 byte region.
                check_memory_boundary_violation(DoubleWord, current_address, data_length);
              }else {
                if(data_length != 0) { // Check to see if there is an associated data, or just a byte mask
                  // Check to see if the write exceeds a 32 byte region.
                  // Remember that a Byte Write has 1 doubleword of mask data.
                  data_length -= 1;
                  check_memory_boundary_violation(Byte, current_address, data_length);
                };
              };
              // Determine where this write is destined for.
              case(control_block.decode_address(current_address)) {
                Memory_Mapped_IO: {
                  write_error = memory_write(current_packet);
                };
                Interrupts: {
                  if(device_function != Host_Bridge) {
                    sys.csco_ht_errors.report_protocol_error(ILLEGAL_INTERRUPT,
                                                         "A non-Host Bridge device accepted an interrupt.",
                                                         control_block, channel.channel_identifier);
                  };
                  if(current_packet.TriggerMode() == Level) { // An EOI is required
                    var eoi_response : csco_ht_packet;
                    eoi_response = gen_directed_eoi(current_packet.passing_rule, current_packet.SeqID(), current_packet.Vector());
                    schedule_packet(channel, eoi_response);
                  };
                  var message : string;
                  messagef(MEDIUM, "Interrupt RX : %s", current_packet.print_csco_ht_packet_string());
                };
                PCI_IO: {
                  write_error = pci_io_write(current_packet);
                };
                PCI_Configuration: {
                  if(current_address[39:24] == 16'hFDFE) { // Type 0 access.
                    if(current_packet.DataType() == Byte_Write) {
                      // Check to ensure that the write is only of length 1 DW.
                      if(data_length != 0) {
                        dut_error("A Byte Write to configuration space was received and is greater than 1 DW in length.");
                      };
                      write_mask = current_data_packet.Byte_Write'command'Mask[3:0];
                      write_data = current_data_packet.Byte_Write'command'Data;
                    }else {
                      // Check to ensure that the write is only of length 1 DW.
                      if(data_length != 0) {
                        dut_error("A DoubleWord Write to configuration space was received and is greater than 1 DW in length.");
                      };
                      write_mask = 4'hF;
                      write_data = current_data_packet.DoubleWord_Write'command'Data;
                    };
                    // Update the control block so it knows which channel received this packet.
                    control_block.csr_inbound_link = channel.channel_identifier;
                    write_error = control_block.csr_write(current_address, write_mask, write_data[write_data.size()-1:0]);
                  }else { // Type 1 access.
                    //Figure out how to deal with PCI type 1 accesses.
                    dut_error("A PCI Type 1 request was accepted and needs to be forwarded appropriately.");
                  };
                };
                default: {
                  dut_error("An HT Device has accepted a packet which does not belong to this device.  Illegal Address.");
                };
              };
              // A response is required.
              if(current_packet.vChan == NonPosted) {
                var UnitID : uint (bits:5);
                if(device_function == Host_Bridge) {
                  UnitID = current_packet.UnitID();
                }else {
                  UnitID = control_block.current_UnitID[0];
                };
                response = generate_TgtDone(PassPW, UnitID, current_packet.SrcTag(), write_error.as_a(bit), 0);

                // call external scoreboard if full scoring is not enabled
                if(scoreboard.score_level != csco_ht_full_scoring) {
                  score_tx_pkt(response);
                };

                schedule_packet(channel, response);
              };
            };
            Read: {
              var expected_size : uint;
              var read_data : list of bit;
              var read_error : bool;

              // call external scoreboard if full scoring is not enabled
              if(scoreboard.score_level != csco_ht_full_scoring) {
                score_rx_pkt(current_packet);
              };

              current_address = current_packet.Address();
              if(current_packet.DataType() == DoubleWord_Read) {
                // Check to see if the read exceeds a 64 byte region.
                expected_size = current_packet.Mask_Count();
                check_memory_boundary_violation(DoubleWord, current_address, expected_size);
              }else {
                // For Byte Reads the size is always 1 doubleword (A size of 0);
                expected_size = 0;
              };
              // Determine where the read is destined for.
              case(control_block.decode_address(current_address)) {
                Memory_Mapped_IO: {
                  read_data = memory_read(current_packet);
                };
                PCI_IO: {
                  read_data = pci_io_read(current_packet);
                };
                PCI_Configuration: {
                  var read_mask : uint (bits:4);
                  if(current_address[39:24] == 16'hFDFE) { // Type 0 access.
                    if(current_packet.DataType() == Byte_Read) {
                      read_mask = current_packet.Mask_Count();
                    }else {
                      if(expected_size != 0) {
                        dut_error("A DoubleWord Read to configuration space was received and is greater than 1 DW in length.");
                      };
                      read_mask = 4'hF;
                    };
                    // Update the control block so it knows which channel received this packet.
                    control_block.csr_inbound_link = channel.channel_identifier;
                    read_data = control_block.csr_read(current_address, read_mask);
                  }else { // Type 1 access.
                    // Figure out how to deal with PCI type 1 accesses.
                    dut_error("A PCI Type 1 request was accepted and needs to be forwarded appropriately.");
                  };
                };
                default: {
                  dut_error("An HT Device has accepted a packet which does not belong to this device.  Illegal Address.");
                };
              };
              if(read_data == {} || read_data.size() != (expected_size+1)*32) {
                // An error was encoutered while reading, return random data and set the error flag.
                gen read_data keeping {
                  .size() == (expected_size+1)*32;
                };
                read_error = TRUE;
              }else {
                read_error = FALSE;
              };

              var UnitID : uint (bits:5);
              if(device_function == Host_Bridge) {
                UnitID = current_packet.UnitID();
              }else {
                UnitID = control_block.current_UnitID[0];
              };

              var endian_type: csco_ht_endian_type;
              // get the endianness based on the address
              endian_type = decode_endian_type(current_address);

              response = generate_RdResponse(current_packet.DataType(), current_packet.Read'command'response_passing_rule,
                                             UnitID, current_packet.SrcTag(), read_error.as_a(bit), 0, read_data,
                                             endian_type);

              // call external scoreboard if full scoring is not enabled
              if(scoreboard.score_level != csco_ht_full_scoring) {
                score_tx_pkt(response);
              };

              schedule_packet(channel, response);
            };
            RdResponse: {
              var scoreboard_error : csco_ht_error_code;
              var token : uint(bits: 10);
              scoreboard_error = scoreboard.add_response(current_packet);
              if(scoreboard_error != NO_ERROR) {
                var error_msg : string;
                error_msg = "RdResponse does not have a associated request in this devices scoreboard";
                sys.csco_ht_errors.report_transaction_error(scoreboard_error, error_msg, control_block, NULL, current_packet);
              };

              token = current_packet.generate_token();

              // If the NXA bit is set, it overrides the importance of the Error bit.
              if(current_packet.NXA() == 1) {
                read_error_counter += 1;
                var error_msg : string;
                var request : csco_ht_packet = scoreboard.get_completed_request(token);
                var response : csco_ht_packet = current_packet;
                error_msg = appendf("%s Request to Unknown Address occured", request.command);
                sys.csco_ht_errors.report_transaction_error(READ_NXA, error_msg, control_block, request, response);
              }else if(current_packet.ErrorFlag() == 1) {
                read_error_counter += 1;
                var error_msg : string;
                var request : csco_ht_packet = scoreboard.get_completed_request(token);
                var response : csco_ht_packet = current_packet;
                error_msg = appendf("%s Error occured : Invalid Data", request.command);
                sys.csco_ht_errors.report_transaction_error(READ_ERROR, error_msg, control_block, request, response);
              };

              // if scoreboard level is not set to score all transactions, then call external scoreboard hook
              // also check to make sure that the transaction does not exist on the completed transaction queue
              if((scoreboard.score_level != csco_ht_full_scoring) && !scoreboard.completed_transactions.key_exists(token)) {
                score_rx_pkt(current_packet);
                control_block.retire_SrcTag(token[9:5], token[4:0]);
              } else if(scoreboard.collect_config_response(token) == FALSE) {
                // If it is a config response, let the initiating method clean it up.
                // If it is not a config response check it against the data and clean up the transaction
                scoreboard.check_response_data(token);
                scoreboard.retire_transaction(token);
                control_block.retire_SrcTag(token[9:5], token[4:0]);
              };
            };
            TgtDone: {
              var scoreboard_error : csco_ht_error_code;
              scoreboard_error = scoreboard.add_response(current_packet);
              if(scoreboard_error != NO_ERROR) {
                var error_msg : string;
                error_msg = "TgTDone does not have a associated request in this devices scoreboard";
                sys.csco_ht_errors.report_transaction_error(scoreboard_error, error_msg, control_block, NULL, current_packet);
              };

              // If the NXA bit is set, it overrides the importance of the Error bit.
              if(current_packet.NXA() == 1) {
                var error_msg : string;
                var request : csco_ht_packet = scoreboard.get_completed_request(current_packet.generate_token());
                var response : csco_ht_packet = current_packet;
                error_msg = appendf("%s Request to Unknown Address occured", request.command);
                case(request.command) {
                  Flush : {
                    flush_error_counter += 1;
                    sys.csco_ht_errors.report_transaction_error(FLUSH_NXA, error_msg, control_block, request, response);
                  };
                  Write : {
                    write_error_counter += 1;
                    sys.csco_ht_errors.report_transaction_error(WRITE_NXA, error_msg, control_block, request, response);
                  };
                  Broadcast : {
                    broadcast_error_counter += 1;
                    sys.csco_ht_errors.report_transaction_error(BROADCAST_NXA, error_msg, control_block, request, response);
                  };
                  default : {
                    dut_error("Attempting to decode a bad TgtDone packet and cannot determine source request.");
                  };
                };
              }else if(current_packet.ErrorFlag() == 1) {
                var error_msg : string;
                var request : csco_ht_packet = scoreboard.get_completed_request(current_packet.generate_token());
                var response : csco_ht_packet = current_packet;
                error_msg = appendf("%s Error occured : Invalid Request", request.command);
                case(request.command) {
                  Flush : {
                    flush_error_counter += 1;
                    sys.csco_ht_errors.report_transaction_error(FLUSH_ERROR, error_msg, control_block, request, response);
                  };
                  Write : {
                    write_error_counter += 1;
                    sys.csco_ht_errors.report_transaction_error(WRITE_ERROR, error_msg, control_block, request, response);
                  };
                  Broadcast : {
                    broadcast_error_counter += 1;
                    sys.csco_ht_errors.report_transaction_error(BROADCAST_ERROR, error_msg, control_block, request, response);
                  };
                  default : {
                    dut_error("Attempting to decode a bad TgtDone packet and cannot determine source request.");
                  };
                };
              };
              var token : uint(bits: 10) = current_packet.generate_token();

              if((scoreboard.score_level != csco_ht_full_scoring) && !scoreboard.completed_transactions.key_exists(token)) {
                score_rx_pkt(current_packet);
                control_block.retire_SrcTag(token[9:5], token[4:0]);
              } else if(scoreboard.transaction_complete(token)) {
                // Check to see if the transaction is complete.  (This is key for non posted broadcasts.)
                if(scoreboard.collect_config_response(token) == FALSE) {
                  // If it is not a config response clean up the transaction;
                  scoreboard.retire_transaction(token);
                  control_block.retire_SrcTag(token[9:5], token[4:0]);
                };
              };
            };
            Broadcast: {
              var message : string;
              messagef(MEDIUM, "Broadcast RX : %s", current_packet.print_csco_ht_packet_string());
              // A response is required.

              // call external scoreboard if full scoring is not enabled
              if(scoreboard.score_level != csco_ht_full_scoring) {
                score_rx_pkt(current_packet);
              };

              if(current_packet.vChan == NonPosted) {
                var UnitID : uint (bits:5);
                if(device_function == Host_Bridge) {
                  UnitID = current_packet.UnitID();
                }else {
                  UnitID = control_block.current_UnitID[0];
                };
                response = generate_TgtDone(PassPW, UnitID, current_packet.SrcTag(), 0, 0);
                schedule_packet(channel, response);
              };
              if(control_block.decode_address(current_packet.Address()) == Interrupts) {
                var message : string;
                messagef(MEDIUM, "Interrupt RX : %s", current_packet.print_csco_ht_packet_string());
              };
            };
            Fence: {
              // The purpose of a fence is to provide a barrier to posted writes, so
              // it can be discarded at this level.
            };
            // Nop
            // Sync
            // Reserved
            // DoubleWord_Read
            // DoubleWord_Write
            // Byte_Read
            // Byte_Write
            default: {
              dut_error("Either a Info or a Data packet has been presented to the Device, this should not happen.");
            };
          };
        };
      };
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  // check_memory_boundary_violation() is a method will ensure that a read or write stays within the necessary 32 or 64 byte
  // regions.  If a transaction violates these rules an error is generated.
  ////////////////////////////////////////////////////////////////////////////////
  check_memory_boundary_violation(access : csco_ht_data_type, addr : uint (bits:40), count : uint (bits:4)) is {
    // A Byte Read by definition only looks at a single doubleword of data, so it is impossible
    // for it to cross a boundary.

    // A DoubleWord Read must request data within a 64-byte region.
    // The count field ranges from 0 to 15, representing 1 to 16 doublewords of data to
    // be operated on.
    if(access == DoubleWord && count >= 5'b10000 - addr[5:2]) {
      var message : string;
      message = appendf("A DoubleWord Read/Write packet of length %d exceeded a 64 byte region of memory at address %x.", count, addr);
      sys.csco_ht_errors.report_transaction_error(ILLEGAL_MEM_ACCESS, message, control_block, NULL, NULL);
    }else if(access == Byte && count >= 4'b1000 - addr[4:2]) {
      var message : string;
      message = appendf("A Byte Write packet of length %d exceeded a 32 byte region of memory at address %x.", count, addr);
      sys.csco_ht_errors.report_transaction_error(ILLEGAL_MEM_ACCESS, message, control_block, NULL, NULL);
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  // forward_packets() is the primary TCM for handling all packets received by this device that are not destined for this
  // device.  If the device is a Tunnel it needs to forward the packet and if it is a Host or an EOC then it needs to generate
  // the correct error response for the request packet.  A device specific method is called from this TCM to determine how the
  // packets should be handled.
  ////////////////////////////////////////////////////////////////////////////////
  forward_packets(inbound_channel : csco_ht_channel_interface, outbound_channel : csco_ht_channel_interface)@csco_ht_device_clock is {
    var current_packet : csco_ht_packet;

    first of {
      {
        // Have the TCM monitor and wait for a reset signal.  This will allow the
        // device to reinitialize itself at any time.  Otherwise, the TCM could get stuck waiting
        // for an event that may never happen.
        wait @reset @reset;
      };
      {
        wait true(control_block.physical_link_initialized(inbound_channel.channel_identifier) == TRUE);
        ////////////////////////////////////////////////////////////////////////////////
        while(control_block.physical_link_initialized(inbound_channel.channel_identifier) && control_block.EOC(inbound_channel.channel_identifier) == FALSE) {
          wait true(!inbound_channel.rx_forwarding_queue.is_empty());

          current_packet = inbound_channel.rx_forwarding_queue.pop0();
          record_history(forwarding, current_packet);
          // Device specific method to handle forwarded traffic.
          process_forwarded_traffic(inbound_channel, outbound_channel, current_packet);
          wait cycle;
        };
      };
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  // generate_traffic() is the primary TCM for creating traffic from this node.  Additinally it implements a watchdog timer that
  // is used at the system level to determine when this device is done generating traffic and is used to determine when to stop
  // the current simulation.  Once the device is up and running there are two specific methods called to provide device level
  // traffic.  The first method called is the standard generate_device_traffic() which is a empty method that users should extend
  // to generate a specific traffic pattern.  The second method is generate_random_traffic() which is  method that will create
  // a random mix of transactions and generate at least min_random_transactions of them.
  ////////////////////////////////////////////////////////////////////////////////

  generate_traffic(channel : csco_ht_channel_interface)@csco_ht_device_clock is {
    first of {
      {
        // Have the TCM monitor and wait for a reset signal.  This will allow the
        // device to reinitialize itself at any time.  Otherwise, the TCM could get stuck waiting
        // for an event that may never happen.
        wait @reset @reset;
      };
      {
        print control_block.logical_link_initialized(channel.channel_identifier);
        wait true(control_block.logical_link_initialized(channel.channel_identifier) == TRUE);
        // Once the link has been initialized, wait for a user specified event to occur before generating traffic.
        //wait @sys.csco_ht_test_control.csco_ht_start_traffic;
        print device_function;
        if(device_function == Host_Bridge || device_function == EOC_IO_Device ||
           (device_function == Tunnel_IO_Device && control_block.read_link_orientation(channel.channel_identifier) == TX_Upstream_RX_Downstream_Interface)) {
             // Check to make sure that this is the only link for a Host or EOC Device, or that it is the Upstream link
             // for a Tunnel before generating traffic. (In general the downstream link on a tunnel does not generate traffic.)
             // If the channel is marked as End Of Chain, don't transmit any packets on it.
             print control_block.EOC(channel.channel_identifier);
             if(control_block.EOC(channel.channel_identifier) == FALSE) {
               if (genCrcError == TRUE && doneGenCrcError == TRUE) {
                 for each (interface) in csco_ht_interfaces {
                   interface.csco_ht_link.generate_bad_CRCs = TRUE;
                 };
               } else {
                 for each (interface) in csco_ht_interfaces {
                   interface.csco_ht_link.generate_bad_CRCs = FALSE;
                 };
               };
               doneGenCrcError = FALSE;
               start generate_Ex_traffic(channel);
               // Go to a device specific method for traffic patterns, or just fall through tcm.
               generate_device_traffic(channel);
               // Wait until a responses have been received.
               wait true(scoreboard.pending_transactions.is_empty() && scoreboard.completed_transactions.is_empty());
             };
             // Only the port on a device which can legally generate traffic can indicate that it is done.
             traffic_generation_completed();
             // Once all of the response have been received fall through, thus causing this device to stop patting the watchdog in sys.
           };
      };
    };
  };

 remoteBoot(channel : csco_ht_channel_interface)@csco_ht_device_clock is {
    first of {
      {
        // Have the TCM monitor and wait for a reset signal.  This will allow the
        // device to reinitialize itself at any time.  Otherwise, the TCM could get stuck waiting
        // for an event that may never happen.
        wait @reset @reset;
      };
      {
        print remote_boot;
        for each (link) in csco_ht_interfaces {
          print link.csco_ht_link.rx_link_initialized;
          wait true(link.csco_ht_link.rx_link_initialized == TRUE);
        };
        start generate_Ex_traffic(channel);
        // Go to a device specific method for traffic patterns, or just fall through tcm.
        generate_device_traffic(channel);
        // Wait until a responses have been received.
        wait true(scoreboard.pending_transactions.is_empty() && scoreboard.completed_transactions.is_empty());
        // Only the port on a device which can legally generate traffic can indicate that it is done.
        // traffic_generation_completed();
        // Once all of the response have been received fall through, thus causing this device to stop patting the watchdog in sys
      };
    };
  };


  ////////////////////////////////////////////////////////////////////////////////
  // generate_device_traffic is the method that will actually kick off the user defined
  // scenarios.
  ////////////////////////////////////////////////////////////////////////////////
  generate_device_traffic(channel : csco_ht_channel_interface)@csco_ht_device_clock is {
    for each in csco_ht_scene {
//      it.run_scenario(channel);
      it.init_scenario(channel);
    };
  };

  generate_Ex_traffic(channel : csco_ht_channel_interface)@csco_ht_device_clock is empty;

  ////////////////////////////////////////////////////////////////////////////////
  // Methods that need to be implemented by child objects. These are device specific methods that are used to provide
  // a specific behavior based on the nature of the device.
  // Note: For write methods a boolean return value of TRUE indicates a write error and FALSE indicates a write success.
  // For read methods an empty list is returned on read errors and a non-empty list indicates a read success.
  ////////////////////////////////////////////////////////////////////////////////
  // A standard access to memory
  memory_write(packet : csco_ht_packet):bool @csco_ht_device_clock is empty;
  memory_read(packet : csco_ht_packet):list of bit @csco_ht_device_clock is empty;
  // A standard access to pci io space
  pci_io_write(packet : csco_ht_packet):bool @csco_ht_device_clock is empty;
  pci_io_read(packet : csco_ht_packet):list of bit @csco_ht_device_clock is empty;
  // Handle traffic received for this device, but not destined for this device.
  process_forwarded_traffic(inbound_channel : csco_ht_channel_interface, outbound_channel : csco_ht_channel_interface, current_packet : csco_ht_packet)@csco_ht_device_clock is empty;
  ////////////////////////////////////////////////////////////////////////////////
  // If a Tunnel device has one of its links shut down (the EOC bit is set) it will jump into this method to drop or return
  // NXA's to all traffic inbound to that channel.
  ////////////////////////////////////////////////////////////////////////////////
  generate_EOC_error(inbound_channel: csco_ht_channel_interface, request : csco_ht_packet)@csco_ht_device_clock is {
    var current_forward_packet : csco_ht_packet;
    var current_accept_packet : csco_ht_packet;
    wait true(control_block.physical_link_initialized(inbound_channel.channel_identifier) == TRUE);
    ////////////////////////////////////////////////////////////////////////////////
    while(control_block.physical_link_initialized(inbound_channel.channel_identifier) && control_block.EOC(inbound_channel.channel_identifier) == TRUE) {
      all of {
        {
          wait true(!inbound_channel.rx_forwarding_queue.is_empty());
          current_forward_packet = inbound_channel.rx_forwarding_queue.pop0();
          record_history(forwarding, current_forward_packet);
          generate_forwarding_error(inbound_channel, current_forward_packet);
          wait cycle;
        };
        {
          wait true(!inbound_channel.rx_acceptance_queue.is_empty());
          current_accept_packet = inbound_channel.rx_acceptance_queue.pop0();
          record_history(rx, current_accept_packet);
          generate_forwarding_error(inbound_channel, current_accept_packet);
          wait cycle;
        };
      };
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  // generate_forwarding_error() is a method that will create the appropriate response packet for a request that is not
  // destined for this device and cannot be forwarded.  (e.g. EOC devices, Host Bridges and Tunnels with a disabled link will
  // call this method.)  Once the response has been generated it will be sent out on the appropriate channel.
  ////////////////////////////////////////////////////////////////////////////////
  generate_forwarding_error(inbound_channel: csco_ht_channel_interface, request : csco_ht_packet) is {
    var response : csco_ht_packet;
    var UnitID : uint (bits:5);
    var Bridge : bit;
    case(request.command) {
      Flush: {
        if(control_block.read_link_orientation(inbound_channel.channel_identifier) == TX_Downstream_RX_Upstream_Interface) {
          // This is an upstream request so the bridge bit must be set even if this is not a Host Bridge
          Bridge = 1;
          UnitID = request.UnitID();
        }else {
          Bridge = 0;
          UnitID = control_block.current_UnitID[0];
        };
        gen response keeping {
          .command == TgtDone;
          .passing_rule == PassPW;
          .TgtDone'command'UnitID == UnitID;
          .TgtDone'command'Bridge == Bridge;
          .TgtDone'command'SrcTag == request.SrcTag();
          .TgtDone'command'ErrorFlag == 1;
          .TgtDone'command'NXA == 1;
        };
        inbound_channel.schedule_csco_ht_tx_packet(response);
      };
      Write: {
        if(request.vChan == NonPosted) { // A response is required for this write.
          if(control_block.read_link_orientation(inbound_channel.channel_identifier) == TX_Downstream_RX_Upstream_Interface) {
            // This is an upstream request so the bridge bit must be set even if this is not a Host Bridge
            Bridge = 1;
            UnitID = request.UnitID();
          }else {
            Bridge = 0;
            UnitID = control_block.current_UnitID[0];
          };
          gen response keeping {
            .command == TgtDone;
            .passing_rule == PassPW;
            .TgtDone'command'UnitID == UnitID;
            .TgtDone'command'Bridge == Bridge;
            .TgtDone'command'SrcTag == request.SrcTag();
            .TgtDone'command'ErrorFlag == 1;
            .TgtDone'command'NXA == 1;
          };
          inbound_channel.schedule_csco_ht_tx_packet(response);
        };
      };
      Read: {
        if(control_block.read_link_orientation(inbound_channel.channel_identifier) == TX_Downstream_RX_Upstream_Interface) {
          // This is an upstream request so the bridge bit must be set even if this is not a Host Bridge
          Bridge = 1;
          UnitID = request.UnitID();
        }else {
          Bridge = 0;
          UnitID = control_block.current_UnitID[0];
        };
        var response_size : uint (bits:4);
        if(request.DataType() == DoubleWord_Read) {
          response_size = request.Mask_Count();
        }else {
          // For Byte Reads the size is always 1 doubleword (A size of 0);
          response_size = 0;
        };
        var read_data : list of bit;
        gen read_data keeping {
          .size() == (response_size + 1) * 32;
          for each (data_bit) in it {
            data_bit == 1;
          };
        };
        gen response keeping {
          .command == RdResponse;
          .passing_rule == request.Read'command'response_passing_rule;
          .RdResponse'command'UnitID == UnitID;
          .RdResponse'command'Bridge == Bridge;
          .RdResponse'command'SrcTag == request.SrcTag();
          .RdResponse'command'ErrorFlag == 1;
          .RdResponse'command'Count == response_size;
          .RdResponse'command'NXA == 1;
          .RdResponse'command'assoc_data.command == request.DataType();
          request.DataType() == Byte_Read => .RdResponse'command'assoc_data.Byte_Read'command'Data == read_data;
          request.DataType() == DoubleWord_Read => .RdResponse'command'assoc_data.DoubleWord_Read'command'Data == read_data;
        };
        inbound_channel.schedule_csco_ht_tx_packet(response);
      };
      Broadcast: {
        // Silently drop this.
      };
      Fence: {
        // Silently drop this.
      };
      default: {
        dut_error("Cannot generate an Error Response to an unknown packet.");
      };
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  // generic_memory_write() will allow a write request to be written to a standard 'e' memory that is modeled as a
  // keyed list.
  ////////////////////////////////////////////////////////////////////////////////
  generic_memory_write(mem : list (key:addr) of csco_ht_memory_element, packet : csco_ht_packet) is {
    var addr : uint (bits:40);
    var data : csco_ht_packet;
    var write_count : uint (bits:4);

    addr = packet.Address();
    data = packet.retrieve_data_packet();
    write_count = packet.Mask_Count();

    if(packet.DataType() == Byte_Write) {
      if(write_count != 0) { // If packet has no associated data with it, only a write mask, do nothing.
        write_count -= 1;
        prep_generic_memory_location(mem, addr[39:2], write_count);
        // The count indicates the number of DW's to be written and the mask is per byte.
        for i from 0 to (write_count+1)*4-1 do {
          if(data.Byte_Write'command'Mask[i:i] == 1) {
            mem.key(addr[39:2]+i/4).data[(i%4+1)*8-1:(i%4)*8] = data.Byte_Write'command'Data[((i+1)*8)-1:(i*8)];
          };
        };
      };
    }else {
      prep_generic_memory_location(mem, addr[39:2], write_count);
      for i from 0 to write_count do {
        mem.key(addr[39:2]+i).data = data.DoubleWord_Write'command'Data[((i+1)*32)-1:(i*32)];
      };
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  // generic_memory_read() will allow a read request to be read to a standard 'e' memory that is modeled as a
  // keyed list.
  ////////////////////////////////////////////////////////////////////////////////
  generic_memory_read(mem : list (key:addr) of csco_ht_memory_element, packet : csco_ht_packet):list of bit is {
    var addr : uint (bits:40);

    addr = packet.Address();

    if(packet.DataType() == Byte_Read) {
      // Just read an entire 32 bit block (one doubleword) and use the mask to determine what's valid.
      // If a read has no side effects then the mask can just be ignored.
      prep_generic_memory_location(mem, addr[39:2], 0);
      result = %{mem.key(addr[39:2]).data};
    }else {
      var read_length : uint (bits:4);
      read_length = packet.Mask_Count();
      prep_generic_memory_location(mem, addr[39:2], read_length);
      for i from 0 to read_length do {
        result.add(%{mem.key(addr[39:2]+i).data});
      };
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  // prep_generic_memory_location() will create an entry in a standard 'e' memory that is modeled as a
  // keyed list. Note: Data in newly created memory entries is considered to be uninitialized and is generated randomly.
  ////////////////////////////////////////////////////////////////////////////////
  prep_generic_memory_location(mem :  list (key:addr) of csco_ht_memory_element, mem_index : uint (bits:38), region_size : uint (bits:4)) is {
    for i from 0 to region_size do {
      if(!mem.key_exists(mem_index+i)) {
        var new_element : csco_ht_memory_element;
        gen new_element keeping {
          .addr == mem_index+i;
        };
        mem.add(new_element);
      };
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  // Directed request packet generation
  // These are a series of methods used for users to create directed HT requests.
  // They allow a user to specify the major fields of the packet without worrying about the
  // guts of an HT request.  Each method will return a newly created packet.
  ////////////////////////////////////////////////////////////////////////////////
  // gen_directed_flush()
  // passing_rule - Dictates if the PassPW bit will be set. NoPassPW or PassPW are the possible choices.
  // SeqID - Is the sequence number to be used for this request.
  ////////////////////////////////////////////////////////////////////////////////
  gen_directed_flush(passing_rule : csco_ht_passing_rules, SeqID : uint (bits:4)):csco_ht_packet @csco_ht_device_clock is {
    var temp_packet : csco_ht_packet;
    var UnitID : uint (bits:5);
    var SrcTag : uint (bits:5);

    UnitID = control_block.current_UnitID[0];
    SrcTag = gen_tx_SrcTag(UnitID);

    gen temp_packet keeping {
      .command == Flush;
      .passing_rule == passing_rule;
      .Flush'command'SeqID == SeqID;
      .Flush'command'UnitID == UnitID;
      .Flush'command'SrcTag == SrcTag;
    };
    result = temp_packet;
  };
  ////////////////////////////////////////////////////////////////////////////////
  // gen_directed_write()
  // data_type - Specifies if the request is to be a DoubleWord_Write or Byte_Write.  DoubleWord or Byte are the choices.
  // vc - Selects either the Posted or NonPosted virtual channels. Posted or NonPosted are the choices.
  // passing_rule - Dictates if the PassPW bit will be set. NoPassPW or PassPW are the possible choices.
  // SeqID - Is the sequence number to be used for this request.
  // data - Is the data to be written.
  // addr - The 40 bit address
  // access_type - Specifies if the request is to be a Coherent or NonCoherent
  // BW_latency_req - Specifies if the request is to be a Normal or Isochronous
  // Compat -Specify the Compat bit

  ////////////////////////////////////////////////////////////////////////////////
  gen_directed_write(data_type : csco_ht_packet_encoding, vc : csco_ht_virtual_channel, access_type:csco_ht_mem_access_type, BW_latency_req:csco_ht_BW_latency_req, passing_rule : csco_ht_passing_rules, SeqID : uint (bits:4), data : list of bit, addr : uint (bits:40)):csco_ht_packet @csco_ht_device_clock is {
    var temp_packet : csco_ht_packet;
    var data_size : uint (bits:4);
    var UnitID : uint (bits:5);
    var SrcTag : uint (bits:5);
    var endian_type : csco_ht_endian_type = decode_endian_type(addr);

    data_size = data.size()/32 - 1;

    UnitID = control_block.current_UnitID[0];
    if(vc == NonPosted) {
      SrcTag = gen_tx_SrcTag(UnitID);
    };

    gen temp_packet keeping {
      .command == Write;
      .endian_type == endian_type;
      .vChan == vc;
      .passing_rule == passing_rule;
      data_type == Byte_Write => .Write'command'data_type == Byte;
      data_type == DoubleWord_Write => .Write'command'data_type == DoubleWord;
      .Write'command'access_type == access_type;
      .Write'command'BW_latency_req == BW_latency_req;
      .Write'command'SeqID == SeqID;
      .Write'command'UnitID == UnitID;
      vc == NonPosted => .Write'command'SrcTag_Rsv == SrcTag;
      .Write'command'Compat == 0;
      .Write'command'Mask_Count == data_size;
      .Write'command'Addr == addr;
      data_type == Byte_Write => .Write'command'assoc_data.Byte_Write'command'temp_mask == data[3:0];
      data_type == Byte_Write => .Write'command'assoc_data.Byte_Write'command'Mask == data[31:0];
      data_type == Byte_Write => .Write'command'assoc_data.Byte_Write'command'Data == %{data[data.size()-1:32]};
      data_type == DoubleWord_Write => .Write'command'assoc_data.DoubleWord_Write'command'Data == data;
    };
    result = temp_packet;
  };
  ////////////////////////////////////////////////////////////////////////////////
  // gen_directed_interrupt()
  // data_type - Specifies if the request is to be a DoubleWord_Write or Byte_Write.  DoubleWord or Byte are the choices.
  // vc - Selects either the Posted or NonPosted virtual channels. Posted or NonPosted are the choices.
  // passing_rule - Dictates if the PassPW bit will be set. NoPassPW or PassPW are the possible choices.
  // SeqID - Is the sequence number to be used for this request.
  // data - Is the data to be written.
  // addr - The 40 bit address
  ////////////////////////////////////////////////////////////////////////////////
  gen_directed_interrupt(passing_rule : csco_ht_passing_rules, SeqID : uint (bits:4), data : uint, MT : csco_ht_interrupt_type, TM : csco_ht_interrupt_mode, DM : csco_ht_interrupt_destination, IntrDest : byte, Vector : byte):csco_ht_packet @csco_ht_device_clock is {
    var temp_packet : csco_ht_packet;
    var UnitID : uint (bits:5);

    UnitID = control_block.current_UnitID[0];

    gen temp_packet keeping {
      .command == Write;
      .Write'command'interrupt == TRUE;
      .passing_rule == passing_rule;
      .Write'command'SeqID == SeqID;
      .Write'command'UnitID == UnitID;
      .Write'command'MT == MT;
      .Write'command'TM == TM;
      .Write'command'DM == DM;
      .Write'command'IntrDest == IntrDest;
      .Write'command'Vector == Vector;
      .Write'command'assoc_data.Byte_Write'command'temp_mask == data[3:0];
      .Write'command'assoc_data.Byte_Write'command'Mask == data[31:0];
    };
    result = temp_packet;
  };
  ////////////////////////////////////////////////////////////////////////////////
  // gen_directed_read()
  // data_type - Specifies if the request is to be a DoubleWord_Read or Byte_Read.  DoubleWord or Byte are the choices.
  // resp_passing_rule - Will dictate if the PassPW bit will be set for the response. NoPassPW or PassPW are the possible choices.
  // passing_rule - Dictates if the PassPW bit will be set. NoPassPW or PassPW are the possible choices.
  // SeqID - Is the sequence number to be used for this request.
  // data_size - Is either the count field for DoubleWord request or the byte mask field for Byte requests.
  // addr - The 40 bit address
  // access_type - Specifies if the request is to be a Coherent or NonCoherent
  // BW_latency_req - Specifies if the request is to be a Normal or Isochronous
  // Compat: - Specify the Compat bit
  ////////////////////////////////////////////////////////////////////////////////
  gen_directed_read(data_type : csco_ht_packet_encoding, access_type:csco_ht_mem_access_type, BW_latency_req:csco_ht_BW_latency_req, resp_passing_rule : csco_ht_passing_rules, passing_rule : csco_ht_passing_rules, SeqID : uint (bits:4), data_size : uint (bits:4), addr : uint (bits:40)):csco_ht_packet @csco_ht_device_clock is {
    var temp_packet : csco_ht_packet;
    var UnitID : uint (bits:5);
    var SrcTag : uint (bits:5);
    var endian_type: csco_ht_endian_type = decode_endian_type(addr);

    UnitID = control_block.current_UnitID[0];
    SrcTag = gen_tx_SrcTag(UnitID);

    gen temp_packet keeping {
      .command == Read;
      .endian_type == endian_type;
      .passing_rule == passing_rule;
      .Read'command'response_passing_rule == resp_passing_rule;
      data_type == Byte_Read => .Read'command'data_type == Byte;
      data_type == DoubleWord_Read => .Read'command'data_type == DoubleWord;
      .Read'command'access_type == access_type;
      .Read'command'BW_latency_req == BW_latency_req;
      .Read'command'SeqID == SeqID;
      .Read'command'UnitID == UnitID;
      .Read'command'SrcTag_Rsv == SrcTag;
      .Read'command'Compat == 0;
      .Read'command'Mask_Count == data_size;
      .Read'command'Addr == addr;
    };

    result = temp_packet;
  };
  ////////////////////////////////////////////////////////////////////////////////
  // gen_directed_broadcast()
  // vc - Selects either the Posted or NonPosted virtual channels. Posted or NonPosted are the choices.
  // passing_rule - Dictates if the PassPW bit will be set. NoPassPW or PassPW are the possible choices.
  // SeqID - Is the sequence number to be used for this request.
  // addr - The 40 bit address
  ////////////////////////////////////////////////////////////////////////////////
  gen_directed_broadcast(vc : csco_ht_virtual_channel, passing_rule : csco_ht_passing_rules, SeqID : uint (bits:4), addr : uint (bits:40)):csco_ht_packet @csco_ht_device_clock is {
    var temp_packet : csco_ht_packet;
    var UnitID : uint (bits:5);
    var SrcTag : uint (bits:5);
    var endian_type : csco_ht_endian_type = decode_endian_type(addr);

    UnitID = control_block.current_UnitID[0];
    if(vc == NonPosted) {
      SrcTag = gen_tx_SrcTag(UnitID);
    };

    gen temp_packet keeping {
      .command == Broadcast;
      .endian_type == endian_type;
      .vChan == vc;
      .passing_rule == passing_rule;
      .Broadcast'command'SeqID == SeqID;
      .Broadcast'command'UnitID == UnitID;
      vc == NonPosted => .Broadcast'command'SrcTag_Rsv == SrcTag;
      .Broadcast'command'Addr == addr;
    };
    result = temp_packet;
  };
  ////////////////////////////////////////////////////////////////////////////////
  // gen_directed_eoi()
  // passing_rule - Dictates if the PassPW bit will be set. NoPassPW or PassPW are the possible choices.
  // SeqID - Is the sequence number to be used for this request.
  // vector - interrupt vector
  ////////////////////////////////////////////////////////////////////////////////
  gen_directed_eoi(passing_rule : csco_ht_passing_rules, SeqID : uint (bits:4), Vector : byte):csco_ht_packet @csco_ht_device_clock is {
    var temp_packet : csco_ht_packet;
    var UnitID : uint (bits:5);

    UnitID = control_block.current_UnitID[0];

    gen temp_packet keeping {
      .command == Broadcast;
      .Broadcast'command'interrupt == TRUE;
      .passing_rule == passing_rule;
      .Broadcast'command'SeqID == SeqID;
      .Broadcast'command'UnitID == UnitID;
      .Broadcast'command'Vector == Vector;
    };
    result = temp_packet;
  };
  ////////////////////////////////////////////////////////////////////////////////
  // gen_directed_fence()
  // passing_rule - Dictates if the PassPW bit will be set. NoPassPW or PassPW are the possible choices.
  // SeqID - Is the sequence number to be used for this request.
  ////////////////////////////////////////////////////////////////////////////////
  gen_directed_fence(passing_rule : csco_ht_passing_rules, SeqID : uint (bits:4)):csco_ht_packet @csco_ht_device_clock is {
    var temp_packet : csco_ht_packet;
    var UnitID : uint (bits:5);

    UnitID = control_block.current_UnitID[0];

    gen temp_packet keeping {
      .command == Fence;
      .passing_rule == passing_rule;
      .Fence'command'SeqID == SeqID;
      .Fence'command'UnitID == UnitID;
    };
    result = temp_packet;
  };
  ////////////////////////////////////////////////////////////////////////////////
  // Directed response packet generation
  // These are a series of methods used for interal methods to create directed HT response.
  // Each method will return a newly created packet.
  ////////////////////////////////////////////////////////////////////////////////
  // generate_RdResponse()
  // data_type - Specifies if the response is to be a DoubleWord or Byte response.  DoubleWord or Byte are the choices.
  // passing_rule - Dictates if the PassPW bit will be set. NoPassPW or PassPW are the possible choices.
  // UnitID - Is the UnitID for the response. This should be generated from the associated request.
  // srctag - Is the SrcTag. This should be generated from the associated request.
  // ErrorFlag - A bit used to indicate that an error occured with the read.
  // NXA - A bit used to indicate that the read address could not be found.
  // data - Is the data that was read..
  ////////////////////////////////////////////////////////////////////////////////
  generate_RdResponse(data_type : csco_ht_packet_encoding, passing_rule : csco_ht_passing_rules, UnitID : uint (bits:5),  srctag : uint (bits:5), errorFlag : bit, NXA : bit, data : list of bit, endian_type: csco_ht_endian_type):csco_ht_packet is {
    var temp_packet : csco_ht_packet;
    var data_size : uint (bits:4);

    data_size = data.size()/32 - 1;

    gen temp_packet keeping {
      .command == RdResponse;
      .endian_type == endian_type;
      .passing_rule == passing_rule;
      .RdResponse'command'UnitID == UnitID;
      me.device_function == Host_Bridge => .RdResponse'command'Bridge == 1;
      me.device_function == Tunnel_IO_Device or me.device_function == EOC_IO_Device => .RdResponse'command'Bridge == 0;
      .RdResponse'command'SrcTag == srctag;
      .RdResponse'command'ErrorFlag == errorFlag;
      .RdResponse'command'Count == data_size;
      .RdResponse'command'NXA == NXA;
      .RdResponse'command'assoc_data.command == data_type;
      data_type == Byte_Read => .RdResponse'command'assoc_data.Byte_Read'command'Data == data;
      data_type == DoubleWord_Read => .RdResponse'command'assoc_data.DoubleWord_Read'command'Data == data;
    };
    result = temp_packet;
  };
  ////////////////////////////////////////////////////////////////////////////////
  // generate_TgtDone()
  // passing_rule - Dictates if the PassPW bit will be set. NoPassPW or PassPW are the possible choices.
  // UnitID - Is the UnitID for the response. This should be generated from the associated request.
  // srctag - Is the SrcTag. This should be generated from the associated request.
  // ErrorFlag - A bit used to indicate that an error occured with the write or broadcast.
  // NXA - A bit used to indicate that the write address could not be found.
  ////////////////////////////////////////////////////////////////////////////////
  generate_TgtDone(passing_rule : csco_ht_passing_rules, UnitID : uint (bits:5),  srctag : uint (bits:5), errorFlag : bit, NXA : bit):csco_ht_packet is {
    var temp_packet : csco_ht_packet;

    gen temp_packet keeping {
      .command == TgtDone;
      .passing_rule == passing_rule;
      .TgtDone'command'UnitID == UnitID;
      me.device_function == Host_Bridge => .TgtDone'command'Bridge == 1;
      me.device_function == Tunnel_IO_Device or me.device_function == EOC_IO_Device => .TgtDone'command'Bridge == 0;
      .TgtDone'command'SrcTag == srctag;
      .TgtDone'command'ErrorFlag == errorFlag;
      .TgtDone'command'NXA == NXA;
    };
    result = temp_packet;
  };
  ////////////////////////////////////////////////////////////////////////////////
  // record_history() is a method that will time stamp all packets and place them in the appropriate
  // history lists if the record_device_history flag is set.
  ////////////////////////////////////////////////////////////////////////////////
  record_history(traffic : csco_ht_traffic_type, packet : csco_ht_packet) is {
    if(record_device_history) {
      var new_packet_record : csco_ht_packet_record;
      gen new_packet_record keeping {
        .packet == packet;
        .time == sys.time;
      };
      case(traffic) {
        tx: {
          tx_history.add(new_packet_record);
        };
        rx: {
          rx_history.add(new_packet_record);
        };
        forwarding: {
          forwarding_history.add(new_packet_record);
        };
        default: {
          dut_error("Trying to record an illegal type of traffic.");
        };
      };
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  // print_traffic_params() extends the csco_ht_device.e method to print the device
  // traffic parameters settings after initialization has completed
  ////////////////////////////////////////////////////////////////////////////////
  print_traffic_params()@csco_ht_device_clock is {
    var message : string;
    first of {
      {
        // Have the TCM monitor and wait for a reset signal.  This will allow the
        // device to reinitialize itself at any time.  Otherwise, the TCM could get stuck waiting
        // for an event that may never happen.
        wait @reset @reset;
      };
      {
        wait @sys.csco_ht_test_control.csco_ht_model_initialized;
        var unit_id: list of uint(bits:5);
        unit_id = control_block.capability_register.read_UnitID();
        messagef(LOW, "Device: %s, UnitID: %d", device_function, unit_id[0]);
        messagef(MEDIUM, "Traffic Parameters:") {
          outf("   max_memory_latency=0x%x", max_memory_latency);
          outf("   max_io_delay=0x%x", max_io_delay);
          outf("   max_tunneling_delay=0x%x", max_tunneling_delay);
        };
      };
    };
  };
  /////////////////////////////////////////////////////////////////////////////
  // print_device_info() is a standard method for formating informational
  // messages to be printed.
  //
  // spadix - print_device_info() was replaced by messagef(), short_name()
  //          and a message logger
  /////////////////////////////////////////////////////////////////////////////

  logger : message_logger;
    keep soft logger.tags == {NORMAL};

  short_name(): string is {
    var function_name : string;
    case(control_block.device_function) {
      Host_Bridge:      { function_name = "(HostBridge)" };
      Tunnel_IO_Device: { function_name = "(TunnelDev)" };
      EOC_IO_Device:    { function_name = "(EOCDev)" };
    };
    result = append(control_block.device_reference_name, function_name);
  };

  ////////////////////////////////////////////////////////////////////////////////
  // print_device_records() is a standard method for outputing text in a standardized format.
  // Additionally, the messages are only printed if the display_records flag is set.
  ////////////////////////////////////////////////////////////////////////////////
  print_device_records() is {
    messagef(LOW, "Transaction Counters for: %s(%s)\n", control_block.device_reference_name, device_function) {
      if(device_function == Host_Bridge) {
        outf("  Flush=%d Write=%d Read=%d Broadcast=%d Fence=%d : Config Write=%d Config Read=%d\n\n", flush_counter, write_counter, read_counter, broadcast_counter, fence_counter, config_write_counter, config_read_counter);
      }else {
        outf("  Flush=%d Write=%d Read=%d Broadcast=%d Fence=%d\n", flush_counter, write_counter, read_counter, broadcast_counter, fence_counter);
      };
    };
    if(display_records == TRUE) {
      messagef(LOW, "\n  Transaction History for: %s(%s)\n", control_block.device_reference_name, device_function) {
        outf("  ////////// TX_History\n");
        for each in tx_history {
          it.print_record();
        };
        outf("  ////////// RX_History\n");
        for each in rx_history {
          it.print_record();
        };
        outf("  ////////// Forwarding_History\n");
        for each in forwarding_history {
          it.print_record();
        };
      };
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  // reset() is a method used to initialize any internal fields that handle current
  // transmitting traffic.  Essentially this is a method to shutdown then entire device
  // and prepare it to reinitialize the HT chain.
  ////////////////////////////////////////////////////////////////////////////////
  reset() is {
    emit reset;
    var message : string;
    messagef(MEDIUM, "HT Model Reset : Reset_Count = %d", reset_count);

    // Dump out the current errors that have occured, and reset the counters.
    sys.csco_ht_errors.reset();

    for each (channel) in csco_ht_interfaces {
      channel.reset();
    };
    // Allow for use hooks
    reset_cleanup();
  };
  ////////////////////////////////////////////////////////////////////////////////
  // reset_cleanup() is a method that allows a user to change the state of the model between
  // resets.
  ////////////////////////////////////////////////////////////////////////////////
  reset_cleanup() is {
    if(reset_count > 1) {
      for each (channel) in csco_ht_interfaces {
        // After the initial reset, disable the transmission of errors.
        channel.allow_tx_errors = FALSE;
      };
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  // activate_TCMs() is a tcm that is used to start all other runtime TCMs for this device.
  // This also allows for proper reset control.
  ////////////////////////////////////////////////////////////////////////////////
  activate_TCMs(reset_type : csco_ht_reset_type)@csco_ht_device_clock is {
    // Once PWROK is asserted for a warm or cold reset start the transmitters.
    for each (link) in csco_ht_interfaces {
      start link.initialize_tx_channel(reset_type);
      start link.initialize_rx_channel(reset_type);

      // Start the TCM's to manage the devices traffic.
      start accept_packets(link);
      case(device_function) {
        Host_Bridge: {
          start forward_packets(link, NULL);
        };
        Tunnel_IO_Device: {
          if(index == 0) {
            start forward_packets(csco_ht_interfaces[0], csco_ht_interfaces[1]);
          }else {
            start forward_packets(csco_ht_interfaces[1], csco_ht_interfaces[0]);
          };
        };
        EOC_IO_Device: {
          start forward_packets(link, NULL);
        };
        default: {
          dut_error("Illegal attempt to start forwarding process.");
        };
      };
      start generate_traffic(link);
      if (remote_boot == TRUE) {
        start remoteBoot(link);
      };

      // print traffic parameters after chain initialization has finished
      start print_traffic_params();
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  // initialize_device() is the primary method that will bring up a device after reset.  It is responsible
  // for monitoring the system wide reset and power ok signals, and then getting things going once they
  // have completed a reset sequence.
  ////////////////////////////////////////////////////////////////////////////////
  initialize_device()@csco_ht_device_clock is {
    var reset_type : csco_ht_reset_type;
    while(!sys.csco_ht_test_control.csco_ht_model_done) {
      first of {
        // Monitor for system reset which is signaled with either RESET_L being asserted or PWROK being deasserted.
        {
          wait true(RESET_L == TRUE);
          messagef(HIGH, "RESET_L is asserted");
        };
        {
          wait true(PWROK == FALSE);
          messagef(HIGH, "PWROK is deasserted");
        };
      };
      reset();

      first of {
        {
          // A cold reset is initiated by the assertion of RESET and the deassertion of PWROK
          wait true(RESET_L == TRUE && PWROK == FALSE);
//           for each (link) in csco_ht_interfaces {
//             start link.pre_reset_tx_initialization();
//           };
          reset_type = Cold;
        };
        {
          wait true(RESET_L == TRUE && PWROK == TRUE);
//           for each (link) in csco_ht_interfaces {
//             start link.pre_reset_tx_initialization();
//           };
          reset_type = Warm;
        };
      };
      // Once the reset sequence is detected, start the pre_reset sequence and then wait for PWROK
      if(reset_type == Cold) {
        wait true(PWROK == TRUE);
      };
      // Initialize the control block for this device.
      control_block.reset(reset_type);
      device_clock_freq = control_block.read_max_current_tx_clock_freq();

      // Initialize the scoreboard
      scoreboard.reset();

      activate_TCMs(reset_type);
      wait true(RESET_L == FALSE);
    };
  };

  ////////////////////////////////////////////////////////////////////////////////
  // initialization_sequence() is responsible for turning the system wide reset and power signals into
  // device level flags that are monitored by the initialize_device() method.  This method will generate
  // it's own initialization sequence if no verilog simulator is attached.
  ////////////////////////////////////////////////////////////////////////////////
  initialization_sequence()@csco_ht_device_clock is {
    if(sys.csco_ht_using_verilog_simulator == FALSE) {
      wait [5];
      // Assume power and clock to be okay, so enable the device RESET
      RESET_L = TRUE;
      wait [5];
      PWROK = TRUE;
      wait [5];
      RESET_L = FALSE;
    }else {
      // Make sure that RESET_L and PWROK get set initially. (Some simulators don't always force a "change" at the
      // beginning of the simulations).
      if('(PowerOk)' == 1) {
        PWROK = TRUE;
      }else {
        PWROK = FALSE;
      };
      if('(Reset_L)' == 0) {
        RESET_L = TRUE;
      }else {
        RESET_L = FALSE;
      };
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  // reset_sequence() is responsible for driving the system wide reset and power signals
  // so that the whole HT chain can be reset.
  ////////////////////////////////////////////////////////////////////////////////
  reset_sequence()@csco_ht_device_clock is {
    messagef(LOW, "HT eVC Model is initiating reset sequence\n");

    var error_string : string;
    error_string = appendf("HT eVC Model is initiating reset sequence");
    sys.csco_ht_errors.report_protocol_error(EVC_HT_CHAIN_RESET, error_string, control_block, 0);

    // Initiate a Soft Reset
    if(sys.csco_ht_using_verilog_simulator == FALSE) {
      sys.csco_ht_sys_RESET_L = TRUE;
      emit sys.csco_ht_sys_reset_change;
      wait [CSCO_HT_RESET_ASSERTION_LENGTH];
      sys.csco_ht_sys_RESET_L = FALSE;
      emit sys.csco_ht_sys_reset_change;
    }else {
      '(Reset_L)' = 0;
      wait [CSCO_HT_RESET_ASSERTION_LENGTH];
      '(Reset_L)' = 1;
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  // Provide a method to register the device with the csco_ht_test_control_block
  ////////////////////////////////////////////////////////////////////////////////
  register_device() is {
    testbench_ID = sys.csco_ht_test_control.register_csco_ht_device();
  };
  ////////////////////////////////////////////////////////////////////////////////
  // Provide a method to indicate that the device has completed generating traffic to the csco_ht_test_control_block
  ////////////////////////////////////////////////////////////////////////////////
  traffic_generation_completed() is {
    sys.csco_ht_test_control.device_traffic_generation_completed(testbench_ID);
  };
  ////////////////////////////////////////////////////////////////////////////////
  // Kick off the initialization methods.
  ////////////////////////////////////////////////////////////////////////////////
  run() is also {
    start initialize_device();
    start initialization_sequence();
    register_device();
  };
  ////////////////////////////////////////////////////////////////////////////////
  // Once the test is complete, print out the device specific information and check that the
  // scoreboards are empty.
  ////////////////////////////////////////////////////////////////////////////////
  check() is also {
    scoreboard.check_scoreboard();
  };
  ////////////////////////////////////////////////////////////////////////////////
  finalize() is also {
    print_device_records();
  };
  ////////////////////////////////////////////////////////////////////////////////
};
////////////////////////////////////////////////////////////////////////////////
extend Host_Bridge'device_function csco_ht_device {
  // A list of profiles for all devices on an HT chain.  This is only used by the Host Bridge.
  chain_topology : list of csco_ht_device_profile;
  keep soft chain_topology.size() == 0;
};
////////////////////////////////////////////////////////////////////////////////
struct csco_ht_device_profile {
  !target_device_function : csco_ht_device_encoding; // Use this to track what each device is on the chain.
  device_header : csco_ht_device_header;
  !capability_header : csco_ht_capability_header;

  // method to return a list of all the regs for this device
  // FIXME (lihuang) leave empty for now until the read_masks and func_write_masks are put in
  //   for the device and capability regs
  get_device_regs(): list of csco_ht_basic_reg is {
  };
};
////////////////////////////////////////////////////////////////////////////////
'>
