////////////////////////////////////////////////////////////////////////////       // -
// Title         : HT IO Device
// Project       : HT Testbench
////////////////////////////////////////////////////////////////////////////       // -
// File          : csco_ht_io_device.e
// Author        : Brian Van Essen  <vanessen@cisco.com>
// Created       : 25.01.2001
// Last modified : 25.01.2001
////////////////////////////////////////////////////////////////////////////       // -
// Description :
// This file is an extension of the csco_ht_device.e file and provides a simple EOC
// device and a Tunnel device.  These devices have a simple memory attached to them
// and will correctly respond to any request that they receive.  Additionally, they
// will properly forward traffic and terminate the chain. Each device is also capable
// of random and directed traffic generation.  This file is essentially
// "when Tunnel_IO_Device'csco_ht_device_function csco_ht_device {};" or
// "when EOC_IO_Device'csco_ht_device_function csco_ht_device {};" block.
////////////////////////////////////////////////////////////////////////////////
// Modification history :
// 25.01.2001 : created
//////////////////////////////////////////////////////////////////////////////-

// ***** BEGIN LICENSE BLOCK *****
// Version: MPL 1.1
// The contents of this file are subject to the Mozilla Public License Version
// 1.1 (the "License"); you may not use this file except in compliance with
// the License. You may obtain a copy of the License at
// http:                                                   // www.mozilla.org/MPL/ or ./mpl_1_1.txt
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
// for the specific language governing rights and limitations under the License.
// The Original Code is Cisco Hypertransport eVC.
// The Initial Developer of the Original Code is
// Brian Van Essen, Lisa Huang, and Sean Smith.
// Portions created by the Initial Developer are Copyright (C) 2002
// the Initial Developer. All Rights Reserved.
// Contributor(s): Brian Van Essen, Lisa Huang, and Sean Smith
// ***** END LICENSE BLOCK *****

<'
////////////////////////////////////////////////////////////////////////////////
// Define a Tunnel Device
////////////////////////////////////////////////////////////////////////////////
extend Tunnel_IO_Device'device_function csco_ht_device {
  // Specify that this device has two HT interfaces.
  keep control_block.num_device_csco_ht_interfaces == 2;
  keep csco_ht_interfaces.size() == 2;
  // Currently implement a flat memory space that is addressable by all three BAR's.
  !device_memory : list (key:addr) of csco_ht_memory_element;
  ////////////////////////////////////////////////////////////////////////////////
  // These are a collection of generic methods that are initially define in csco_ht_device.e and
  // are redefined here to provide specific functionality for an individual device.
  ////////////////////////////////////////////////////////////////////////////////
  // memory_write() defines how accesses to the Tunnels memory are handled.
  ////////////////////////////////////////////////////////////////////////////////
  memory_write(packet : csco_ht_packet):bool @csco_ht_device_clock is {
    var mem_latency : uint;
    gen mem_latency keeping {
      it <= max_memory_latency;
    };
    wait [mem_latency] * cycle;

    generic_memory_write(device_memory, packet);
    result = FALSE;
  };
  ////////////////////////////////////////////////////////////////////////////////
  // pci_io_write() defines how accesses to the Tunnels IO space are handled.
  ////////////////////////////////////////////////////////////////////////////////
  pci_io_write(packet : csco_ht_packet):bool @csco_ht_device_clock is {
    dut_error("A Sized_Write was sent to the PCI IO space on this device, which is not implemented.");
    var io_delay : uint;
    gen io_delay keeping {
      it <= max_io_delay;
    };
    wait [io_delay] * cycle;
    result = TRUE;
  };
  ////////////////////////////////////////////////////////////////////////////////
  // memory_read() defines how accesses to the Tunnels memory are handled.
  ////////////////////////////////////////////////////////////////////////////////
  memory_read(packet : csco_ht_packet):list of bit @csco_ht_device_clock is {
    var mem_latency : uint;
    gen mem_latency keeping {
      it <= max_memory_latency;
    };
    wait [mem_latency] * cycle;

    result = generic_memory_read(device_memory, packet);
  };
  ////////////////////////////////////////////////////////////////////////////////
  // pci_io_read() defines how accesses to the Tunnels IO space are handled.
  ////////////////////////////////////////////////////////////////////////////////
  pci_io_read(packet : csco_ht_packet):list of bit @csco_ht_device_clock is {
    dut_error("A Sized_Read was sent to the PCI IO space on this device, which is not implemented.");
    var io_delay : uint;
    gen io_delay keeping {
      it <= max_io_delay;
    };
    wait [io_delay] * cycle;
    result = {};
  };
  ////////////////////////////////////////////////////////////////////////////////
  // process_forwarded_traffic() defines how the Tunnel will respond to inbound packets that are
  // not destined for the Tunnel device.  In this case, all traffic is forwarded to the other link
  // unless the next link's EOC bit is set, in which case the Tunnel acts as a end of chain device
  // and will send error responses.
  ////////////////////////////////////////////////////////////////////////////////
  process_forwarded_traffic(inbound_channel : csco_ht_channel_interface, outbound_channel : csco_ht_channel_interface, current_packet : csco_ht_packet)@csco_ht_device_clock is {
    var tunneling_delay : uint;
    gen tunneling_delay keeping {
      it <= max_tunneling_delay;
    };
    wait [tunneling_delay] * cycle;
    if(control_block.EOC(outbound_channel.channel_identifier) == TRUE) {
      // The outbound link has been set to be the end of chain so return an error for any packet destined for that link.
      generate_forwarding_error(inbound_channel, current_packet);
    }else {
      outbound_channel.schedule_csco_ht_tx_packet(current_packet)
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  // print_device_records() extends the csco_ht_device.e method to include the print out of the Tunnel memory.
  ////////////////////////////////////////////////////////////////////////////////
  print_device_records() is also {
    if(display_records == TRUE) {
      outf("////////// Main Memory\n");
      for each in device_memory.sort(.addr) {
        it.print_mem();
      };
    };
  };
};
////////////////////////////////////////////////////////////////////////////////
// Define an EOC Device (End Of Chain)
////////////////////////////////////////////////////////////////////////////////
extend EOC_IO_Device'device_function csco_ht_device {
  // Specify that this device has a single HT interface.
  keep control_block.num_device_csco_ht_interfaces == 1;
  keep csco_ht_interfaces.size() == 1;
  // Currently implement a flat memory space that is addressable by all three BAR's.
  !device_memory : list (key:addr) of csco_ht_memory_element;
  ////////////////////////////////////////////////////////////////////////////////
  // These are a collection of generic methods that are initially define in csco_ht_device.e and
  // are redefined here to provide specific functionality for an individual device.
  ////////////////////////////////////////////////////////////////////////////////
  // memory_write() defines how accesses to the EOCs memory are handled.
  ////////////////////////////////////////////////////////////////////////////////
  memory_write(packet : csco_ht_packet):bool @csco_ht_device_clock is {
    var mem_latency : uint;
    gen mem_latency keeping {
      it <= max_memory_latency;
    };
    wait [mem_latency] * cycle;

    generic_memory_write(device_memory, packet);
    result = FALSE;
  };
  ////////////////////////////////////////////////////////////////////////////////
  // pci_io_write() defines how accesses to the EOCs IO space are handled.
  ////////////////////////////////////////////////////////////////////////////////
  pci_io_write(packet : csco_ht_packet):bool @csco_ht_device_clock is {
    dut_error("A Sized_Write was sent to the PCI IO space on this device, which is not implemented.");
    var io_delay : uint;
    gen io_delay keeping {
      it <= max_io_delay;
    };
    wait [io_delay] * cycle;
    result = TRUE;
  };
  ////////////////////////////////////////////////////////////////////////////////
  // memory_read() defines how accesses to the EOCs memory are handled.
  ////////////////////////////////////////////////////////////////////////////////
  memory_read(packet : csco_ht_packet):list of bit @csco_ht_device_clock is {
    var mem_latency : uint;
    gen mem_latency keeping {
      it <= max_memory_latency;
    };
    wait [mem_latency] * cycle;

    result = generic_memory_read(device_memory, packet);
  };
  ////////////////////////////////////////////////////////////////////////////////
  // pci_io_read() defines how accesses to the EOCs IO space are handled.
  ////////////////////////////////////////////////////////////////////////////////
  pci_io_read(packet : csco_ht_packet):list of bit @csco_ht_device_clock is {
    dut_error("A Sized_Read was sent to the PCI IO space on this device, which is not implemented.");
    var io_delay : uint;
    gen io_delay keeping {
      it <= max_io_delay;
    };
    wait [io_delay] * cycle;
    result = {};
  };
  ////////////////////////////////////////////////////////////////////////////////
  // process_forwarded_traffic() defines how the EOC will respond to inbound packets that are
  // not destined for the EOC device.  In this case, all traffic is responded to with an error
  // response.
  ////////////////////////////////////////////////////////////////////////////////
  process_forwarded_traffic(inbound_channel : csco_ht_channel_interface, outbound_channel : csco_ht_channel_interface, current_packet : csco_ht_packet)@csco_ht_device_clock is {
    var tunneling_delay : uint;
    gen tunneling_delay keeping {
      it <= max_tunneling_delay;
    };
    wait [tunneling_delay] * cycle;
    generate_forwarding_error(inbound_channel, current_packet);
  };
  ////////////////////////////////////////////////////////////////////////////////
  // print_device_records() extends the csco_ht_device.e method to include the print out of the EOC memory.
  ////////////////////////////////////////////////////////////////////////////////
  print_device_records() is also {
    if(display_records == TRUE) {
      outf("////////// Main Memory\n");
      for each in device_memory.sort(.addr) {
        it.print_mem();
      };
    };
  };
};
////////////////////////////////////////////////////////////////////////////////
'>
