////////////////////////////////////////////////////////////////////////////
// Title         : HT Scenario Definitions
// Project       : HT Testbench
////////////////////////////////////////////////////////////////////////////
// File          : csco_ht_scenario_definitions.e
// Author        : Brian Van Essen  <vanessen@braves.cisco.com>
// Created       : 04.05.2001
// Last modified : 04.05.2001
////////////////////////////////////////////////////////////////////////////
// Description :
// This file is for the user to extend and define new scenario's that are
// used for traffic generation.
////////////////////////////////////////////////////////////////////////////////
// Modification history :
// 04.05.2001 : created
////////////////////////////////////////////////////////////////////////////

// ***** BEGIN LICENSE BLOCK *****
// Version: MPL 1.1
// The contents of this file are subject to the Mozilla Public License Version
// 1.1 (the "License"); you may not use this file except in compliance with
// the License. You may obtain a copy of the License at
// http:                                                   // www.mozilla.org/MPL/ or ./mpl_1_1.txt
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
// for the specific language governing rights and limitations under the License.
// The Original Code is Cisco Hypertransport eVC.
// The Initial Developer of the Original Code is
// Brian Van Essen, Lisa Huang, and Sean Smith.
// Portions created by the Initial Developer are Copyright (C) 2002
// the Initial Developer. All Rights Reserved.
// Contributor(s): Brian Van Essen, Lisa Huang, and Sean Smith
// ***** END LICENSE BLOCK *****


<'

////////////////////////////////////////////////////////////////////////////////
// Define the different types of scenarios that are available.
////////////////////////////////////////////////////////////////////////////////

extend csco_ht_scenario_kind : [INTERMIX, INTERMIX_OVERLAY, WRITE_IN_ORDER, READ_IN_ORDER, READ_WRITE_IN_ORDER,
                            INTERRUPT, READ_POR_REGS, WALK_ONES_REGS, INTERMIX_REGS];

////////////////////////////////////////////////////////////////////////////////
// INTERMIX is a scenario that performs random read/write/flush/fence/broadcast
// traffic to a targeted address range.  The transactions are unique in that they
// target no overlapping regions of memory.
////////////////////////////////////////////////////////////////////////////////

extend INTERMIX'kind csco_ht_scenario {
  // In the INTERMIX scenario each transaction is unique and non overlapping.
  !outstanding_writes : list of csco_ht_packet;

  generate_transactions()@parent_device.csco_ht_device_clock is {
    var transaction_cmd : csco_ht_packet_encoding;
    var transaction : csco_ht_packet;
    var current_target : csco_ht_scenario_element;
    var endian_type: csco_ht_endian_type;

    // Check to see if there are still outstanding writes or address that need to be written too.
    if(target_address_list.size() == 0 && outstanding_writes.size() == 0) {
      done = TRUE;
      return;
    };

    // Generate a legal transaction type.
    gen transaction_cmd keeping {
      parent_device.device_function == Host_Bridge => it == Write or it == Read or it == Broadcast;
      parent_device.device_function == Tunnel_IO_Device => it == Write or it == Read or it == Flush or it == Fence;
//      parent_device.device_function == EOC_IO_Device => it == Write or it == Read or it == Flush or it == Fence;
      parent_device.device_function == EOC_IO_Device && parent_device.NoFence == FALSE => it == Write or it == Read or it == Flush or it == Fence;
      parent_device.device_function == EOC_IO_Device && parent_device.NoFence == TRUE => it == Write or it == Read or it == Flush;

      // Once all of the address have been written to, send reads for all outstanding writes.
      target_address_list.size() == 0 => it == Read;
    };

    case(transaction_cmd) {
      Flush: {
        var write_transaction : csco_ht_packet;
        var flush_transaction : csco_ht_packet;
        // To generate a meaningful flush sequence a posted write must occur, then a flush, followed by a read to the
        // same location. The read should be allowed to pass all posted traffic, but not be sent until the TgtDone for the
        // flush is received.  (Thus the write should already be commited.)

        // Generate a Posted Write
        var target_address_index : uint;
        // Select a random list index
        gen target_address_index keeping {
          it < target_address_list.size();
        };
        // Use that random number to select an address block
        current_target = target_address_list[target_address_index];
        target_address_list.delete(target_address_index);

        // get the endianness based on the address
        endian_type = parent_device.decode_endian_type(current_target.addr);

//KMM        gen write_transaction keeping {
//KMM          .command == Write;
//KMM          .endian_type == endian_type;
//KMM          .vChan == Posted;
//KMM          .Write'command'UnitID == parent_device.control_block.current_UnitID[0];
//KMM          .Write'command'SeqID == 0;
//KMM          .Write'command'Compat == 0;                 // By default always force the compat bit to 0.  FIX THIS LATER.
//KMM          .Write'command'data_type == Byte => .Write'command'Mask_Count <= current_target.num_doublewords+1;
//KMM          .Write'command'data_type == DoubleWord => .Write'command'Mask_Count <= current_target.num_doublewords;
//KMM          .Write'command'Addr == current_target.addr;
//KMM        };
        gen write_transaction keeping {
          it is a Write'command csco_ht_packet(tmp)                                                 and
          tmp.UnitID == parent_device.control_block.current_UnitID[0]           and
          tmp.SeqID == 0                                                        and
          tmp.Compat == 0                                                       and// By default always force the compat bit to 0.  FIX THIS LATER.
          tmp.Addr == current_target.addr;

          it is a Write'command csco_ht_packet(tmp2) and tmp2.data_type == Byte => tmp2.Mask_Count <= current_target.num_doublewords+1;
          it is a Write'command csco_ht_packet(tmp3) and tmp3.data_type == DoubleWord => tmp3.Mask_Count <= current_target.num_doublewords;

          .endian_type == endian_type;
          .vChan == Posted;
        };
        // Check to see if the transaction has occupied the entire address space of the address block.
        gen_replacement_address(write_transaction.Mask_Count(), current_target);
        parent_device.scoreboard.post(write_transaction);
        schedule_transaction(write_transaction);

        var SrcTag : uint (bits:5) = parent_device.gen_tx_SrcTag(parent_device.control_block.current_UnitID[0]); // Wait until a SrcTag is available.
        // Generate a Flush
//KMM        gen flush_transaction keeping {
//KMM          .command == Flush;
//KMM          .passing_rule == NoPassPW;
//KMM          .Flush'command'SeqID == 0;
//KMM          .Flush'command'UnitID == parent_device.control_block.current_UnitID[0];
//KMM          .Flush'command'SrcTag == SrcTag;
//KMM        };
        gen flush_transaction keeping {
          it is a Flush'command csco_ht_packet(tmp)                                 and
          .passing_rule == NoPassPW                                             and
          tmp.SeqID == 0                                                        and
          tmp.UnitID == parent_device.control_block.current_UnitID[0]           and
          tmp.SrcTag == SrcTag;
        };
        if(parent_device.scoreboard.score_level == csco_ht_full_scoring) {
          parent_device.scoreboard.post(flush_transaction);
        } else {
          parent_device.score_tx_pkt(flush_transaction);
        };
        schedule_transaction(flush_transaction);
        var response_token : uint (bits:10);
        response_token = %{flush_transaction.UnitID(), flush_transaction.SrcTag()};
        // Wait until the response has been received and dequeued before sending the followup read.
        wait true(parent_device.scoreboard.transaction_pending(response_token) == FALSE);

        // Generate a Read, but make it unordered so that it will pass the write if the flush failed.
        // Check to see if the write was a byte or a doubleword.  If it was a byte, then multiple byte reads will have to be
        // generated to check the write.  (One read for each doubleword of byte write, since byte reads are only 1 doubleword long.)
        // There must be at least 1 read transaction for each write.
        var num_read_transactions : uint = 1;
        var i : uint = 0;
        if(write_transaction.DataType() == Byte_Write) {
          num_read_transactions = write_transaction.Mask_Count();
        };
        repeat {
          var SrcTag : uint (bits:5) = parent_device.gen_tx_SrcTag(write_transaction.UnitID()); // Wait until a SrcTag is available
          var write_addr: uint(bits: 40) = (write_transaction.Address() + (i*4));

          endian_type = parent_device.decode_endian_type(write_addr);
//KMM          gen transaction keeping {
//KMM            .command == Read;
//KMM            .endian_type == endian_type;
//KMM            .passing_rule == PassPW;
//KMM            write_transaction.DataType() == Byte_Write => .Read'command'data_type == Byte;
//KMM            write_transaction.DataType() == DoubleWord_Write => .Read'command'data_type == DoubleWord;
//KMM            .Read'command'BW_latency_req == write_transaction.Write'command'BW_latency_req;
//KMM            .Read'command'access_type == write_transaction.Write'command'access_type;
//KMM            .Read'command'SeqID == 0;
//KMM            .Read'command'UnitID == write_transaction.UnitID();
//KMM            .Read'command'SrcTag_Rsv == SrcTag;
//KMM            .Read'command'Compat == write_transaction.Write'command'Compat;
//KMM            // Make sure that each subsequent read in a string of byte reads uses the next nibble of byte mask.
//KMM            write_transaction.DataType() == Byte_Write => .Read'command'Mask_Count == write_transaction.Write'command'assoc_data.Byte_Write'command'Mask[(i+1)*4-1:i*4];
//KMM            write_transaction.DataType() == DoubleWord_Write => .Read'command'Mask_Count == write_transaction.Mask_Count();
//KMM            // Make sure that each subsequent read in a string of byte reads uses the next doubleword address..
//KMM            .Read'command'Addr == write_addr;
//KMM          };
          var tmp_write_transaction : Write'command  csco_ht_packet;
          var tmp_assoc_data        : Byte_Write'command csco_ht_packet;
          tmp_write_transaction = write_transaction.as_a(Write'command csco_ht_packet);
          tmp_assoc_data  = tmp_write_transaction.assoc_data.as_a(Byte_Write'command csco_ht_packet);
          gen transaction keeping {
            it is a Read'command csco_ht_packet(tmp)                                                and
            tmp.endian_type == endian_type                                                              and
            tmp.passing_rule == PassPW                                                          and
            tmp.BW_latency_req == tmp_write_transaction.BW_latency_req                          and
            tmp.access_type == tmp_write_transaction.access_type                                and
            tmp.SeqID == 0                                                                      and
            tmp.UnitID == write_transaction.UnitID()                                            and
            tmp.SrcTag_Rsv == SrcTag                                                            and
            tmp.Compat == tmp_write_transaction.Compat                                          and
            tmp.Addr == write_addr                                                              and
            // Make sure that each subsequent read in a string of byte reads uses the next nibble of byte mask.
            (write_transaction.DataType() == Byte_Write => tmp.Mask_Count == tmp_assoc_data.Mask[(i+1)*4-1:i*4] )       and
            (write_transaction.DataType() == DoubleWord_Write => tmp.Mask_Count == write_transaction.Mask_Count() )     and
            // Make sure that each subsequent read in a string of byte reads uses the next doubleword address..
            (write_transaction.DataType() == Byte_Write => tmp.data_type == Byte )                                      and
            (write_transaction.DataType() == DoubleWord_Write => tmp.data_type == DoubleWord );
          };
          if(num_read_transactions > 1) {
            // If there are multiple reads being associated with this write request, just schedule them, but don't
            // increment the transaction counter.
            if(parent_device.scoreboard.score_level != csco_ht_no_scoring) {
              parent_device.scoreboard.post(transaction);
            } else {
              parent_device.score_tx_pkt(transaction);
            };
            schedule_transaction(transaction);
          };
          num_read_transactions -= 1;
          i += 1;
        }until(num_read_transactions == 0);
      };
      Write:{
        // If a write is generated, pick a address and data to be written, then scoreboard them for self checking.
        var target_address_index : uint;
        // Select a random list index
        gen target_address_index keeping {
          it < target_address_list.size();
        };
        // Use that random number to select an address block
        current_target = target_address_list[target_address_index];
        target_address_list.delete(target_address_index);

        endian_type = parent_device.decode_endian_type(current_target.addr);

//KMM        gen transaction keeping {
//KMM          .command == transaction_cmd;
//KMM          .endian_type == endian_type;
//KMM          parent_device.device_function == Host_Bridge => .Write'command'UnitID == 0;
//KMM          parent_device.device_function == Tunnel_IO_Device or parent_device.device_function == EOC_IO_Device => .Write'command'UnitID == parent_device.control_block.current_UnitID[0];
//KMM          // If there are no free SrcTag's then force a Posted Transaction.
//KMM          parent_device.control_block.free_SrcTag(parent_device.control_block.current_UnitID[0]) == FALSE => .vChan == Posted;
//KMM          .vChan == NonPosted => .Write'command'SrcTag_Rsv == parent_device.control_block.new_SrcTag(.Write'command'UnitID);
//KMM          .Write'command'Compat == 0; // By default always force the compat bit to 0.  FIX THIS LATER.
//KMM          .Write'command'data_type == Byte => .Write'command'Mask_Count <= current_target.num_doublewords+1;
//KMM          .Write'command'data_type == DoubleWord => .Write'command'Mask_Count <= current_target.num_doublewords;
//KMM          .Write'command'Addr == current_target.addr;
//KMM        };

        gen transaction keeping {
          it is a Write'command csco_ht_packet(tmp)                         and
          tmp.endian_type == endian_type                                and
          tmp.Addr == current_target.addr                               and
          tmp.Compat == 0; // By default always force the compat bit to 0.  FIX THIS LATER.

          it is a Write'command csco_ht_packet(tmp2) and parent_device.device_function == Host_Bridge => tmp2.UnitID == 0;
          it is a Write'command csco_ht_packet(tmp3) and (parent_device.device_function == Tunnel_IO_Device or parent_device.device_function== EOC_IO_Device )
                                                => tmp3.UnitID == parent_device.control_block.current_UnitID[0];

          // If there are no free SrcTag's then force a Posted Transaction.
          parent_device.control_block.free_SrcTag(parent_device.control_block.current_UnitID[0]) == FALSE => .vChan == Posted;

         it is a Write'command csco_ht_packet(tmp4) and .vChan == NonPosted => tmp4.SrcTag_Rsv == parent_device.control_block.new_SrcTag(tmp4.UnitID);
         it is a Write'command csco_ht_packet(tmp5) and tmp5.data_type == Byte => tmp5.Mask_Count <= current_target.num_doublewords+1;
         it is a Write'command csco_ht_packet(tmp6) and tmp6.data_type == DoubleWord => tmp6.Mask_Count <= current_target.num_doublewords;
        };

        // Check to see if the transaction has occupied the entire address space of the address block.
        gen_replacement_address(transaction.Mask_Count(), current_target);
        outstanding_writes.add(transaction);
      };
      Read: {
        var order_after_Posted_Write : bool = FALSE;
        var order_after_NonPosted_Sequential_Write : bool = FALSE;

        // Never create a read to an unwritten location.
        if(outstanding_writes.is_empty() == TRUE) {
          // Just loop through again.
          return;
        };
        var write_index : uint;
        gen write_index keeping {
          it < outstanding_writes.size();
        };
        var write_req : csco_ht_packet;
        write_req = outstanding_writes[write_index];

        // In order to ensure that a read will read valid data one of two things must occur. If the original write is NonPosted
        // and has a non zero SeqID then the write must be in an ordered sequence.  If the original write is NonPosted and
        // has a zero SeqID then the read cannot occur until the TgtDone has been received.
        // If the original write is Posted, then the SeqID does not matter and the PassPW bit needs to be forced to FALSE.
        if(write_req.vChan == Posted) {
          order_after_Posted_Write = TRUE;
        }else {
          if(write_req.SeqID() == 0) {
            var response_token : uint (bits:10);
            response_token = %{write_req.UnitID(), write_req.SrcTag()};

            if(parent_device.scoreboard.transaction_pending(response_token)) {
              // Since a entry in the scoreboard still exists, it means that a TgtDone has not yet been received.
              // Therefore it is not yet safe to send a read, as it still might pass the write.
              return;
            };
          }else {
            order_after_NonPosted_Sequential_Write = TRUE;
          };
        };

        // Check to see if the write was a byte or a doubleword.  If it was a byte, then multiple byte reads will have to be
        // generated to check the write.  (One read for each doubleword of byte write, since byte reads are only 1 doubleword long.)
        // There must be at least 1 read transaction for each write.
        var num_read_transactions : uint = 1;
        var i : uint = 0;
        if(write_req.DataType() == Byte_Write) {
          num_read_transactions = write_req.Mask_Count();
        };
        repeat {
          var SrcTag : uint (bits:5) = parent_device.gen_tx_SrcTag(write_req.UnitID()); // Wait until a SrcTag is available.
          var write_addr: uint(bits: 40) = write_req.Address()+(i*4);

          endian_type = parent_device.decode_endian_type(write_addr);
//KMM          gen transaction keeping {
//KMM            .command == transaction_cmd;
//KMM            .endian_type == endian_type;
//KMM            order_after_Posted_Write == TRUE => .passing_rule == NoPassPW;
//KMM            write_req.DataType() == Byte_Write => .Read'command'data_type == Byte;
//KMM            write_req.DataType() == DoubleWord_Write => .Read'command'data_type == DoubleWord;
//KMM            .Read'command'BW_latency_req == write_req.Write'command'BW_latency_req;
//KMM            .Read'command'access_type == write_req.Write'command'access_type;
//KMM            order_after_NonPosted_Sequential_Write == TRUE => .Read'command'SeqID == write_req.SeqID();
//KMM            .Read'command'UnitID == write_req.UnitID();
//KMM            .Read'command'SrcTag_Rsv == SrcTag;
//KMM            .Read'command'Compat == write_req.Write'command'Compat;
//KMM            // Make sure that each subsequent read in a string of byte reads uses the next nibble of byte mask.
//KMM            write_req.DataType() == Byte_Write => .Read'command'Mask_Count == write_req.Write'command'assoc_data.Byte_Write'command'Mask[(i+1)*4-1:i*4];
//KMM            write_req.DataType() == DoubleWord_Write => .Read'command'Mask_Count == write_req.Mask_Count();
//KMM            // Make sure that each subsequent read in a string of byte reads uses the next doubleword address..
//KMM            .Read'command'Addr == write_addr;
//KMM          };
          var tmp_write_req : Write'command csco_ht_packet;
          tmp_write_req = write_req.as_a(Write'command csco_ht_packet);
          var tmp_assoc_data : Byte_Write'command csco_ht_packet;
          tmp_assoc_data  = tmp_write_req.assoc_data.as_a(Byte_Write'command csco_ht_packet);
          gen transaction keeping {
            it is a Read'command csco_ht_packet(tmp)                and
            tmp.endian_type == endian_type                      and
            tmp.BW_latency_req == tmp_write_req.BW_latency_req  and
            tmp.access_type == tmp_write_req.access_type                and
            tmp.UnitID == write_req.UnitID()                    and
            tmp.SrcTag_Rsv == SrcTag                            and
            tmp.Compat == tmp_write_req.Compat                  and
            tmp.Addr == write_addr                              and
            (order_after_Posted_Write == TRUE => .passing_rule == NoPassPW)                                             and
            (write_req.DataType() == Byte_Write => tmp.data_type == Byte)                                               and
            (write_req.DataType() == DoubleWord_Write => tmp.data_type == DoubleWord)                                   and
            (order_after_NonPosted_Sequential_Write == TRUE => tmp.SeqID == write_req.SeqID()  )                        and
            // Make sure that each subsequent read in a string of byte reads uses the next nibble of byte mask.
            (write_req.DataType() == Byte_Write => tmp.Mask_Count == tmp_assoc_data.Mask[(i+1)*4-1:i*4] )       and
            (write_req.DataType() == DoubleWord_Write => tmp.Mask_Count == write_req.Mask_Count() );
            // Make sure that each subsequent read in a string of byte reads uses the next doubleword address..
          };
          if(num_read_transactions > 1) {
            // If there are multiple reads being associated with this write request, just schedule them, but don't
            // increment the transaction counter.
            if(parent_device.scoreboard.score_level == csco_ht_full_scoring) {
              parent_device.scoreboard.post(transaction);
            } else {
              parent_device.score_tx_pkt(transaction);
            };
            schedule_transaction(transaction);
          };
          num_read_transactions -= 1;
          i += 1;
        }until(num_read_transactions == 0);

        outstanding_writes.delete(write_index);
      };
      Broadcast: {
//KMM        gen transaction keeping {
//KMM          .command == transaction_cmd;
//KMM          .Broadcast'command'UnitID == parent_device.control_block.current_UnitID[0];
//KMM          parent_device.control_block.free_SrcTag(parent_device.control_block.current_UnitID[0]) == FALSE => .vChan == Posted;
//KMM          .vChan == NonPosted => .Broadcast'command'SrcTag_Rsv == parent_device.control_block.new_SrcTag(.Broadcast'command'UnitID);
//KMM        };
        gen transaction keeping {
          it is a Broadcast'command csco_ht_packet(tmp)                                     and
          tmp.UnitID == parent_device.control_block.current_UnitID[0]           and
          (parent_device.control_block.free_SrcTag(parent_device.control_block.current_UnitID[0]) == FALSE => .vChan == Posted )        and
          (.vChan == NonPosted => tmp.SrcTag_Rsv == parent_device.control_block.new_SrcTag(tmp.UnitID) );
        };
      };
      Fence: {
        var write_transaction : csco_ht_packet;
        var fence_transaction : csco_ht_packet;
        // To generate a meaningful fence sequence a posted write must occur, then a fence, followed by a read to the
        // same location. The read should be allowed to pass all posted traffic, and sent immediately after the fence.

        // Generate a Posted Write
        var target_address_index : uint;
        // Select a random list index
        gen target_address_index keeping {
          it < target_address_list.size();
        };
        // Use that random number to select an address block
        current_target = target_address_list[target_address_index];
        target_address_list.delete(target_address_index);

        endian_type = parent_device.decode_endian_type(current_target.addr);

//KMM        gen write_transaction keeping {
//KMM          .command == Write;
//KMM          .endian_type == endian_type;
//KMM          .vChan == Posted;
//KMM          .Write'command'UnitID == parent_device.control_block.current_UnitID[0];
//KMM          .Write'command'SeqID == 0;
//KMM          .Write'command'Compat == 0; // By default always force the compat bit to 0.  FIX THIS LATER.
//KMM          .Write'command'data_type == Byte => .Write'command'Mask_Count <= current_target.num_doublewords+1;
//KMM          .Write'command'data_type == DoubleWord => .Write'command'Mask_Count <= current_target.num_doublewords;
//KMM          .Write'command'Addr == current_target.addr;
//KMM        };
        gen write_transaction keeping {
          it is a Write'command csco_ht_packet(tmp)                                         and
          tmp.endian_type == endian_type                                and
          tmp.vChan == Posted                                           and
          tmp.UnitID == parent_device.control_block.current_UnitID[0]   and
          tmp.SeqID == 0                                                and
          tmp.Compat == 0                                               and  // By default always force the compat bit to 0.  FIX THIS LATER.
          (tmp.data_type == Byte => tmp.Mask_Count <= current_target.num_doublewords+1 )        and
          (tmp.data_type == DoubleWord => tmp.Mask_Count <= current_target.num_doublewords )    and
          tmp.Addr == current_target.addr;
        };
        // Check to see if the transaction has occupied the entire address space of the address block.
        gen_replacement_address(write_transaction.Mask_Count(), current_target);
        if(parent_device.scoreboard.score_level == csco_ht_full_scoring) {
          parent_device.scoreboard.post(write_transaction);
        } else {
          parent_device.score_tx_pkt(write_transaction);
        };
        schedule_transaction(write_transaction);

        // Generate a Fence
//KMM        gen fence_transaction keeping {
//KMM          .command == Fence;
//KMM          .passing_rule == NoPassPW;
//KMM          .Fence'command'SeqID == 0;
//KMM          .Fence'command'UnitID == parent_device.control_block.current_UnitID[0];
//KMM        };
        gen fence_transaction keeping {
          it is a Fence'command csco_ht_packet(tmp)                                 and
          .passing_rule == NoPassPW                                             and
          tmp.SeqID == 0                                                        and
          tmp.UnitID == parent_device.control_block.current_UnitID[0];
        };
        if(parent_device.scoreboard.score_level == csco_ht_full_scoring) {
          parent_device.scoreboard.post(fence_transaction);
        } else {
          parent_device.score_tx_pkt(fence_transaction);
        };
        schedule_transaction(fence_transaction);

        // Generate a Read, but make it unordered so that it will pass the write if the fence failed.
        // Force it to have a clear PassPW bit so that the fence can work.
        // Check to see if the write was a byte or a doubleword.  If it was a byte, then multiple byte reads will have to be
        // generated to check the write.  (One read for each doubleword of byte write, since byte reads are only 1 doubleword long.)
        // There must be at least 1 read transaction for each write.
        var num_read_transactions : uint = 1;
        var i : uint = 0;
        if(write_transaction.DataType() == Byte_Write) {
          num_read_transactions = write_transaction.Mask_Count();
        };
        repeat {
          var SrcTag : uint (bits:5) = parent_device.gen_tx_SrcTag(write_transaction.UnitID()); // Wait until a SrcTag is available.
          var write_addr: uint(bits: 40) = write_transaction.Address() + (i*4);

          endian_type = parent_device.decode_endian_type(write_addr);
//KMM          gen transaction keeping {
//KMM            .command == Read;
//KMM            .endian_type == endian_type;
//KMM            .passing_rule == NoPassPW;
//KMM            write_transaction.DataType() == Byte_Write => .Read'command'data_type == Byte;
//KMM            write_transaction.DataType() == DoubleWord_Write => .Read'command'data_type == DoubleWord;
//KMM            .Read'command'BW_latency_req == write_transaction.Write'command'BW_latency_req;
//KMM            .Read'command'access_type == write_transaction.Write'command'access_type;
//KMM            .Read'command'SeqID == 0;
//KMM            .Read'command'UnitID == write_transaction.UnitID();
//KMM            .Read'command'SrcTag_Rsv == SrcTag;
//KMM            .Read'command'Compat == write_transaction.Write'command'Compat;
//KMM            // Make sure that each subsequent read in a string of byte reads uses the next nibble of byte mask.
//KMM        write_transaction.DataType() == Byte_Write => .Read'command'Mask_Count == write_transaction.Write'command'assoc_data.Byte_Write'command'Mask[(i+1)*4-1:i*4];
//KMM            write_transaction.DataType() == DoubleWord_Write => .Read'command'Mask_Count == write_transaction.Mask_Count();
//KMM            // Make sure that each subsequent read in a string of byte reads uses the next doubleword address..
//KMM            .Read'command'Addr == write_addr;
//KMM          };
          var tmp_write_transaction : Write'command csco_ht_packet;
          tmp_write_transaction = write_transaction.as_a(Write'command csco_ht_packet);
          var tmp_assoc_data  : Byte_Write'command csco_ht_packet;
          tmp_assoc_data = tmp_write_transaction.assoc_data.as_a(Byte_Write'command csco_ht_packet);
          gen transaction keeping {
            it is a Read'command csco_ht_packet(tmp)                                                                        and
            .endian_type == endian_type                                                                                 and
            .passing_rule == NoPassPW                                                                                   and
            (write_transaction.DataType() == Byte_Write => tmp.data_type == Byte)                                       and
            (write_transaction.DataType() == DoubleWord_Write => tmp.data_type == DoubleWord)                           and
            tmp.BW_latency_req == tmp_write_transaction.BW_latency_req                                                  and
            tmp.access_type == tmp_write_transaction.access_type                                                        and
            tmp.SeqID == 0                                                                                              and
            tmp.UnitID == write_transaction.UnitID()                                                                    and
            tmp.SrcTag_Rsv == SrcTag                                                                                    and
            tmp.Compat == tmp_write_transaction.Compat                                                                  and
            // Make sure that each subsequent read in a string of byte reads uses the next nibble of byte mask.
            (write_transaction.DataType() == Byte_Write => tmp.Mask_Count == tmp_assoc_data.Mask[(i+1)*4-1:i*4] )       and
            (write_transaction.DataType() == DoubleWord_Write => tmp.Mask_Count == write_transaction.Mask_Count() )     and
            // Make sure that each subsequent read in a string of byte reads uses the next doubleword address..
            tmp.Addr == write_addr;
          };
          if(num_read_transactions > 1) {
            // If there are multiple reads being associated with this write request, just schedule them, but don't
            // increment the transaction counter.
            if(parent_device.scoreboard.score_level == csco_ht_full_scoring) {
              parent_device.scoreboard.post(transaction);
            } else {
              parent_device.score_tx_pkt(transaction);
            };
            schedule_transaction(transaction);
          };
          num_read_transactions -= 1;
          i += 1;
        }until(num_read_transactions == 0);
      };
      default: {
        dut_error("Illegal random transaction selected.\n");
      };
    };
    if(transaction != NULL) {
      if(parent_device.scoreboard.score_level == csco_ht_full_scoring) {
        parent_device.scoreboard.post(transaction);
      } else {
        parent_device.score_tx_pkt(transaction);
      };
      schedule_transaction(transaction);
    };
  };
};
////////////////////////////////////////////////////////////////////////////////
// INTERMIX_OVERLAY is a scenario that performs random read/write/flush/fence/broadcast
// traffic to a targeted address range.  The transactions are non unique in that they
// target overlapping regions of memory.
////////////////////////////////////////////////////////////////////////////////
extend INTERMIX_OVERLAY'kind csco_ht_scenario {
  !outstanding_writes : list of csco_ht_packet;
  // A flag used to indicate that multiple writes to an address region should not be generated until a read has been sent.
  never_overwrite_mem_until_read : bool;
  keep soft never_overwrite_mem_until_read == FALSE;
  // A special flag that only imposes Write ordering on 32 byte regions instead of 64byte regions.  (Used for random data generation.)
  allow_32byte_region_ordering : bool;
  keep soft allow_32byte_region_ordering == FALSE;
  // A course map of the target memory that keeps track of 32 byte regions that are targeted by currently outstanding reads
  // and writes.  This is used to ensure that mulitple requests to a single 32/64 byte region of memory are ordered so that they
  // can be accessed in an orderly and predicatable fashion.
  !target_memory_transaction_map : list (key:addr) of csco_ht_memory_transaction_element;

  generate_transactions()@parent_device.csco_ht_device_clock is {
    var transaction_cmd : csco_ht_packet_encoding;
    var transaction : csco_ht_packet;
    var current_target : csco_ht_scenario_element;
    var endian_type : csco_ht_endian_type;

    // Check to see if there are still outstanding writes or address that need to be written too.
    if(target_address_list.size() == 0 && outstanding_writes.size() == 0) {
      done = TRUE;
      return;
    };

    // Generate a legal transaction type.
    gen transaction_cmd keeping {
      parent_device.device_function == Host_Bridge => it == Write or it == Read or it == Broadcast;
      parent_device.device_function == Tunnel_IO_Device => it == Write or it == Read or it == Flush or it == Fence;
      parent_device.device_function == EOC_IO_Device => it == Write or it == Read or it == Flush or it == Fence;
      // Once all of the address have been written to, send reads for all outstanding writes.
      target_address_list.size() == 0 => it == Read;
    };

    case(transaction_cmd) {
      Flush: {
        var write_transaction : csco_ht_packet;
        var flush_transaction : csco_ht_packet;
        // To generate a meaningful flush sequence a posted write must occur, then a flush, followed by a read to the
        // same location. The read should be allowed to pass all posted traffic, but not be sent until the TgtDone for the
        // flush is received.  (Thus the write should already be commited.)

        var non_overlapping_address : bool;
        var target_address_index : uint;
        repeat {
          // Generate a Posted Write
          // Select a random list index
          gen target_address_index keeping {
            it < target_address_list.size();
          };
          // Use that random number to select an address block
          current_target = target_address_list[target_address_index];
          // Check to see if the address selected is in the same 64 byte block as an outstanding write transaction.
          // If it is, select another address. (Or 32 byte block if the flag is set.)
          if(target_memory_transaction_map.key_exists(current_target.addr[39:5]) ||
             (!allow_32byte_region_ordering && current_target.addr[5:5] == 0 && target_memory_transaction_map.key_exists(current_target.addr[39:5] + 1))) {
               non_overlapping_address = FALSE;
               wait cycle;
             }else {
               non_overlapping_address = TRUE;
             };
        }until(non_overlapping_address);
        target_address_list.delete(target_address_index);

        endian_type = parent_device.decode_endian_type(current_target.addr);

//KMM        gen write_transaction keeping {
//KMM          .command == Write;
//KMM          .endian_type == endian_type;
//KMM          .vChan == Posted;
//KMM          .Write'command'UnitID == parent_device.control_block.current_UnitID[0];
//KMM          .Write'command'SeqID == 0;
//KMM          .Write'command'Compat == 0; // By default always force the compat bit to 0.  FIX THIS LATER.
//KMM          .Write'command'data_type == Byte => .Write'command'Mask_Count <= current_target.num_doublewords+1;
//KMM          .Write'command'data_type == DoubleWord => .Write'command'Mask_Count <= current_target.num_doublewords;
//KMM          .Write'command'Addr == current_target.addr;
//KMM          // If the 32 byte flag is set, force the DW writes to be in a 32 byte region.
//KMM          allow_32byte_region_ordering == TRUE && .Write'command'data_type == DoubleWord => .Write'command'Mask_Count < 4'b1000 - .Write'command'Addr[4:2];
//KMM        };
        gen write_transaction keeping {
          it is a Write'command csco_ht_packet(tmp)                                         and
          .endian_type == endian_type                                                   and
          .vChan == Posted                                                              and
          tmp.UnitID == parent_device.control_block.current_UnitID[0]                   and
          tmp.SeqID == 0                                                                and
          tmp.Compat == 0                                                               and // By default always force the compat bit to 0.  FIX THIS LATER.
          (tmp.data_type == Byte => tmp.Mask_Count <= current_target.num_doublewords+1) and
          (tmp.data_type == DoubleWord => tmp.Mask_Count <= current_target.num_doublewords )    and
          tmp.Addr == current_target.addr                                                       and
          // If the 32 byte flag is set, force the DW writes to be in a 32 byte region.
          (allow_32byte_region_ordering == TRUE && tmp.data_type == DoubleWord => tmp.Mask_Count < 4'b1000 - tmp.Addr[4:2]);
        };

        // Don't scoreboard the transaction, just update the map and send it.
        memory_transaction_map_update(target_memory_transaction_map, write_transaction);
        // Check to see if the transaction has occupied the entire address space of the address block.
        gen_replacement_address(write_transaction.Mask_Count(), current_target);
        if(parent_device.scoreboard.score_level == csco_ht_full_scoring) {
          parent_device.scoreboard.post(write_transaction);
        } else {
          parent_device.score_tx_pkt(write_transaction);
        };
        schedule_transaction(write_transaction);

        var SrcTag : uint (bits:5) = parent_device.gen_tx_SrcTag(parent_device.control_block.current_UnitID[0]); // Wait until a SrcTag is available.
        // Generate a Flush
//KMM        gen flush_transaction keeping {
//KMM          .command == Flush;
//KMM          .endian_type == endian_type;
//KMM          .passing_rule == NoPassPW;
//KMM          .Flush'command'SeqID == 0;
//KMM          .Flush'command'UnitID == parent_device.control_block.current_UnitID[0];
//KMM          .Flush'command'SrcTag == SrcTag;
//KMM        };
        gen flush_transaction keeping {
          it is a Flush'command csco_ht_packet(tmp)                         and
          .endian_type == endian_type                                   and
          .passing_rule == NoPassPW                                     and
          tmp.SeqID == 0                                                and
          tmp.UnitID == parent_device.control_block.current_UnitID[0]   and
          tmp.SrcTag == SrcTag;
        };
        if(parent_device.scoreboard.score_level == csco_ht_full_scoring) {
          parent_device.scoreboard.post(flush_transaction);
        } else {
          parent_device.score_tx_pkt(flush_transaction);
        };
        schedule_transaction(flush_transaction);
        var response_token : uint (bits:10);
        response_token = %{flush_transaction.UnitID(), flush_transaction.SrcTag()};
        // Wait until the response has been received and dequeued before sending the followup read.
        wait true(parent_device.scoreboard.transaction_pending(response_token) == FALSE);

        // Generate a Read, but make it unordered so that it will pass the write if the flush failed.
        // Check to see if the write was a byte or a doubleword.  If it was a byte, then multiple byte reads will have to be
        // generated to check the write.  (One read for each doubleword of byte write, since byte reads are only 1 doubleword long.)
        // There must be at least 1 read transaction for each write.
        var num_read_transactions : uint = 1;
        var i : uint = 0;
        if(write_transaction.DataType() == Byte_Write) {
          num_read_transactions = write_transaction.Mask_Count();
        };
        repeat {
          var SrcTag : uint (bits:5) = parent_device.gen_tx_SrcTag(write_transaction.UnitID()); // Wait until a SrcTag is available.
          var write_addr: uint(bits: 40) = write_transaction.Address() + (i*4);

          endian_type = parent_device.decode_endian_type(write_addr);
//KMM          gen transaction keeping {
//KMM            .command == Read;
//KMM            .endian_type == endian_type;
//KMM            .passing_rule == PassPW;
//KMM            write_transaction.DataType() == Byte_Write => .Read'command'data_type == Byte;
//KMM            write_transaction.DataType() == DoubleWord_Write => .Read'command'data_type == DoubleWord;
//KMM            .Read'command'BW_latency_req == write_transaction.Write'command'BW_latency_req;
//KMM            .Read'command'access_type == write_transaction.Write'command'access_type;
//KMM            .Read'command'SeqID == 0;
//KMM            .Read'command'UnitID == write_transaction.UnitID();
//KMM            .Read'command'SrcTag_Rsv == SrcTag;
//KMM            .Read'command'Compat == write_transaction.Write'command'Compat;
//KMM            // Make sure that each subsequent read in a string of byte reads uses the next nibble of byte mask.
//KMM            write_transaction.DataType() == Byte_Write => .Read'command'Mask_Count == write_transaction.Write'command'assoc_data.Byte_Write'command'Mask[(i+1)*4-1:i*4];
//KMM            write_transaction.DataType() == DoubleWord_Write => .Read'command'Mask_Count == write_transaction.Mask_Count();
//KMM            // Make sure that each subsequent read in a string of byte reads uses the next doubleword address..
//KMM            .Read'command'Addr == write_addr;
//KMM          };
          var tmp_write_transaction : Write'command csco_ht_packet;
          var tmp_assoc_data        : Byte_Write'command csco_ht_packet;
          tmp_write_transaction = write_transaction.as_a(Write'command csco_ht_packet);
          tmp_assoc_data = tmp_write_transaction.assoc_data.as_a(Byte_Write'command csco_ht_packet);
          gen transaction keeping {
            it is a Read'command csco_ht_packet(tmp)                                                and
            .endian_type == endian_type                                                         and
            .passing_rule == PassPW                                                             and
            (write_transaction.DataType() == Byte_Write => tmp.data_type == Byte)               and
            (write_transaction.DataType() == DoubleWord_Write => tmp.data_type == DoubleWord)   and
            tmp.BW_latency_req == tmp_write_transaction.BW_latency_req                          and
            tmp.access_type == tmp_write_transaction.access_type                                and
            tmp.SeqID == 0                                                                      and
            tmp.UnitID == write_transaction.UnitID()                                            and
            tmp.SrcTag_Rsv == SrcTag                                                            and
            tmp.Compat == tmp_write_transaction.Compat                                          and
            // Make sure that each subsequent read in a string of byte reads uses the next nibble of byte mask.
            (write_transaction.DataType() == Byte_Write => tmp.Mask_Count == tmp_assoc_data.Mask[(i+1)*4-1:i*4] )       and
            (write_transaction.DataType() == DoubleWord_Write => tmp.Mask_Count == write_transaction.Mask_Count() )     and
            // Make sure that each subsequent read in a string of byte reads uses the next doubleword address..
            tmp.Addr == write_addr;
          };
          memory_transaction_map_update(target_memory_transaction_map, transaction);
          if(num_read_transactions > 1) {
            // If there are multiple reads being associated with this write request, just schedule them, but don't
            // increment the transaction counter.
            if(parent_device.scoreboard.score_level == csco_ht_full_scoring) {
              parent_device.scoreboard.post(transaction);
            } else {
              parent_device.score_tx_pkt(transaction);
            };
            schedule_transaction(transaction);
          };
          num_read_transactions -= 1;
          i += 1;
        }until(num_read_transactions == 0);
      };
      Write:{
        // If a write is generated, pick a address and data to be written, then scoreboard them for self checking.
        var force_ordered_seq : bool;
        var forced_SeqID : uint (bits:4);
        var target_address_index : uint;
        // Select a random list index
        gen target_address_index keeping {
          it < target_address_list.size();
        };
        // Use that random number to select an address block
        current_target = target_address_list[target_address_index];
        // In order to preserve the correct ordering of sequential writes the following rule applies.
        // If a target address is in the same 64 byte block as an outstanding write, then the current write needs to be non posted
        // so that reads to the previous write location do not pass the current write.  (This can also be reduced to a
                                                                                         // 32 byte region restriction.)
        if(target_memory_transaction_map.key_exists(current_target.addr[39:5]) ||
           (!allow_32byte_region_ordering && current_target.addr[5:5] == 0 && target_memory_transaction_map.key_exists(current_target.addr[39:5] + 1))) {
             var transaction_map_element : csco_ht_memory_transaction_element;
             if(target_memory_transaction_map.key_exists(current_target.addr[39:5])) {
               transaction_map_element = target_memory_transaction_map.key(current_target.addr[39:5]);
             }else if(target_memory_transaction_map.key_exists(current_target.addr[39:5]+1)){
               transaction_map_element = target_memory_transaction_map.key(current_target.addr[39:5]+1);
             }else {
               dut_error("No relevant transaction map element\n");
             };

             // The proposed target address is in the same 64/32 byte block as an existing outstanding write request.
             // Force the new transaction to be in the non posted channel and have an ordered SeqID.
             if(parent_device.control_block.free_SrcTag(parent_device.control_block.current_UnitID[0]) == FALSE || never_overwrite_mem_until_read == TRUE
                || transaction_map_element.read_outstanding == TRUE || transaction_map_element.SeqID_outstanding == 0) {
                  // There are either no free SrcTag's, a read is outstanding or the previous write is in SeqID 0 so an ordered sequence cannot be created.
                  // Thus a transaction cannot be non posted.
                  // Go through the loop again.
                  return;
                }else {
                  force_ordered_seq = TRUE;
                  forced_SeqID = transaction_map_element.SeqID_outstanding;
                };
           };
        // Wait to delete the target address from the list until it is certain that it will be used.
        target_address_list.delete(target_address_index);

        endian_type = parent_device.decode_endian_type(current_target.addr);

//KMM        gen transaction keeping {
//KMM          .command == transaction_cmd;
//KMM          .endian_type == endian_type;
//KMM          force_ordered_seq == TRUE => .vChan == NonPosted;
//KMM          parent_device.device_function == Host_Bridge => .Write'command'UnitID == 0;
//KMM          parent_device.device_function == Tunnel_IO_Device or parent_device.device_function == EOC_IO_Device => .Write'command'UnitID == parent_device.control_block.current_UnitID[0];
//KMM          force_ordered_seq == TRUE => .Write'command'SeqID == forced_SeqID;
//KMM          // If there are no free SrcTag's then force a Posted Transaction.
//KMM          parent_device.control_block.free_SrcTag(parent_device.control_block.current_UnitID[0]) == FALSE => .vChan == Posted;
//KMM          .vChan == NonPosted => .Write'command'SrcTag_Rsv == parent_device.control_block.new_SrcTag(.Write'command'UnitID);
//KMM          .Write'command'Compat == 0; // By default always force the compat bit to 0.  FIX THIS LATER.
//KMM          .Write'command'data_type == Byte => .Write'command'Mask_Count <= current_target.num_doublewords+1;
//KMM          .Write'command'data_type == DoubleWord => .Write'command'Mask_Count <= current_target.num_doublewords;
//KMM          // If the 32 byte flag is set, force the DW writes to be in a 32 byte region.
//KMM          allow_32byte_region_ordering == TRUE && .Write'command'data_type == DoubleWord => .Write'command'Mask_Count < 4'b1000 - .Write'command'Addr[4:2];
//KMM          .Write'command'Addr == current_target.addr;
//KMM        };

        gen transaction keeping {
          it is a Write'command csco_ht_packet(tmp)                                         and
          .endian_type == endian_type                                                   and
          (force_ordered_seq == TRUE => .vChan == NonPosted)                            and
          (parent_device.device_function == Host_Bridge => tmp.UnitID == 0)             and
          (parent_device.device_function == Tunnel_IO_Device or parent_device.device_function == EOC_IO_Device
                        => tmp.UnitID == parent_device.control_block.current_UnitID[0]) and
          (force_ordered_seq == TRUE => tmp.SeqID == forced_SeqID)                      and
          // If there are no free SrcTag's then force a Posted Transaction.
          (parent_device.control_block.free_SrcTag(parent_device.control_block.current_UnitID[0]) == FALSE => .vChan == Posted) and
          (.vChan == NonPosted => tmp.SrcTag_Rsv == parent_device.control_block.new_SrcTag(tmp.UnitID) )                        and
          tmp.Compat == 0                                                               and  // By default always force the compat bit to 0.  FIX THIS LATER.
          (tmp.data_type == Byte => tmp.Mask_Count <= current_target.num_doublewords+1 ) and
          (tmp.data_type == DoubleWord => tmp.Mask_Count <= current_target.num_doublewords) and
          // If the 32 byte flag is set, force the DW writes to be in a 32 byte region.
          (allow_32byte_region_ordering == TRUE && tmp.data_type == DoubleWord => tmp.Mask_Count < 4'b1000 - tmp.Addr[4:2] ) and
          tmp.Addr == current_target.addr;
        };
        // Check to see if the transaction has occupied the entire address space of the address block.
        gen_replacement_address(transaction.Mask_Count(), current_target);
        outstanding_writes.add(transaction);
        memory_transaction_map_update(target_memory_transaction_map, transaction);
      };
      Read: {
        var order_after_Posted_Write : bool = FALSE;
        var order_after_NonPosted_Sequential_Write : bool = FALSE;

        // Never create a read to an unwritten location.
        if(outstanding_writes.is_empty() == TRUE) {
          // Just loop through again.
          return;
        };
        var write_index : uint;
        gen write_index keeping {
          it < outstanding_writes.size();
        };
        var write_req : csco_ht_packet;
        write_req = outstanding_writes[write_index];

        // In order to ensure that a read will read valid data one of two things must occur. If the original write is NonPosted
        // and has a non zero SeqID then the write must be in an ordered sequence.  If the original write is NonPosted and
        // has a zero SeqID then the read cannot occur until the TgtDone has been received.
        // If the original write is Posted, then the SeqID does not matter and the PassPW bit needs to be forced to FALSE.
        if(write_req.vChan == Posted) {
          order_after_Posted_Write = TRUE;
        }else {
          if(write_req.SeqID() == 0) {
            var response_token : uint (bits:10);
            response_token = %{write_req.UnitID(), write_req.SrcTag()};

            if(parent_device.scoreboard.transaction_pending(response_token)) {
              // Since a entry in the scoreboard still exists, it means that a TgtDone has not yet been received.
              // Therefore it is not yet safe to send a read, as it still might pass the write.
              return;
            };
          }else {
            order_after_NonPosted_Sequential_Write = TRUE;
          };
        };

        // Check to see if the write was a byte or a doubleword.  If it was a byte, then multiple byte reads will have to be
        // generated to check the write.  (One read for each doubleword of byte write, since byte reads are only 1 doubleword long.)
        // There must be at least 1 read transaction for each write.
        var num_read_transactions : uint = 1;
        var i : uint = 0;
        if(write_req.DataType() == Byte_Write) {
          num_read_transactions = write_req.Mask_Count();
        };
        repeat {
          var SrcTag : uint (bits:5) = parent_device.gen_tx_SrcTag(write_req.UnitID()); // Wait until a SrcTag is available.
          var write_addr: uint(bits: 40) = write_req.Address()+(i*4);

          endian_type = parent_device.decode_endian_type(write_addr);
//KMM          gen transaction keeping {
//KMM            .command == transaction_cmd;
//KMM            .endian_type == endian_type;
//KMM            order_after_Posted_Write == TRUE => .passing_rule == NoPassPW;
//KMM            write_req.DataType() == Byte_Write => .Read'command'data_type == Byte;
//KMM            write_req.DataType() == DoubleWord_Write => .Read'command'data_type == DoubleWord;
//KMM            .Read'command'BW_latency_req == write_req.Write'command'BW_latency_req;
//KMM            .Read'command'access_type == write_req.Write'command'access_type;
//KMM            order_after_NonPosted_Sequential_Write == TRUE => .Read'command'SeqID == write_req.SeqID();
//KMM            .Read'command'UnitID == write_req.UnitID();
//KMM            .Read'command'SrcTag_Rsv == SrcTag;
//KMM            .Read'command'Compat == write_req.Write'command'Compat;
//KMM            // Make sure that each subsequent read in a string of byte reads uses the next nibble of byte mask.
//KMM            write_req.DataType() == Byte_Write => .Read'command'Mask_Count == write_req.Write'command'assoc_data.Byte_Write'command'Mask[(i+1)*4-1:i*4];
//KMM            write_req.DataType() == DoubleWord_Write => .Read'command'Mask_Count == write_req.Mask_Count();
//KMM            // Make sure that each subsequent read in a string of byte reads uses the next doubleword address..
//KMM            .Read'command'Addr == write_addr;
//KMM          };
          var tmp_write_req : Write'command csco_ht_packet;
          tmp_write_req = write_req.as_a(Write'command csco_ht_packet);
          var tmp_assoc_data : Byte_Write'command csco_ht_packet;
          tmp_assoc_data = tmp_write_req.assoc_data.as_a(Byte_Write'command csco_ht_packet);
          gen transaction keeping {
            it is a Read'command csco_ht_packet(tmp)                                        and
            .endian_type == endian_type                                                 and
            (order_after_Posted_Write == TRUE => .passing_rule == NoPassPW)             and
            (write_req.DataType() == Byte_Write => tmp.data_type == Byte )              and
            (write_req.DataType() == DoubleWord_Write => tmp.data_type == DoubleWord )  and
            tmp.BW_latency_req == tmp_write_req.BW_latency_req                          and
            tmp.access_type == tmp_write_req.access_type                                and
            (order_after_NonPosted_Sequential_Write == TRUE => tmp.SeqID == write_req.SeqID() ) and
            tmp.UnitID == write_req.UnitID()                                            and
            tmp.SrcTag_Rsv == SrcTag                                                    and
            tmp.Compat == tmp_write_req.Compat                                          and
            // Make sure that each subsequent read in a string of byte reads uses the next nibble of byte mask.
            ( write_req.DataType() == Byte_Write => tmp.Mask_Count == tmp_assoc_data.Mask[(i+1)*4-1:i*4] )      and
            ( write_req.DataType() == DoubleWord_Write => tmp.Mask_Count == write_req.Mask_Count() )            and
            // Make sure that each subsequent read in a string of byte reads uses the next doubleword address..
            tmp.Addr == write_addr;
          };

          memory_transaction_map_update(target_memory_transaction_map, transaction);
          if(num_read_transactions > 1) {
            // If there are multiple reads being associated with this write request, just schedule them, but don't
            // increment the transaction counter.
            if(parent_device.scoreboard.score_level == csco_ht_full_scoring) {
              parent_device.scoreboard.post(transaction);
            } else {
              parent_device.score_tx_pkt(transaction);
            };
            schedule_transaction(transaction);
          };
          num_read_transactions -= 1;
          i += 1;
        }until(num_read_transactions == 0);

        outstanding_writes.delete(write_index);
      };
      Broadcast: {
//KMM        gen transaction keeping {
//KMM          .command == transaction_cmd;
//KMM          .Broadcast'command'UnitID == parent_device.control_block.current_UnitID[0];
//KMM          parent_device.control_block.free_SrcTag(parent_device.control_block.current_UnitID[0]) == FALSE => .vChan == Posted;
//KMM          .vChan == NonPosted => .Broadcast'command'SrcTag_Rsv == parent_device.control_block.new_SrcTag(.Broadcast'command'UnitID);
//KMM        };
        gen transaction keeping {
          it is a Broadcast'command csco_ht_packet(tmp)                                     and
          tmp.UnitID == parent_device.control_block.current_UnitID[0]                   and
          (parent_device.control_block.free_SrcTag(parent_device.control_block.current_UnitID[0]) == FALSE => .vChan == Posted) and
          (.vChan == NonPosted => tmp.SrcTag_Rsv == parent_device.control_block.new_SrcTag(tmp.UnitID) );
        };
      };
      Fence: {
        var write_transaction : csco_ht_packet;
        var fence_transaction : csco_ht_packet;
        // To generate a meaningful fence sequence a posted write must occur, then a fence, followed by a read to the
        // same location. The read should be allowed to pass all posted traffic, and sent immediately after the fence.

        var non_overlapping_address : bool;
        var target_address_index : uint;
        repeat {
          // Generate a Posted Write
          // Select a random list index
          gen target_address_index keeping {
            it < target_address_list.size();
          };
          // Use that random number to select an address block
          current_target = target_address_list[target_address_index];
          // Check to see if the address selected is in the same 64 byte block as an outstanding write transaction.
          // If it is, select another address. (Or 32 byte block if the flag is set.)
          if(target_memory_transaction_map.key_exists(current_target.addr[39:5]) ||
             (!allow_32byte_region_ordering && current_target.addr[5:5] == 0 && target_memory_transaction_map.key_exists(current_target.addr[39:5] + 1))) {
               non_overlapping_address = FALSE;
               wait cycle;
             }else {
               non_overlapping_address = TRUE;
             };
        }until(non_overlapping_address);
        target_address_list.delete(target_address_index);

        endian_type = parent_device.decode_endian_type(current_target.addr);

//KMM        gen write_transaction keeping {
//KMM          .command == Write;
//KMM          .endian_type == endian_type;
//KMM          .vChan == Posted;
//KMM          .Write'command'UnitID == parent_device.control_block.current_UnitID[0];
//KMM          .Write'command'SeqID == 0;
//KMM          .Write'command'Compat == 0; // By default always force the compat bit to 0.  FIX THIS LATER.
//KMM          .Write'command'data_type == Byte => .Write'command'Mask_Count <= current_target.num_doublewords+1;
//KMM          .Write'command'data_type == DoubleWord => .Write'command'Mask_Count <= current_target.num_doublewords;
//KMM          .Write'command'Addr == current_target.addr;
//KMM          // If the 32 byte flag is set, force the DW writes to be in a 32 byte region.
//KMM          allow_32byte_region_ordering == TRUE && .Write'command'data_type == DoubleWord => .Write'command'Mask_Count < 4'b1000 - .Write'command'Addr[4:2];
//KMM        };
        gen write_transaction keeping {
          it is a Write'command  csco_ht_packet(tmp)                                and
          .endian_type == endian_type                                           and
          .vChan == Posted                                                      and
          tmp.UnitID == parent_device.control_block.current_UnitID[0]           and
          tmp.SeqID == 0                                                        and
          tmp.Compat == 0                                                       and  // By default always force the compat bit to 0.  FIX THIS LATER.
          (tmp.data_type == Byte => tmp.Mask_Count <= current_target.num_doublewords+1)         and
          (tmp.data_type == DoubleWord => tmp.Mask_Count <= current_target.num_doublewords)     and
          (tmp.Addr == current_target.addr)                                                     and
          // If the 32 byte flag is set, force the DW writes to be in a 32 byte region.
          (allow_32byte_region_ordering == TRUE && tmp.data_type == DoubleWord => tmp.Mask_Count < 4'b1000 - tmp.Addr[4:2] );
        };
        // Don't scoreboard the transaction, just update the map and send it.
        memory_transaction_map_update(target_memory_transaction_map, write_transaction);
        // Check to see if the transaction has occupied the entire address space of the address block.
        gen_replacement_address(write_transaction.Mask_Count(), current_target);
        if(parent_device.scoreboard.score_level == csco_ht_full_scoring) {
          parent_device.scoreboard.post(write_transaction);
        } else {
          parent_device.score_tx_pkt(write_transaction);
        };
        schedule_transaction(write_transaction);

        // Generate a Fence
//KMM        gen fence_transaction keeping {
//KMM          .command == Fence;
//KMM          .passing_rule == NoPassPW;
//KMM          .Fence'command'SeqID == 0;
//KMM          .Fence'command'UnitID == parent_device.control_block.current_UnitID[0];
//KMM        };
        gen fence_transaction keeping {
          it is a Fence'command csco_ht_packet(tmp)                         and
          .passing_rule == NoPassPW                                     and
          tmp.SeqID == 0                                                and
          tmp.UnitID == parent_device.control_block.current_UnitID[0];
        };
        if(parent_device.scoreboard.score_level == csco_ht_full_scoring) {
          parent_device.scoreboard.post(fence_transaction);
        } else {
          parent_device.score_tx_pkt(fence_transaction);
        };
        schedule_transaction(fence_transaction);

        // Generate a Read, but make it unordered so that it will pass the write if the fence failed.
        // Force it to have a clear PassPW bit so that the fence can work.
        // Check to see if the write was a byte or a doubleword.  If it was a byte, then multiple byte reads will have to be
        // generated to check the write.  (One read for each doubleword of byte write, since byte reads are only 1 doubleword long.)
        // There must be at least 1 read transaction for each write.
        var num_read_transactions : uint = 1;
        var i : uint = 0;
        if(write_transaction.DataType() == Byte_Write) {
          num_read_transactions = write_transaction.Mask_Count();
        };
        repeat {
          var SrcTag : uint (bits:5) = parent_device.gen_tx_SrcTag(write_transaction.UnitID()); // Wait until a SrcTag is available.
          var write_addr: uint(bits: 40) = write_transaction.Address() + (i*4);

          endian_type = parent_device.decode_endian_type(write_addr);
//KMM          gen transaction keeping {
//KMM            .command == Read;
//KMM            .endian_type == endian_type;
//KMM            .passing_rule == NoPassPW;
//KMM            write_transaction.DataType() == Byte_Write => .Read'command'data_type == Byte;
//KMM            write_transaction.DataType() == DoubleWord_Write => .Read'command'data_type == DoubleWord;
//KMM            .Read'command'BW_latency_req == write_transaction.Write'command'BW_latency_req;
//KMM            .Read'command'access_type == write_transaction.Write'command'access_type;
//KMM            .Read'command'SeqID == 0;
//KMM            .Read'command'UnitID == write_transaction.UnitID();
//KMM            .Read'command'SrcTag_Rsv == SrcTag;
//KMM            .Read'command'Compat == write_transaction.Write'command'Compat;
//KMM            // Make sure that each subsequent read in a string of byte reads uses the next nibble of byte mask.
//KMM            write_transaction.DataType() == Byte_Write => .Read'command'Mask_Count == write_transaction.Write'command'assoc_data.Byte_Write'command'Mask[(i+1)*4-1:i*4];
//KMM            write_transaction.DataType() == DoubleWord_Write => .Read'command'Mask_Count == write_transaction.Mask_Count();
//KMM            // Make sure that each subsequent read in a string of byte reads uses the next doubleword address..
//KMM            .Read'command'Addr == write_addr;
//KMM          };

          var tmp_write_transaction : Write'command csco_ht_packet;
          tmp_write_transaction = write_transaction.as_a(Write'command csco_ht_packet);
          var tmp_assoc_data : Byte_Write'command csco_ht_packet;
          tmp_assoc_data = tmp_write_transaction.assoc_data.as_a(Byte_Write'command csco_ht_packet);

          gen transaction keeping {
            it is a Read'command csco_ht_packet(tmp)                                        and
            .endian_type == endian_type                                                 and
            .passing_rule == NoPassPW                                                   and
            (write_transaction.DataType() == Byte_Write => tmp.data_type == Byte )      and
            (write_transaction.DataType() == DoubleWord_Write => tmp.data_type == DoubleWord )  and
            tmp.BW_latency_req == tmp_write_transaction.BW_latency_req                          and
            tmp.access_type == tmp_write_transaction.access_type                                and
            tmp.SeqID == 0                                                                      and
            tmp.UnitID == write_transaction.UnitID()                                            and
            tmp.SrcTag_Rsv == SrcTag                                                            and
            tmp.Compat == tmp_write_transaction.Compat                                          and
            // Make sure that each subsequent read in a string of byte reads uses the next nibble of byte mask.
            (write_transaction.DataType() == Byte_Write => tmp.Mask_Count == tmp_assoc_data.Mask[(i+1)*4-1:i*4])        and
            (write_transaction.DataType() == DoubleWord_Write => tmp.Mask_Count == write_transaction.Mask_Count() )     and
            // Make sure that each subsequent read in a string of byte reads uses the next doubleword address..
            tmp.Addr == write_addr;
          };
          memory_transaction_map_update(target_memory_transaction_map, transaction);
          if(num_read_transactions > 1) {
            // If there are multiple reads being associated with this write request, just schedule them, but don't
            // increment the transaction counter.
            if(parent_device.scoreboard.score_level == csco_ht_full_scoring) {
              parent_device.scoreboard.post(transaction);
            } else {
              parent_device.score_tx_pkt(transaction);
            };
            schedule_transaction(transaction);
          };
          num_read_transactions -= 1;
          i += 1;
        }until(num_read_transactions == 0);
      };
      default: {
        dut_error("Illegal random transaction selected.\n");
      };
    };
    if(transaction != NULL) {
      if(parent_device.scoreboard.score_level == csco_ht_full_scoring) {
        parent_device.scoreboard.post(transaction);
      } else {
        parent_device.score_tx_pkt(transaction);
      };
      schedule_transaction(transaction);
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  // After a given address block is used for the INTERMIX_OVERLAY scenario
  // check to see if a subset or superset of those address should be added back into
  // the target_address_list to allow for overlap.
  ////////////////////////////////////////////////////////////////////////////////
  gen_replacement_address(transaction_size : uint (bits:4), current_target : csco_ht_scenario_element) is also {
    var reuse_address_block : bool;
    gen reuse_address_block;
    if(reuse_address_block == TRUE) {
      var new_target : csco_ht_scenario_element;
      var max_reused_addr : uint (bits:40);
      var current_address_region : csco_ht_address_map;

      if (random_Address_aligned != TRUE) then {
        max_reused_addr = current_target.addr+((current_target.num_doublewords+1)<<2);
        gen new_target keeping {
          it.addr >= current_target.addr;
          max_reused_addr <= addr_map.top_addr => it.addr <= max_reused_addr;
          max_reused_addr > addr_map.top_addr => it.addr <= addr_map.top_addr;
          //it.addr <= max_reused_addr;
          it.addr[1:0] == 2'b00;
          it.tmp1 <= ((addr_map.top_addr+1-it.addr)>>2)-1;
          it.tmp1 <= 0 => it.num_doublewords == 0;
          it.tmp1 >  0 => it.num_doublewords == it.tmp1;
        }
      } else {
         max_reused_addr = current_target.addr+((current_target.num_doublewords+1)<<3);
         gen new_target keeping {
          it.addr >= current_target.addr;
          max_reused_addr <= addr_map.top_addr => it.addr <= max_reused_addr;
          max_reused_addr > addr_map.top_addr => it.addr <= addr_map.top_addr;
          //it.addr <= max_reused_addr;
          it.addr[2:0] == 3'b000;
          it.tmp1 <= ((addr_map.top_addr+1-it.addr)>>3)-1;
          it.tmp1 <= 0 => it.num_doublewords == 0;
          it.tmp1 >  0 => it.num_doublewords == it.tmp1;
        }
      };
      target_address_list.push(new_target); // Push the new address block back onto the list.
      var AccessType_: csco_ht_mem_access_type;
      var BWLatencyReq_: csco_ht_BW_latency_req;
      var vChan_: csco_ht_virtual_channel;
      num_transactions += 1;
      gen AccessType_ keeping {it == NonCoherent;};
      listOfAccessType.add(AccessType_);
      gen BWLatencyReq_ keeping {it == Normal;};
      listOfBWLatencyReq.add(BWLatencyReq_);
      gen vChan_ keeping {it == Posted;};
      listOfVChanType.add(vChan_);
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
  // memory_transaction_map_update() will take an HT request and create or delete an entry in the memory map.  This map is
  // used to record the 32 or 64 byte blocks that all outstanding transactions are accessing.  This allows reads and writes
  // to be properly ordered if they overlap in memory space so that access results are predicatable.
  ////////////////////////////////////////////////////////////////////////////////
  memory_transaction_map_update(transaction_map : list (key:addr) of csco_ht_memory_transaction_element, packet : csco_ht_packet) is {
    var addr : uint (bits:40);
    var transaction_length : uint (bits:4);
    var transaction_element : csco_ht_memory_transaction_element;

    case(packet.command()) {
      Write: {
        addr = packet.Address();
        transaction_length = packet.Mask_Count();

        if(!transaction_map.key_exists(addr[39:5])) {
          gen transaction_element keeping {
            .addr == addr[39:5];
            .SeqID_outstanding == packet.SeqID();
          };
          transaction_map.add(transaction_element);
        };

        if(packet.DataType() == Byte_Write) {
          transaction_length -= 1;
        };
        var temp_addr : uint (bits:40) = addr + (transaction_length << 2);
        if(!transaction_map.key_exists(temp_addr[39:5])) {
          gen transaction_element keeping {
            .addr == temp_addr[39:5];
            .SeqID_outstanding == packet.SeqID();
          };
          transaction_map.add(transaction_element);
        };
      };
      Read: {
        addr = packet.Address();
        transaction_length = packet.Mask_Count();

        if(transaction_map.key_exists(addr[39:5])) {
          transaction_map.key(addr[39:5]).read_outstanding = TRUE;
        }else {
          dut_error("A read is being created for a memory location that has not been written.\n");
        };

//KMM        if(packet.Read'command'data_type == DoubleWord) {
        if(packet.as_a(Read'command csco_ht_packet).data_type == DoubleWord) {
          var temp_addr : uint (bits:40) = addr + (transaction_length << 2);
          if(transaction_map.key_exists(temp_addr[39:5])) {
            transaction_map.key(temp_addr[39:5]).read_outstanding = TRUE;
          }else {
            dut_error("A read is being created for a memory location that has not been written.\n");
          };
        };
      };
      RdResponse: {
        var response_token : uint (bits:10);
        if(packet.BridgeBit() == 1) {
          // This is a response from a Host Bridge so the UnitID is the source of the request.
          // Hence the UnitID is used to compute the response token.
          response_token = %{packet.UnitID(), packet.SrcTag()};
        }else {
          // Otherwise this is a response from an IO Device and the UnitID is the target of the request.
          // Thus the UnitID of 0 should be used to compute the response token to align with the requestors UnitID.
          response_token = %{5'b00000, packet.SrcTag()};
        };
        addr = parent_device.scoreboard.pending_transactions.key(response_token).request.Address();
        transaction_length = parent_device.scoreboard.pending_transactions.key(response_token).request.Mask_Count();

        if(transaction_map.key_exists(addr[39:5])) {
          transaction_map.delete(transaction_map.key_index(addr[39:5]));
        };

        var temp_addr : uint (bits:40) = addr + (transaction_length << 2);
        if(transaction_map.key_exists(temp_addr[39:5])) {
          transaction_map.delete(transaction_map.key_index(temp_addr[39:5]));
        };
      };
      default: {
        dut_error("A packet other than a read or a write was attempting to update the transaction map.\n");
      };
    };
  };
  ////////////////////////////////////////////////////////////////////////////////
};
////////////////////////////////////////////////////////////////////////////////
// WRITE_IN_ORDER is a scenario that performs in order write request, of random size,
// to a targeted_address range.
////////////////////////////////////////////////////////////////////////////////
extend WRITE_IN_ORDER'kind csco_ht_scenario {
  SeqID : uint (bits:4);
  keep SeqID != 0;  // Select a SeqID so that all traffic from this scenario is strongly ordered.

  generate_transactions()@parent_device.csco_ht_device_clock is {
    var current_target : csco_ht_scenario_element;
    var transaction : csco_ht_packet;
    var endian_type: csco_ht_endian_type;

    if(target_address_list.size() > 0) {
      current_target = target_address_list.pop0();
      endian_type = parent_device.decode_endian_type(current_target.addr);

//KMM      gen transaction keeping {
//KMM        .command == Write;
//KMM        .endian_type == endian_type;
//KMM        // In order to make sure that these are sequenced have them use the same SeqID and can't pass each other in different VC's.
//KMM        .vChan == NonPosted => .passing_rule == NoPassPW;
//KMM        parent_device.device_function == Host_Bridge => .Write'command'UnitID == 0;
//KMM        parent_device.device_function == Tunnel_IO_Device or parent_device.device_function == EOC_IO_Device => .Write'command'UnitID == parent_device.control_block.current_UnitID[0];
//KMM        .Write'command'SeqID == SeqID;
//KMM        // If there are no free SrcTag's then force a Posted Transaction.
//KMM        parent_device.control_block.free_SrcTag(parent_device.control_block.current_UnitID[0]) == FALSE => .vChan == Posted;
//KMM        .vChan == NonPosted => .Write'command'SrcTag_Rsv == parent_device.control_block.new_SrcTag(.Write'command'UnitID);
//KMM        .Write'command'Compat == 0; // By default always force the compat bit to 0.  FIX THIS LATER.
//KMM        .Write'command'data_type == Byte => .Write'command'Mask_Count <= current_target.num_doublewords+1;
//KMM        .Write'command'data_type == DoubleWord => .Write'command'Mask_Count <= current_target.num_doublewords;
//KMM        .Write'command'Addr == current_target.addr;
//KMM      };

      gen transaction keeping {
        it is a Write'command csco_ht_packet(tmp)                                                                   and
        .endian_type == endian_type                                                                             and
        // In order to make sure that these are sequenced have them use the same SeqID and can't pass each other in different VC's.
        (.vChan == NonPosted => .passing_rule == NoPassPW)                                                      and
        (parent_device.device_function == Host_Bridge => tmp.UnitID == 0)                                       and
        (parent_device.device_function == Tunnel_IO_Device or parent_device.device_function == EOC_IO_Device =>
                                tmp.UnitID == parent_device.control_block.current_UnitID[0])                    and
        tmp.SeqID == SeqID                                                                                      and
        // If there are no free SrcTag's then force a Posted Transaction.
        (parent_device.control_block.free_SrcTag(parent_device.control_block.current_UnitID[0]) == FALSE => .vChan == Posted ) and
        (.vChan == NonPosted => tmp.SrcTag_Rsv == parent_device.control_block.new_SrcTag(tmp.UnitID) )          and
        tmp.Compat == 0                                                 and  // By default always force the compat bit to 0.  FIX THIS LATER.
        (tmp.data_type == Byte => tmp.Mask_Count <= current_target.num_doublewords+1 )                          and
        (tmp.data_type == DoubleWord => tmp.Mask_Count <= current_target.num_doublewords)                       and
        tmp.Addr == current_target.addr;
      };

      // Check to see if the transaction has occupied the entire address space of the address block.
      gen_replacement_address(transaction.Mask_Count(), current_target);
      if(parent_device.scoreboard.score_level == csco_ht_full_scoring) {
        parent_device.scoreboard.post(transaction);
      } else {
        parent_device.score_tx_pkt(transaction);
      };
      schedule_transaction(transaction);
    }else {
      done = TRUE;
    };
  };
};
////////////////////////////////////////////////////////////////////////////////
// READ_WRITE_IN_ORDER is a scenario that performs in order and alternating write and read request,
// of random size, to a targeted_address range.
////////////////////////////////////////////////////////////////////////////////
extend READ_WRITE_IN_ORDER'kind csco_ht_scenario {
  next_transaction : csco_ht_packet_encoding;
  keep next_transaction == Write; // Start off with a write packet.
  SeqID : uint (bits:4);
  keep SeqID != 0;  // Select a SeqID so that all traffic from this scenario is strongly ordered.
  !outstanding_writes : list of csco_ht_packet;

  generate_transactions()@parent_device.csco_ht_device_clock is {
    var current_target : csco_ht_scenario_element;
    var transaction : csco_ht_packet;
    var endian_type: csco_ht_endian_type;
    var target_address_index : uint;

    if(target_address_list.size() == 0 && outstanding_writes.size() == 0) {
      done = TRUE;
    }else {
      case(next_transaction) {
        Write: {
          if(target_address_list.is_empty() == TRUE) {
            dut_error("The target address list should never be empty when this scenario gets here.");
          };
          current_target = target_address_list.pop0();
          endian_type = parent_device.decode_endian_type(current_target.addr);
//KMM          gen transaction keeping {
//KMM            .command == Write;
//KMM            .endian_type == endian_type;
//KMM            // In order to make sure that these are sequenced have them use the same SeqID and can't pass each other in different VC's.
//KMM            .vChan == NonPosted => .passing_rule == NoPassPW;
//KMM            parent_device.device_function == Host_Bridge => .Write'command'UnitID == 0;
//KMM            parent_device.device_function == Tunnel_IO_Device or parent_device.device_function == EOC_IO_Device => .Write'command'UnitID == parent_device.control_block.current_UnitID[0];
//KMM            .Write'command'SeqID == SeqID;
//KMM            // If there are no free SrcTag's then force a Posted Transaction.
//KMM            parent_device.control_block.free_SrcTag(parent_device.control_block.current_UnitID[0]) == FALSE => .vChan == Posted;
//KMM            .vChan == NonPosted => .Write'command'SrcTag_Rsv == parent_device.control_block.new_SrcTag(.Write'command'UnitID);
//KMM            .Write'command'Compat == 0; // By default always force the compat bit to 0.  FIX THIS LATER.
//KMM            .Write'command'data_type == Byte => .Write'command'Mask_Count <= current_target.num_doublewords+1;
//KMM            .Write'command'data_type == DoubleWord => .Write'command'Mask_Count <= current_target.num_doublewords;
//KMM            .Write'command'Addr == current_target.addr;
//KMM          };

          gen transaction keeping {
            it is a Write'command csco_ht_packet(tmp)                                                       and
            .endian_type == endian_type                                                                 and
            // In order to make sure that these are sequenced have them use the same SeqID and can't pass each other in different VC's.
            (.vChan == NonPosted => .passing_rule == NoPassPW)                                          and
            (parent_device.device_function == Host_Bridge => tmp.UnitID == 0)                           and
            (parent_device.device_function == Tunnel_IO_Device or parent_device.device_function == EOC_IO_Device =>
                                tmp.UnitID == parent_device.control_block.current_UnitID[0] )           and
            tmp.SeqID == SeqID                                                                          and
            // If there are no free SrcTag's then force a Posted Transaction.
            (parent_device.control_block.free_SrcTag(parent_device.control_block.current_UnitID[0]) == FALSE => .vChan == Posted) and
            (.vChan == NonPosted => tmp.SrcTag_Rsv == parent_device.control_block.new_SrcTag(tmp.UnitID) )                      and
            tmp.Compat == 0                     and // By default always force the compat bit to 0.  FIX THIS LATER.
            (tmp.data_type == Byte => tmp.Mask_Count <= current_target.num_doublewords+1) and
            (tmp.data_type == DoubleWord => tmp.Mask_Count <= current_target.num_doublewords )                  and
            tmp.Addr == current_target.addr;
          };


          // Check to see if the transaction has occupied the entire address space of the address block.
          gen_replacement_address(transaction.Mask_Count(), current_target);
          outstanding_writes.add(transaction);

          if(parent_device.scoreboard.score_level == csco_ht_full_scoring) {
            parent_device.scoreboard.post(transaction);
          } else {
            parent_device.score_tx_pkt(transaction);
          };
          schedule_transaction(transaction);
          // After the write, follow up with a read.
          next_transaction = Read;
        };
        Read: {
          var order_after_Posted_Write : bool = FALSE;
          var write_req : csco_ht_packet;
          // Never create a read to an unwritten location.
          if(outstanding_writes.is_empty() == TRUE) {
            dut_error("The outstanding writes list should never be empty when this scenario gets here.");
          };
          write_req = outstanding_writes.pop();

          // In order to ensure that a read will read valid data one of two things must occur. If the original write is NonPosted
          // and has a non zero SeqID then the write must be in an ordered sequence.  If the original write is NonPosted and
          // has a zero SeqID then the read cannot occur until the TgtDone has been received.
          // If the original write is Posted, then the SeqID does not matter and the PassPW bit needs to be forced to FALSE.
          if(write_req.vChan == Posted) {
            order_after_Posted_Write = TRUE;
          };

          // Check to see if the write was a byte or a doubleword.  If it was a byte, then multiple byte reads will have to be
          // generated to check the write.  (One read for each doubleword of byte write, since byte reads are only 1 doubleword long.)
          // There must be at least 1 read transaction for each write.
          var num_read_transactions : uint = 1;
          var i : uint = 0;
          if(write_req.DataType() == Byte_Write) {
            num_read_transactions = write_req.Mask_Count();
          };
          repeat {
            var SrcTag : uint (bits:5) = parent_device.gen_tx_SrcTag(write_req.UnitID()); // Wait until a SrcTag is available.
            var write_addr: uint(bits: 40) = write_req.Address()+(i*4);

            endian_type = parent_device.decode_endian_type(write_addr);
//KMM            gen transaction keeping {
//KMM              .command == Read;
//KMM              .endian_type == endian_type;
//KMM              order_after_Posted_Write == TRUE => .passing_rule == NoPassPW;
//KMM              write_req.DataType() == Byte_Write => .Read'command'data_type == Byte;
//KMM              write_req.DataType() == DoubleWord_Write => .Read'command'data_type == DoubleWord;
//KMM              .Read'command'BW_latency_req == write_req.Write'command'BW_latency_req;
//KMM              .Read'command'access_type == write_req.Write'command'access_type;
//KMM              .Read'command'SeqID == SeqID;
//KMM              .Read'command'UnitID == write_req.UnitID();
//KMM              .Read'command'SrcTag_Rsv == SrcTag;
//KMM              .Read'command'Compat == write_req.Write'command'Compat;
//KMM              // Make sure that each subsequent read in a string of byte reads uses the next nibble of byte mask.
//KMM              write_req.DataType() == Byte_Write => .Read'command'Mask_Count == write_req.Write'command'assoc_data.Byte_Write'command'Mask[(i+1)*4-1:i*4];
//KMM              write_req.DataType() == DoubleWord_Write => .Read'command'Mask_Count == write_req.Mask_Count();
//KMM              // Make sure that each subsequent read in a string of byte reads uses the next doubleword address..
//KMM              .Read'command'Addr == write_addr;
//KMM            };

            var tmp_write_req : Write'command csco_ht_packet;
            tmp_write_req = write_req.as_a(Write'command csco_ht_packet);
            var tmp_assoc_data : Byte_Write'command csco_ht_packet;
            tmp_assoc_data = tmp_write_req.assoc_data.as_a(Byte_Write'command csco_ht_packet);
            gen transaction keeping {
              it is a Read'command  csco_ht_packet(tmp)                                     and
              .endian_type == endian_type                                               and
              (order_after_Posted_Write == TRUE => .passing_rule == NoPassPW)           and
              (write_req.DataType() == Byte_Write => tmp.data_type == Byte)             and
              (write_req.DataType() == DoubleWord_Write => tmp.data_type == DoubleWord) and
              tmp.BW_latency_req == tmp_write_req.BW_latency_req                        and
              tmp.access_type == tmp_write_req.access_type                              and
              tmp.SeqID == SeqID                                                        and
              tmp.UnitID == write_req.UnitID()                                          and
              tmp.SrcTag_Rsv == SrcTag                                                  and
              tmp.Compat == tmp_write_req.Compat                                        and
              // Make sure that each subsequent read in a string of byte reads uses the next nibble of byte mask.
              (write_req.DataType() == Byte_Write => tmp.Mask_Count == tmp_assoc_data.Mask[(i+1)*4-1:i*4])      and
              (write_req.DataType() == DoubleWord_Write => tmp.Mask_Count == write_req.Mask_Count() )           and
              // Make sure that each subsequent read in a string of byte reads uses the next doubleword address..
              tmp.Addr == write_addr;
            };

            if(parent_device.scoreboard.score_level == csco_ht_full_scoring) {
              parent_device.scoreboard.post(transaction);
            } else {
              parent_device.score_tx_pkt(transaction);
            };
            schedule_transaction(transaction);
            num_read_transactions -= 1;
            i += 1;
          }until(num_read_transactions == 0);
          // After the read, follow up with a write.
          next_transaction = Write;
        };
        default: {
          dut_error("Illegal packet type selected for READ_WRITE_IN_ORDER scenario.");
        };
      };
    };
  };
};
////////////////////////////////////////////////////////////////////////////////
// BATCH_WRITE_BATCH_READ is a scenario that performs write then read requests,
// of random size, to a targeted_address range.
////////////////////////////////////////////////////////////////////////////////

extend csco_ht_scenario_kind : [BATCH_WRITE_BATCH_READ];

extend BATCH_WRITE_BATCH_READ'kind csco_ht_scenario {
  next_transaction : csco_ht_packet_encoding;
  keep next_transaction == Write; // Start off with a write packet.
  SeqID : uint (bits:4);
  keep SeqID != 0;  // Select a SeqID so that all traffic from this scenario is strongly ordered.
  !outstanding_writes : list of csco_ht_packet;

  generate_transactions()@parent_device.csco_ht_device_clock is {
    var current_target : csco_ht_scenario_element;
    var transaction : csco_ht_packet;
    var endian_type: csco_ht_endian_type;
    var target_address_index : uint;
    print target_address_list.size();
    print outstanding_writes.size();
    if(target_address_list.size() == 0 && outstanding_writes.size() == 0) {
      done = TRUE;
    }else {
      //randomIndex = get_next_uid();
      case(next_transaction) {
        Write: {
          if(target_address_list.is_empty() == TRUE) {
            dut_error("The target address list should never be empty when this scenario gets here.");
          };
          current_target = target_address_list.pop();

          endian_type = parent_device.decode_endian_type(current_target.addr);

          //print target_address_list;
          randomIndex = get_next_uid();
          gen transaction keeping {
            .command == Write;
            .endian_type == endian_type;
            // In order to make sure that these are sequenced have them use the same SeqID and can't pass each other in different VC's.
            .vChan == NonPosted => .passing_rule == NoPassPW;
            parent_device.device_function == Host_Bridge => .Write'command'UnitID == 0;
            parent_device.device_function == Tunnel_IO_Device or parent_device.device_function == EOC_IO_Device => .Write'command'UnitID == parent_device.control_block.current_UnitID[0];
            .Write'command'SeqID == SeqID;
            // If there are no free SrcTag's then force a Posted Transaction.
            parent_device.control_block.free_SrcTag(parent_device.control_block.current_UnitID[0]) == FALSE => .vChan == Posted;
            .vChan == listOfVChanType[randomIndex];
            .vChan == NonPosted => .Write'command'SrcTag_Rsv == parent_device.control_block.new_SrcTag(.Write'command'UnitID);
            .Write'command'Compat == 0; // By default always force the compat bit to 0.  FIX THIS LATER.
            parent_device.ReadWrite_double==TRUE => .Write'command'data_type == DoubleWord;
            .Write'command'data_type == Byte => .Write'command'Mask_Count <= current_target.num_doublewords+1;
            .Write'command'data_type == DoubleWord => .Write'command'Mask_Count <= current_target.num_doublewords;
            .Write'command'Addr == current_target.addr;
            .Write'command'access_type == listOfAccessType[randomIndex];
            .Write'command'BW_latency_req == listOfBWLatencyReq[randomIndex];
          };

          // Check to see if the transaction has occupied the entire address space of the address block.
          gen_replacement_address(transaction.Mask_Count(), current_target);
          outstanding_writes.add(transaction);

          if(parent_device.scoreboard.score_level == csco_ht_full_scoring) {
            parent_device.scoreboard.post(transaction);
          } else {
            parent_device.score_tx_pkt(transaction);
          };
          schedule_transaction(transaction);
          // After the write, follow up with a read.
          if target_address_list.size() != 0 {
            next_transaction = Write;
          } else {
            next_transaction = Read;
          };
        };
        Read: {
          var order_after_Posted_Write : bool = FALSE;
          var write_req : csco_ht_packet;
          // Never create a read to an unwritten location.
          if(outstanding_writes.is_empty() == TRUE) {
            dut_error("The outstanding writes list should never be empty when this scenario gets here.");
          };
          write_req = outstanding_writes.pop();

          // In order to ensure that a read will read valid data one of two things must occur. If the original write is NonPosted
          // and has a non zero SeqID then the write must be in an ordered sequence.  If the original write is NonPosted and
          // has a zero SeqID then the read cannot occur until the TgtDone has been received.
          // If the original write is Posted, then the SeqID does not matter and the PassPW bit needs to be forced to FALSE.
          if(write_req.vChan == Posted) {
            order_after_Posted_Write = TRUE;
          };

          // Check to see if the write was a byte or a doubleword.  If it was a byte, then multiple byte reads will have to be
          // generated to check the write.  (One read for each doubleword of byte write, since byte reads are only 1 doubleword long.)
          // There must be at least 1 read transaction for each write.
          var num_read_transactions : uint = 1;
          var i : uint = 0;
          if(write_req.DataType() == Byte_Write) {
            num_read_transactions = write_req.Mask_Count();
          };
          repeat {
            var SrcTag : uint (bits:5) = parent_device.gen_tx_SrcTag(write_req.UnitID()); // Wait until a SrcTag is available.
            var write_addr: uint(bits: 40) = write_req.Address()+(i*4);

            endian_type = parent_device.decode_endian_type(write_addr);
            gen transaction keeping {
              .command == Read;
              .endian_type == endian_type;
              order_after_Posted_Write == TRUE => .passing_rule == NoPassPW;
              write_req.DataType() == Byte_Write => .Read'command'data_type == Byte;
              write_req.DataType() == DoubleWord_Write => .Read'command'data_type == DoubleWord;
              .Read'command'BW_latency_req == write_req.Write'command'BW_latency_req;
              .Read'command'access_type == write_req.Write'command'access_type;
              .Read'command'SeqID == SeqID;
              .Read'command'UnitID == write_req.UnitID();
              .Read'command'SrcTag_Rsv == SrcTag;
              .Read'command'Compat == write_req.Write'command'Compat;
              // Make sure that each subsequent read in a string of byte reads uses the next nibble of byte mask.
              write_req.DataType() == Byte_Write => .Read'command'Mask_Count == write_req.Write'command'assoc_data.Byte_Write'command'Mask[(i+1)*4-1:i*4];
              write_req.DataType() == DoubleWord_Write => .Read'command'Mask_Count == write_req.Mask_Count();
              // Make sure that each subsequent read in a string of byte reads uses the next doubleword address..
              .Read'command'Addr == write_addr;
            };
            if(parent_device.scoreboard.score_level == csco_ht_full_scoring) {
              parent_device.scoreboard.post(transaction);
            } else {
              parent_device.score_tx_pkt(transaction);
            };
            schedule_transaction(transaction);
            num_read_transactions -= 1;
            i += 1;
          }until(num_read_transactions == 0);
          // After the read, follow up with a write.
          // next_transaction = Write;
        };
        default: {
          dut_error("Illegal packet type selected for BATCH_WRITE_BATCH_READ scenario.");
        };
      };
    };
  };
};
////////////////////////////////////////////////////////////////////////////////
// INTERRUPT is a scenario that performs in interrupts targeted towards the Host Bridge,
////////////////////////////////////////////////////////////////////////////////
extend INTERRUPT'kind csco_ht_scenario {
  keep addr_map.region == Interrupts;
  SeqID : uint (bits:4);
  keep SeqID != 0;  // Select a SeqID so that all traffic from this scenario is strongly ordered.

  generate_transactions()@parent_device.csco_ht_device_clock is {
    var current_target : csco_ht_scenario_element;
    var transaction : csco_ht_packet;
    if(target_address_list.size() == 0) {
      done = TRUE;
    }else {
      if(target_address_list.is_empty() == TRUE) {
        dut_error("The target address list should never be empty when this scenario gets here.");
      };
      current_target = target_address_list.pop0();

//KMM      gen transaction keeping {
//KMM        .command == Write;
//KMM        .Write'command'interrupt == TRUE;
//KMM        parent_device.device_function == Host_Bridge => .Write'command'UnitID == 0;
//KMM        parent_device.device_function == Tunnel_IO_Device or parent_device.device_function == EOC_IO_Device => .Write'command'UnitID == parent_device.control_block.current_UnitID[0];
//KMM        .Write'command'SeqID == SeqID;
//KMM        // Derive the vector and destination from the generated address
//KMM        .Write'command'IntrDest == current_target.addr[15:8];
//KMM        .Write'command'Vector == current_target.addr[23:16];
//KMM      };

      gen transaction keeping {
        it is a Write'command csco_ht_packet(tmp)                                   and
        tmp.interrupt == TRUE                                                   and
        (parent_device.device_function == Host_Bridge => tmp.UnitID == 0)       and
        (parent_device.device_function == Tunnel_IO_Device or parent_device.device_function == EOC_IO_Device =>
                tmp.UnitID == parent_device.control_block.current_UnitID[0])    and
        tmp.SeqID == SeqID                                                      and
        // Derive the vector and destination from the generated address
        tmp.IntrDest == current_target.addr[15:8]                               and
        tmp.Vector == current_target.addr[23:16];
      };

      if(parent_device.scoreboard.score_level == csco_ht_full_scoring) {
        parent_device.scoreboard.post(transaction);
      } else {
        parent_device.score_tx_pkt(transaction);
      };
      schedule_transaction(transaction);
    };
  };
};
////////////////////////////////////////////////////////////////////////////////
// this scenario reads and tests POR values of all registers for a HT device in
// the chain from the Host Bridge
////////////////////////////////////////////////////////////////////////////////
extend READ_POR_REGS'kind csco_ht_scenario {
  keep target_address_gen_complete == TRUE; // Don't self generate target addresses
  num_devices : uint; // number of devices in the chain
  // soft constraint, assume at least 1 device in the chain, size of chain is not known yet :P
  keep soft num_devices == 1;

  generate_transactions()@parent_device.csco_ht_device_clock is {
    var delay          : uint (bits:4);          // random delay to wait between transactions
    var device_profile : csco_ht_device_profile;     // pointer to the device profile
    var target_regs    : list of csco_ht_basic_reg;  // list of regs to test
    var ret            : csco_ht_error_code;         // return value from read

    for i from 0 to (num_devices - 1) {
      // get pointer to the device profile and the list of target regs
//KMM      device_profile = parent_device.Host_Bridge'device_function'chain_topology[i];
      device_profile = parent_device.as_a(Host_Bridge'device_function csco_ht_device ).chain_topology[i];
      target_regs = device_profile.get_device_regs();

      // Iterate through all of the registers for the device
      for each (reg) in target_regs do {
        // generate random delay between reads
        gen delay keeping {
          it > 0;
        };

        ret = reg.read_ht(channel);
        wait [delay];
      };
    };

    done = TRUE;
  };
};
////////////////////////////////////////////////////////////////////////////////
// this scenario writes a '1' to every bit in all the registers for a HT device
// in the chain from the Host Bridge. only bits that are not masked off in the
// csco_ht_basic_reg func_write_mask will be written. the write will be followed by
// a read to the register as a check.
////////////////////////////////////////////////////////////////////////////////
extend WALK_ONES_REGS'kind csco_ht_scenario {
  keep target_address_gen_complete == TRUE; // Don't self generate target addresses
  num_devices : uint; // number of devices in the chain
  // soft constraint, assume at least 1 device in the chain
  keep soft num_devices == 1;

  generate_transactions()@parent_device.csco_ht_device_clock is {
    var delay          : uint (bits:4);          // random delay to wait between transactions
    var device_profile : csco_ht_device_profile;     // pointer to the device profile
    var target_regs    : list of csco_ht_basic_reg;  // list of regs to test
    var ret            : csco_ht_error_code;         // return value from read

    for i from 0 to (num_devices - 1) {
      // get pointer to the device profile and the list of target regs
//KMM      device_profile = parent_device.Host_Bridge'device_function'chain_topology[i];
      device_profile = parent_device.as_a(Host_Bridge'device_function csco_ht_device).chain_topology[i];
      target_regs = device_profile.get_device_regs();

      // Iterate through all of the registers for the device
      for each (reg) in target_regs do {
        // generate random delay between reads
        gen delay keeping {
          it > 0;
        };

        ret = reg.walk_ones_ht(channel);
        wait [delay];
      };
    };

    done = TRUE;
  };
};
////////////////////////////////////////////////////////////////////////////////
// this scenario intermixes reads and writes to all the registers for a HT device
// in the chain from the Host Bridge. the write data is randomly generated and then
// and-ed with the func_write_mask.
////////////////////////////////////////////////////////////////////////////////
extend INTERMIX_REGS'kind csco_ht_scenario {
  // Don't self generate target addresses
  keep target_address_gen_complete == TRUE;

  num_devices : uint; // number of devices in the chain
  // soft constraint, assume at least 1 device in the chain
  keep soft num_devices == 1;

  // list of regs written to but not read
  !outstanding_writes : list (key: reg_name) of csco_ht_basic_reg;

  !target_regs : list of csco_ht_basic_reg; // list of target registers

  generate_transactions()@parent_device.csco_ht_device_clock is {
    var delay          : uint (bits:4);      // random delay
    var ret            : csco_ht_error_code;     // return value from read or write
    var device_profile : csco_ht_device_profile; // Create a profile of the pion

    var trans_cmd : csco_ht_packet_encoding; // transaction cmd to generate;
    var reg_idx   : uint;                // index into list of regs which
                                         // to read or write
    var reg       : csco_ht_basic_reg;       // register to read or write

    for i from 0 to (num_devices - 1) {
//KMM      device_profile = parent_device.Host_Bridge'device_function'chain_topology[i];
      device_profile = parent_device.as_a(Host_Bridge'device_function csco_ht_device).chain_topology[i];

      target_regs = device_profile.get_device_regs();

      while(!target_regs.is_empty() || !outstanding_writes.is_empty()) {
        gen delay keeping {
          it > 0;
        };

        gen trans_cmd keeping {
          it in [Write, Read];
          (target_regs.is_empty()) => it == Read;
        };

        // if the target register list is empty randomly pick
        // a regstier to read from the outstanding_writes list
        if(!target_regs.is_empty()) {
          // randomly generate index into target list of regs
          gen reg_idx keeping {
            it < target_regs.size();
          };

          // pull out reg to write from list and delete
          reg = target_regs[reg_idx];
        } else {
          // randomly generate index into target list of regs
          gen reg_idx keeping {
            it < outstanding_writes.size();
          };

          // pull out reg to write from list and delete
          reg = outstanding_writes[reg_idx];
        };

        case(trans_cmd) {
          Read : {
            // only read if there's an outstanding write or the reg has
            // never been read before.
            if(outstanding_writes.key_exists(reg.reg_name)) {
              reg.read_flag = TRUE;
              outstanding_writes.delete(outstanding_writes.key_index(reg.reg_name));
              ret = reg.read_ht(channel);
            } else if(!reg.read_flag) {
              reg.read_flag = TRUE;
              ret = reg.read_ht(channel);
            } else {
              continue;
            };
          };
          Write : {
            var data    : uint(bits: 64); // random write data
            var wr_mask : uint(bits: 64); // copy of func_wr_mask as a uint;
            var wr_data : uint(bits: 64);

            // only write the register if we havent already
            if(!reg.write_flag) {
              // randomly generate data
              gen data keeping {
                it > 0;
                it < reg.reg_size.as_a(uint);
              };

              // get copy of func_wr_mask in uint form :P
              wr_mask = reg.func_write_mask[:];

              wr_data = (data & wr_mask);

              // mark the register as being written
              reg.write_flag = TRUE;

              if(wr_data != 0) {
                // update shadow copy of reg and push out transaction
                reg.write_local(%{4'hf}, %{wr_data});
                ret = reg.write_ht(channel);

                // add register to the outstanding writes list
                outstanding_writes.add(reg);
              };
            } else {
              continue;
            };
          };
          default : {
            outf("%12d ns %12s(%10s) ERROR: in transaction w/cmd %s gen-ed!\n",
                 sys.time, parent_device.control_block.device_reference_name,
                 parent_device.control_block.device_function, trans_cmd);
          };
        };

        // if the register has been written and read, then take it off
        // the target list.
        if(reg.read_flag && reg.write_flag && !target_regs.is_empty()) {
          target_regs.delete(reg_idx);
        };

        wait[delay];
      };
    };
    done = TRUE;
  };
};
////////////////////////////////////////////////////////////////////////////////
// This is a structure that is used to represent a 32 byte region of memory that has an outstanding read
// or write to it.  This is used by the random traffic generator to ensure that memory accesses to a given
// region are ordered such that they are predicatable.
// Address Decode :
// Doubleword -> 4 bytes -> 32 bits
// Quadword -> 8 bytes -> 64 bits
// Bits [4:0] are all part of a single 32 byte region. Bit[5:5] is used to mark two independent 32 byte regions.
// Bits [5:0] are all part of a single 64 byte region. Bit[6:6] is used to mark two independent 64 byte regions.
////////////////////////////////////////////////////////////////////////////////
struct csco_ht_memory_transaction_element {
  // Use addr[39:5] to index into the memory hierarchy
  addr : uint (bits:35);
  full_addr : uint (bits:40);
  keep full_addr == addr << 5;
  read_outstanding : bool;
  keep read_outstanding == FALSE;
  SeqID_outstanding : uint (bits:4);
};
////////////////////////////////////////////////////////////////////////////////

extend csco_ht_scenario {
  schedule_transaction(transaction : csco_ht_packet) is only {
    parent_device.schedule_packet(channel,transaction );

    if (kind!=INTERMIX && kind!=INTERMIX_OVERLAY && kind!=WRITE_IN_ORDER && kind!=READ_IN_ORDER
        && kind!=READ_WRITE_IN_ORDER && kind!=BATCH_WRITE_BATCH_READ && kind!=INTERRUPT && kind!=READ_POR_REGS
        && kind!=WALK_ONES_REGS && kind!=INTERMIX_REGS && kind!=BASIC_INIT && kind!=FAST_BASIC_INIT) {
      if(parent_device.scoreboard.score_level != csco_ht_no_scoring ) {
        parent_device.scoreboard.post(transaction);
      } else {
        parent_device.score_tx_pkt(transaction);
      };
    };
    transaction_history.add(transaction);
  };
};                                                         // End Extend

'>
